using System;
using System.Diagnostics;
using Zcu.Mve.Core;
using Zcu.Mve.Numerics;

namespace Examples
{
  /// <summary>
  /// Sample module. This demostrates a possibility of module to add and remove ports.
  /// </summary>

  [ModuleInfo("Milan Frank", "Sample module. Example of varying ports.",
     Category = "Example")]

  [PortInfo("input", "Only ScalarNumber and VectorND can be connected.")]
  public class ChangablePorts : Zcu.Mve.Core.Module
  {
    /// <summary>
    /// Adds one default port and registers callbacks called when someone
    /// connects to ports or disconnects.
    /// </summary>
	  public ChangablePorts()
    {
      // Only one default port
      AddInPort("input", typeof(DataObject));

      // Registration of callbacks.
      this.PortConnecting += new PortConnectingEventHandler(OnConnect);
      this.PortDisconnecting += new PortDisconnectingEventHandler(OnDisconnect);
    }

    /// <summary>
    /// Called during connecting to ports.
    /// </summary>
    /// <param name="source"></param>
    /// <param name="portFrom"></param>
    /// <param name="portTo"></param>
    /// <param name="allowed"></param>
    void OnConnect(object source, OutPort portFrom, InPort portTo, out bool allowed) 
    {
      allowed = true;
      switch (portFrom.PortType.Name)
      {
        // Scalar number is about to connect.
        case "ScalarNumber":
          // Check whether output already exists and remove it.
          if (this.ContainsOutPort("output"))
            RemoveOutPort("output");
          // Add new port called output.
          AddOutPort("output", typeof(Scalar));
          break;

        case "VectorND":
          if (this.ContainsInPort("output"))
            RemoveOutPort("output");
          AddOutPort("output", typeof(VectorND));
          break;

        // No other types are allowed.
        default:
          allowed = false;
          break;
      }     
    }

    /// <summary>
    /// Let us remove outport in case of disconnection.
    /// </summary>
    /// <param name="source"></param>
    /// <param name="portFrom"></param>
    /// <param name="portTo"></param>
    /// <param name="allowed"></param>
    void OnDisconnect(object source, OutPort portFrom, InPort portTo, out bool allowed)
    {
      allowed = true;
      if (GetOutPort("output") != null)
        RemoveOutPort("output");      
    }
	
    /// <summary>
    /// Module does nothing.
    /// </summary>
    public override void Execute()
    {
      Console.WriteLine("Module " + this.Name + " do nothing.");
    }
  }
}
