using System;
using System.Globalization;
using Zcu.Mve.Core;
using Zcu.Mve.Numerics;
using System.ComponentModel;

namespace Examples
{
	// properties representing the following Coordinates class will be shown nested in the setup-dialog
	[TypeConverter(typeof(CoordsConverter))]
	public class Coordinates
	{
		private int x;
		private int y;

		public int X
		{
			get { return this.x; }
			set { this.x = value; }
		}

		public int Y
		{
			get { return this.y; }
			set { this.y = value; }
		}
	}

	// TypeConverter is needed to convert the class to string, whitch is needed for the PropertyGrid
	// in setup-dialog
	internal class CoordsConverter : ExpandableObjectConverter 
	{

//-- need for nested property example---------------------------------------------------
		public override bool CanConvertFrom(ITypeDescriptorContext context, Type t) 
		{
			if (t == typeof(string)) 
			{
				return true;
			}
			return base.CanConvertFrom(context, t);
		}

		public override object ConvertFrom(
			ITypeDescriptorContext context, 
			CultureInfo info,
			object value) 
		{
			if (value is string) 
			{
				try 
				{
					string s = (string) value;
					int comma = s.IndexOf(',');
					if (comma != -1) 
					{
						string stringX = s.Substring(0, comma);
						string stringY = s.Substring(comma + 1, s.Length - comma - 1);
						int x = int.Parse(stringX);
						int y = int.Parse(stringY);
						Coordinates c = new Coordinates();
						c.X = x;
						c.Y = y;
						return c;
					}
				}
				catch {}
					throw new ArgumentException("Can not convert '" + (string)value + "' to type Coordinates");
			}
			return base.ConvertFrom(context, info, value);
		}
                                 
		public override object ConvertTo(
			ITypeDescriptorContext context, 
			CultureInfo culture, 
			object value,    
			Type destType) 
		{
			if (destType == typeof(string) && value is Coordinates) 
			{
				Coordinates c = (Coordinates)value;
				return c.X.ToString() + ", " + c.Y.ToString();
			}
			return base.ConvertTo(context, culture, value, destType);
		}
//-- end of: needed for nested property example -----------------------------------------------------

//-- needed for default properties value -----------------------------------------------------
		private static StandardValuesCollection defaultValues = new StandardValuesCollection(
			new string[]{"1, 0", "1, 1", "1, 2", "1, 3", "1, 4"});
		
		public override bool GetStandardValuesSupported(ITypeDescriptorContext context) 
		{
			return true;
		}

		public override bool GetStandardValuesExclusive(
			ITypeDescriptorContext context) 
		{
			// returning false here means the property will
			// have a drop down and a value that can be manually
			// entered.      
			return false;
		}

		public override StandardValuesCollection GetStandardValues(ITypeDescriptorContext context) 
		{
			return defaultValues;
		}
//-- end of: needed for default properties value -----------------------------------------------------
	}


	/// <summary>
	/// Sample module to show automatic setup GUI attributes
	/// </summary>
	[ModuleInfo("Pemek Zitka", "Source of one scalar number", IconName = "numbersource.ico", Assemblied = "2005-05-27")]
	[PortInfo("output", "Vrt slo")]
	[DefaultProperty("Coords")]
	public class GuifiedScalarSource : Zcu.Mve.Core.Module
	{
		internal double number;
		Scalar output = new Scalar(0.0);

		// following variables absolutely nothing, they serve just for GUI demonstration
		internal bool bnumber;
		internal Coordinates coords = new Coordinates();


//-- GUI basic capabilities demonstration ---------------------------------------------------------------

		// PropertyGrid in GuifiedModuleSetup shows all public properties of given modules with the help
		// of few attributes, you can slightly modifie the way your properties are being displayed.
		// At first you must declare "using System.ComponentModel;" in order to be able to use those
		// attributes:

		// This property will be shown in the setup dialog under "Module Setup" category
		// and will have the "The value for output" description.
		[CategoryAttribute("Module Setup"), Description("The value for output.")]
		public double Number
		{
			get { return this.number; }
			set { this.number = value; }
		}

		[CategoryAttribute("Module Setup"), Description("Showing bool representation in setup GUI.")]
		[RefreshProperties(RefreshProperties.Repaint)]
		public bool BoolNumber
		{
			get { return this.bnumber; }
			set { this.bnumber = value; }
		}

		
		[CategoryAttribute("Module Setup"), Description("Showing nested properties representation.")]
		public Coordinates Coords
		{
			get { return this.coords; }
			set { this.coords = value; }
		}

		[CategoryAttribute("Module Info"), Description("Version of the module.")]
		public string Version
		{
			get { return "0.5.1"; }
		}

		[Browsable(false)]
		public int NetSetupProperty
		{
			get { return 0; }
			set { /* ... */ }
		}

		// more and thoroughly described possibilities can be currently found at:
		// http://msdn.microsoft.com/library/default.asp?url=/library/en-us/dndotnet/html/vsnetpropbrow.asp

//-- end of GUI capabilities demonstration -------------------------------------------------------------


		
		public GuifiedScalarSource()
		{
			AddOutPort("output", typeof (Scalar));
		}

		public override void Execute()
		{
			output.Value = number;
			SetOutput("output", output);
		}

		public override void WriteConfig(System.Xml.XmlElement config)
		{
			config.InnerText = number.ToString(Globals.nfi);
		}

		public override void ReadConfig(System.Xml.XmlElement config)
		{
			number = Double.Parse(config.InnerText, Globals.nfi);
		}

		public override Module DeepCopy()
		{
			GuifiedScalarSource module = (GuifiedScalarSource) base.DeepCopy ();
			module.number = this.number;
			return module;
		}
	}
}
