using System;
using System.Collections.Generic;
using Zcu.Mve.Core;
using Zcu.Mve.Numerics;
using Zcu.Mve.Visualization;
using System.Diagnostics;
using System.ComponentModel;

namespace Examples
{
	/// <summary>
	/// A module that computes the number of edges of a mesh. This module is useful for checking of consistence of meshes. Adds a new point attribute to the input mesh, that determines the number of open edges for each vertex.
	/// </summary>
	[ModuleInfo("Libor Va", "A module that computes the number of edges of a mesh. This module is useful for checking of consistence of meshes. Adds a new point attribute to the input mesh, that determines the number of open edges for each vertex.")]
	[PortInfo("TriMesh", "A mesh to be checked.")]
	[PortInfo("OutMesh", "Reference to the input TriMesh.")]
	public class VertexEdgeDetector:Module
	{

        private class LineEqualityComparer : IEqualityComparer<Line>
        {
            public static int pointCount = 1;

            public bool Equals(Line l1, Line l2)
            {
                if ((l1.V1 == l2.V1) && (l1.V2 == l2.V2)) return (true);
                else return (false);
            }

            public int GetHashCode(Line l)
            {
                return (l.V1 * pointCount + l.V2);
            }
        }

		private string triangleCellsName = "";

		private TriangleMesh output;

		private bool paramsChanged = true;

        /// <summary>
        /// Gets or sets the identifier of triangle cells that will be used for computation. 
        /// If set to an empty string, then the first triangle cells will be used.
        /// </summary>
        [Description("Gets or sets the identifier of triangle cells that will be used for computation. If set to an empty string, then the first triangle cells will be used.")]
		public string TriangleCellsName
		{
			get{return(triangleCellsName);}
			set
			{
				triangleCellsName = value;
				paramsChanged = true;
			}
		}

		private string outputAttributeName = "connectivity";

        /// <summary>
        /// Gets or sets the name of the output point attribute.
        /// </summary>
        [Description("Gets or sets the name of the output point attribute.")]
		public string OutputAttributeName
		{
			get{return(outputAttributeName);}
			set
			{
				if (value!="")
				{
					outputAttributeName = value;
					paramsChanged = true;
				}
				else
				{
					throw new MveException(this, "Output attribute name for VertexEdgeDetector cannot be an empty string.");
				}
			}
		}

        /// <summary>
        /// Default constructor.
        /// </summary>
		public VertexEdgeDetector()
		{
			AddInPort("TriMesh",typeof(TriangleMesh));
			AddOutPort("OutMesh", typeof(TriangleMesh));
		}

        /// <summary>
        /// Adds a new point attribute to the input mesh, that determines the number of open edges for each vertex.
        /// Total number of open edges is written to the debug output.
        /// If no parameters and inputs were changed since the last run, then computation is not performed and the last result is returned with dataSame set to true.
        /// </summary>
		public override void Execute()
		{
			if ((paramsChanged)||(!IsAllDataSame())||(output == null))
			{
				TriangleMesh input = (TriangleMesh)GetInput("TriMesh");
				UniformDataArray triangles;
				if (triangleCellsName!="")
					triangles = input.GetCells(triangleCellsName);
				else
					triangles = input.GetCells(typeof(Triangle))[0];
				if (triangles == null)
				{
					throw new MveException(this, "Triangle cells not found.");
				}
	
				LineEqualityComparer.pointCount = input.Points.Count;
                Dictionary<Line,object> lines = new Dictionary<Line,object>(2 * triangles.Count, new LineEqualityComparer());
				foreach (Triangle t in triangles)
				{
					Line l;

					// line 1-2
					if (t.V1<t.V2)
					{
						l.V1 = t.V1;
						l.V2 = t.V2;
					}
					else
					{
						l.V1 = t.V2;
						l.V2 = t.V1;
					}
					if (lines.ContainsKey(l))
						lines.Remove(l);
					else
						lines.Add(l,null);

					// line 1-3
					if (t.V1<t.V3)
					{
						l.V1 = t.V1;
						l.V2 = t.V3;
					}
					else
					{
						l.V1 = t.V3;
						l.V2 = t.V1;
					}
					if (lines.ContainsKey(l))
						lines.Remove(l);
					else
						lines.Add(l,null);

					// line 2-3
					if (t.V2<t.V3)
					{
						l.V1 = t.V2;
						l.V2 = t.V3;
					}
					else
					{
						l.V1 = t.V3;
						l.V2 = t.V2;
					}
					if (lines.ContainsKey(l))
						lines.Remove(l);
					else
						lines.Add(l,null);
				}

				int count = 0;
				int[] counts = new int[input.Points.Count];
				foreach(Line l in lines.Keys)
				{
					counts[l.V1]++;
					counts[l.V2]++;
					count++;
				}

				UniformDataArray attribute = new UniformDataArray(typeof(Integer),input.Points.Count);
				for (int i = 0;i<counts.Length;i++)
				{
					attribute[i] = new Integer(counts[i]);
				}

				Debug.WriteLine("Border edges count: "+count.ToString(Globals.nfi));

				output = (TriangleMesh)input.ShallowCopy();
				output.AddPointAttrs(outputAttributeName,attribute);
				SetOutput("OutMesh",output,false);
				paramsChanged = false;
			}
			else
			{
				SetOutput("OutMesh",output,true);
			}
		}
	}
}
