using System;
using Zcu.Mve.Core;
using Zcu.Mve.Numerics;
using Zcu.Mve.Visualization;
using System.ComponentModel;

namespace Zcu.Mve.Testing
{
    /// <summary>
    /// Module for slicing volume data
    /// </summary>
    /// 
    [ModuleInfo("Libor Va", "Module for slicing volume data.", CategoryType.Filter, CategoryData.Volume, IconName = "OrthoSlicer.ico")]
    [PortInfo("Input array", "MveArray of any kind of data and any dimension to be sliced.")]
    [PortInfo("Output array", "UniformDataArray2D that represent a slice of the input data.")]
    public class OrthoSlicer : Module
    {
        string coordinates = "";

        UniformDataArray2D output = null;

        bool paramsChanged = true;

        /// <summary>
        /// Comma separated integer values that represent a point within the input data through which the slice will pass.
        /// Number of elements must be equal to number of dimensions of the input data.
        /// </summary>
        [Description("Comma separated integer values that represent a point within the input data through which the slice will pass. Number of elements must be equal to number of dimensions of the input data.")]
        public string Coordinates
        {
            get { return (coordinates); }
            set
            {
                coordinates = value;
                paramsChanged = true;
            }
        }

        int widthDim = 0;
        int heightDim = 1;

        /// <summary>
        /// Index of the dimension that will be used as width of the slice.
        /// Must be >=0 and cannot exceed dimension of the input data.
        /// </summary>
        [Description("Index of the dimension that will be used as width of the slice. Must be >=0 and cannot exceed dimension of the input data.")]
        public int WidthDim
        {
            get { return (widthDim); }
            set
            {
                if (value >= 0)
                {
                    widthDim = value;
                    paramsChanged = true;
                }
                else
                {
                    throw new MveException(this, "Value of width dimension cannot be lower than zero.");
                }
            }
        }

        /// <summary>
        /// Index of the dimension that will be used as height of the slice.
        /// Must be >=0 and cannot exceed dimension of the input data.
        /// </summary>
        [Description("Index of the dimension that will be used as height of the slice. Must be >=0 and cannot exceed dimension of the input data.")]
        public int HeightDim
        {
            get { return (heightDim); }
            set
            {
                if (value >= 0)
                {
                    heightDim = value;
                    paramsChanged = true;
                }
                else
                {
                    throw new MveException(this, "Value of height dimension cannot be lower than zero.");
                }
            }
        }

        /// <summary>
        /// Constructor.
        /// </summary>
        public OrthoSlicer()
        {
            AddInPort("Input array", typeof(MveArray));
            AddOutPort("Output array", typeof(UniformDataArray2D));
        }

        /// <inheritdoc />
        public override void Execute()
        {
            if ((paramsChanged) || (!IsDataSame("Input array")) || (output == null))
            {
                if (widthDim == heightDim)
                {
                    throw new MveException(this, "Width dimension cannot be equal to height dimension.");
                }
                string[] coordinateStrings = coordinates.Split(new char[] { ',' });

                MveArray input = (MveArray)GetInput("Input array");

                if (input.ElemType != typeof(Scalar))
                {
                    throw new MveException(this, "Input array element type must be Scalar.");
                }

                if (coordinateStrings.Length != input.Lengths.Length)
                {
                    throw new MveException(this, "Wrong number of coordinates set, must be equal to dimension of the input array.");
                }

                int[] coords = new int[coordinateStrings.Length];

                for (int i = 0; i < coords.Length; i++)
                {
                    coords[i] = Int32.Parse(coordinateStrings[i]);
                    if (coords[i] < 0)
                    {
                        throw new MveException(this, i + ". coordinate lower than zero.");
                    }
                    if (coords[i] >= input.Lengths[i])
                    {
                        throw new MveException(this, i + ". coordinate out of bound.");
                    }
                }

                output = new UniformDataArray2D(typeof(Scalar), input.Lengths[heightDim], input.Lengths[widthDim]);

                for (int x = 0; x < input.Lengths[widthDim]; x++)
                    for (int y = 0; y < input.Lengths[heightDim]; y++)
                    {
                        coords[widthDim] = x;
                        coords[heightDim] = y;
                        output[y, x] = input[coords];
                    }

                SetOutput("Output array", output);
                paramsChanged = false;
            } // recalculation
            else
            {
                SetOutput("Output array", output, true);
            }
        }
    }
}
