using System;
using System.Collections.Generic;
using System.Text;
using Zcu.Mve.Core;
using Zcu.Mve.Visualization;
using Zcu.Mve.Numerics;
using System.Xml;

namespace Zcu.Mve.Visualization.Processing
{
    /// <summary>
    /// Generic module for iteration through UniformDataArray.
    /// Returns an element at a specified index.
    /// Index is specified by the Index port, array is specified by the Array input port.
    /// Output port updates its type upon connecting to output.
    /// Checks for type compatibility, throws runtime error.
    /// </summary>
    [ModuleInfo("Libor Va", "Generic module for iteration through UniformDataArray. Returns an element at a specified index. Index is specified by the Index port, array is specified by the Array input port. Output port updates its type upon connecting to output. Checks for type compatibility, throws runtime error.", CategoryType.Generic, CategoryData.Generic)]
    [PortInfo("Index", "Zero-based index of the desired element of array.")]
    [PortInfo("Array", "The input UniformdataArray.")]
    [PortInfo("Out", "Array element at specified index. Updates port type on first connection.")]
    class ArrayIterator : Module
    {
        private OutPort outPort;
        private IDataObject result = null;

        /// <summary>
        /// Constructs a new ArrayIterator module.
        /// </summary>
        public ArrayIterator()
        {
            AddInPort("Array", typeof(UniformDataArray));
            AddInPort("Index", typeof(Integer));
            AddOutPort("Out", typeof(IDataObject));

            outPort = GetOutPort("Out");

            OutPortDisconnected += new PortDisconnectedEventHandler(ArrayIterator_PortDisconnected);
            OutPortConnecting += new PortConnectingEventHandler(ArrayIterator_PortConnecting);
        }

        /// <summary>
        /// Port Disconnected handler.
        /// </summary>
        /// <param name="source">event source</param>
        /// <param name="portFrom">output port</param>
        /// <param name="portTo">input port</param>
        private void ArrayIterator_PortDisconnected(object source, OutPort portFrom, InPort portTo)
        {
            if (outPort.LinkList.Count == 0)
            {
                outPort.PortType = typeof(IDataObject);
            }
        }

        /// <summary>
        /// Port Connecting handler.
        /// </summary>
        /// <param name="source">event source</param>
        /// <param name="portFrom">output port</param>
        /// <param name="portTo">input port</param>
        /// <param name="allowed">should the connection be allowed?</param>
        private void ArrayIterator_PortConnecting(object source, OutPort portFrom, InPort portTo, out bool allowed)
        {
            if (outPort.PortType == typeof(IDataObject))
            {
                outPort.PortType = portTo.PortType;
            }
            allowed = true;
        }

        /// <inheritdoc />
        public override void Execute()
        {
            if (result != null && IsAllDataSame() && IsAllUserSettingsSame())
            {
                SetOutput("Out", result, true);
            }
            else
            {
                UniformDataArray input = (UniformDataArray)GetInput("Array");
                Integer index = (Integer)GetInput("Index");
                result = input[index];
                SetOutput("Out", result, false);
            }
        }

        /// <inheritdoc />
        public override void WriteConfig(System.Xml.XmlElement config)
        {
            XmlElement typeNameElement = config.OwnerDocument.CreateElement("TypeName");
            typeNameElement.InnerText = outPort.PortType.FullName;
            config.AppendChild(typeNameElement);
        }

        /// <inheritdoc />
        public override void ReadConfig(XmlElement config)
        {
            foreach (XmlElement e in config.ChildNodes)
            {
                if (e.Name == "TypeName")
                {
                    string typeName = e.InnerText;
                    if (typeName.Length > 0)
                    {
                        try
                        {
                            Type type = Globals.FindType(typeName);
                            if (type != null) outPort.PortType = type;
                        }
                        catch (Exception)
                        {
                            WriteDebug("Failed to set output type - possibly missing dll. Specified type: " + typeName);
                        }
                    }
                }
            }
        }
    }
}
