using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Drawing;
using System.IO;
using System.Text;

using Zcu.Mve.Core;
using Zcu.Mve.Numerics;
using Zcu.Mve.Strings;
using Zcu.Mve.Visualization;

namespace Zcu.Mve.Visualization.Loaders
{
    /// <summary>
    /// Reader that can read only one attribute, that can be specified by user.
    /// </summary>
    [ModuleInfo("Josef Vykovsk", "Module for loading raw data from file to RegGridND.",
        CategoryType.Loader,
        CategoryData.Volume,
    Assemblied = "2008-09-02",
    IconName = "BinaryLoader.ico")]
    [PortInfo("OutputDataGrid", "Loaded data in the RegGridND data structure.")]
    [PortInfo("FileName", "Path to the loaded file.")]
    [PortInfo("DataSubsetDefinition", "ND vector of a dimension two times the data dimension. It specifies the data subset.")]
    public class BinaryReaderSingleAttribute:Module, BL_IProgressInfo
    {
        // port name constants
        private readonly string fileNamePort = "FileName";
        private readonly string dataSubsetDefinitionPort = "DataSubsetDefinition";
        private readonly string vifHeaderFilePort = "VifHeaderFile";
        private readonly string outputPort = "OutputDataGrid";

        #region properties

        #region dataFilePath
        /// <summary>
        /// Path to the source file.
        /// </summary>
        private string dataFilePath;
        /// <summary>
        /// Path to the source file.
        /// </summary>
        [Category ("File names")]
        [Description("File to load the data from.")]
        [Editor(typeof(System.Windows.Forms.Design.FileNameEditor), typeof(System.Drawing.Design.UITypeEditor))]
        public string DataFilePath
        {
            get { return dataFilePath; }
            set { dataFilePath = value; }
        }
        #endregion

        #region headerFilePath
        /// <summary>
        /// Path to the header file.
        /// </summary>
        private string headerFilePath;
        /// <summary>
        /// Path to the header file.
        /// </summary>
        [Category("File names")]
        [Description("File to load the data header from.")]
        [Editor(typeof(System.Windows.Forms.Design.FileNameEditor), typeof(System.Drawing.Design.UITypeEditor))]
        public string HeaderFilePath
        {
            get { return headerFilePath; }
            set { headerFilePath = value; }
        }
        #endregion

        #region dataDim
        /// <summary>
        /// Dimension of the data.
        /// </summary>
        private int dataDim = (int)BinaryLoaderGlobalConstants.AllowedDimension._3D_;
        /// <summary>
        /// Dimension of the data.
        /// </summary>
        [Category ("Data settings")]
        [Description ("Dimension of the loaded data grid.\nMust be from {2,3,4}")]
        public BinaryLoaderGlobalConstants.AllowedDimension DataGridDimension
        {
            get { return (BinaryLoaderGlobalConstants.AllowedDimension)dataDim; }
            set { dataDim = (int)value; }
        }
        #endregion

        #region headerSize
        /// <summary>
        /// Size in bytes of an embedded header in the data file.
        /// </summary>
        private int headerSize = 0;
        /// <summary>
        /// Size in bytes of an embedded header in the data file.
        /// </summary>
        [Category ("Data settings")]
        [Description("Size in bytes of an embedded header in the data file.")]
        public int HeaderSize
        {
            get { return headerSize; }
            set { headerSize = value; }
        }
        #endregion

        #region attrType
        /// <summary>
        /// String describing the data attributes. See manual for more info. 
        /// </summary>
        private string attrType;
        /// <summary>
        /// String describing the data attributes. See manual for more info. 
        /// </summary>
        [Category("DataDescription")]
        [Description("Description of the data attributes.\nSyntax: <attribute>;<attribute>\nAttribute syntax:" + 
            "<string atrtibute name>,<string atrtibute type>[<int dimension>]\n" + 
            "Known types: int8, int16, int32, int64, uint8, uint16, unit32, uint64, float32, float64")]
        public BinaryLoaderGlobalConstants.DataTypeNames AttrTypeName
        {
            get { return (BinaryLoaderGlobalConstants.DataTypeNames)Array.IndexOf(BinaryLoaderGlobalConstants.dataTypeNames, attrType); }
            set { attrType = BinaryLoaderGlobalConstants.dataTypeNames[(int)value]; }
        }
        #endregion

        #region attrName
        /// <summary>
        /// String describing the data attributes. See manual for more info. 
        /// </summary>
        private string attrName = "Values";
        /// <summary>
        /// String describing the data attributes. See manual for more info. 
        /// </summary>
        [Category("DataDescription")]
        [Description("Description of the data attributes.\nSyntax: <attribute>;<attribute>\nAttribute syntax:" +
            "<string atrtibute name>,<string atrtibute type>[<int dimension>]\n" +
            "Known types: int8, int16, int32, int64, uint8, uint16, unit32, uint64, float32, float64")]
        public string AttrName
        {
            get { return attrName; }
            set { attrName = value; }
        }
        #endregion

        #region attrDim
        /// <summary>
        /// String describing the data attributes. See manual for more info. 
        /// </summary>
        private string attrDim;
        /// <summary>
        /// String describing the data attributes. See manual for more info. 
        /// </summary>
        [Category("DataDescription")]
        [Description("Description of the data attributes.\nSyntax: <attribute>;<attribute>\nAttribute syntax:" +
            "<string atrtibute name>,<string atrtibute type>[<int dimension>]\n" +
            "Known types: int8, int16, int32, int64, uint8, uint16, unit32, uint64, float32, float64")]
        public string AttrDim
        {
            get { return attrDim; }
            set { attrDim = value; }
        }
        #endregion

        #region invertEndian
        /// <summary>
        /// If the data bits shall be reversed after reading.
        /// </summary>
        private bool invertEndian;
        /// <summary>
        /// If the data bits shall be reversed after reading.
        /// </summary>
        [Category("Misc")]
        [Description("Shall the readed data have their endian inverted?")]
        public bool InvertEndian
        {
            get { return invertEndian; }
            set { invertEndian = value; }
        }
        #endregion

        #region minsS
        /// <summary>
        /// Minimum data space coordinates in each axis. The data samples spans from minsS to maxsS.
        /// </summary>
        private string mins = "0,0,0";
        /// <summary>
        /// Minimum data space coordinates in each axis. The data samples spans from minsS to maxsS.
        /// </summary>
        [Category ("Data settings")]
        [Description ("Minimal coordinate values in each axis. Syntax: <double>;<double>;...;<double>")]
        public string MinimalValues
        {
            get { return mins; }
            set { mins = value; }
        }
        #endregion

        #region maxsS
        /// <summary>
        /// Maximum data space coordinates in each axis. The data samples spans from minsS to maxsS.
        /// </summary>
        private string maxs = "1,1,1";
        /// <summary>
        /// Maximum data space coordinates in each axis. The data samples spans from minsS to maxsS.
        /// </summary>
        [Category("Data settings")]
        [Description("Maximal coordinate values in each axis. Syntax: <double>;<double>;...;<double>")]
        public string MaximalValues
        {
            get { return maxs; }
            set { maxs = value; }
        }
        #endregion

        #region dataSizes
        /// <summary>
        /// Grid dataSizes in each axis. (nuber of data samples)
        /// </summary>
        private string sampleNums;
        /// <summary>
        /// Grid dataSizes in each axis. (nuber of data samples)
        /// </summary>
        [Category("Data settings")]
        [Description("Number of samples in each axis. Syntax: <integer>;<integer>;...;<integer> ")]
        public string NumberOfSamples
        {
            get { return sampleNums; }
            set { sampleNums = value; }
        }
        #endregion

        #region minIndeces
        /// <summary>
        /// Minimal sample indeces of the desired data subset.
        /// </summary>
        private string minIndeces;
        /// <summary>
        /// Minimal sample indeces of the desired data subset.
        /// </summary>
        [Category("Data subset settings")]
        [Description("Minimal indeces of the selection.\nSyntax: <int>;<int>;...;<int>")]
        public string MinIndeces
        {
            get { return minIndeces; }
            set { minIndeces = value; }
        }
        #endregion

        #region maxIndeces
        /// <summary>
        /// Maximal sample indeces of the desired data subset.
        /// </summary>
        private string maxIndeces;
        /// <summary>
        /// Maximal sample indeces of the desired data subset.
        /// </summary>
        [Category("Data subset settings")]
        [Description("Maximal indeces of the selection.\nSyntax: <int>;<int>;...;<int>")]
        public string MaxIndeces
        {
            get { return maxIndeces; }
            set { maxIndeces = value; }
        }
        #endregion

        #region points/cells
        /// <summary>
        /// Specifies wether the attributes will be of point or cell type.
        /// </summary>
        private BinaryLoaderGlobalConstants.PointsOrCells pointOrCells = BinaryLoaderGlobalConstants.PointsOrCells.point;
        /// <summary>
        /// Specifies wether the attributes will be of point or cell type.
        /// </summary>
        [Category("DataDescription")]
        public BinaryLoaderGlobalConstants.PointsOrCells PointOrCells
        {
            get { return pointOrCells; }
            set { pointOrCells = value; }
        }
        #endregion

        #endregion

        private BL_ModuleSettings moduleSetings;
        private BL_ModuleSettings lastSettings;

        private BL_DataDescription dataDescription;

        private VL_InputValidatorAndParser inputValidator;
        private BL_FileReader multipleAttributeFileReader;
        private VifReader vifReader;

        private GridND outputGrid;

        /// <summary>
        /// Creates a new instance of the module and creates its ports.
        /// </summary>
        public BinaryReaderSingleAttribute()
        {
            // Add the input port for file name, non-mandatory.
            AddInPort(fileNamePort, typeof(MveString), false);

            // Add the input port for header file, non-mandatory.
            AddInPort(vifHeaderFilePort, typeof(MveString), false);

            // Add the input port fot data subset definition, non-mandatory.
            AddInPort(dataSubsetDefinitionPort, typeof(VectorND), false);
            
            
            // Add the output port for the resulting data grid.
            AddOutPort(outputPort, typeof(RegGridND));

            lastSettings = null;
        }



        /// <summary>
        /// The execute method - the workhorse of the module.
        /// </summary>
        public override void Execute()
        {
            // Varibles for time measurement.
            DateTime start, end;
            TimeSpan duration;


            // Start the measurement of the module run time.
            start = DateTime.Now;


            // Create needed instances.
            moduleSetings = new BL_ModuleSettings();
            inputValidator = new VL_InputValidatorAndParser();
            multipleAttributeFileReader = new BL_FileReader();
            vifReader = new VifReader();

            // Set the moduleSettings values from the module properties.
            ReadDataDescFromProperties();

            // Override some of the moduleSettings values by values from input ports if connected.
            OverridePropSettingsByPorts();

            OverrideSettingsByVif();

            // Set the file times.
            SetFileTime();

            
            // Retrieve the validated and parsed settings.
            // If the validation fails errors will be reported and an MveException thrown.
            dataDescription = inputValidator.ValidateAndParseInput(moduleSetings);

            // Check, if the loaded data are not same as those previously loaded. If not, load them.
            if (BL_ModuleSettings.AreSettingsSame(moduleSetings, lastSettings) == false)
            {
                // Prepare the file reader for reading.
                multipleAttributeFileReader.SetTheReader(dataDescription, BinaryLoaderGlobalConstants.readerBufferSize, this);

                // Select an approptiate read method and read the file. 
                #region file reading

                // If the data dimension is 2.
                if (dataDescription.DataDim == 2)
                {
                    // If the whole data range shall be readed.
                    if (dataDescription.MinIndeces == null)
                    {
                        multipleAttributeFileReader.ReadWhole2DdataFromFile();
                    }
                    // Else just a data subset will be readed.
                    else
                    {
                        multipleAttributeFileReader.ReadPartial2DdataFromFile();
                    }
                }

                // If the data dimension is 3.
                if (dataDescription.DataDim == 3)
                {
                    // If the whole data range shall be readed.
                    if (dataDescription.MinIndeces == null)
                    {
                        multipleAttributeFileReader.ReadWhole3DdataFromFile();
                    }
                    // Else just a data subset will be readed.
                    else
                    {
                        multipleAttributeFileReader.ReadPartial3DdataFromFile();
                    }
                }

                // If the data dimension is 4.
                if (dataDescription.DataDim == 4)
                {
                    // If the whole data range shall be readed.
                    if (dataDescription.MinIndeces == null)
                    {
                        multipleAttributeFileReader.ReadWhole4DdataFromFile();
                    }
                    // Else just a data subset will be readed.
                    else
                    {
                        multipleAttributeFileReader.ReadPartial4DdataFromFile();
                    }
                }

                #endregion

                // Get the readed data.
                outputGrid = multipleAttributeFileReader.OutputGrid;

                outputGrid.UserTags.Add("FileName", new MveString(dataFilePath));

                // Set the readed data on the output.
                SetOutput(outputPort, outputGrid);
            }
            else
            {
                // Set the data on the port again and set the progress to "completed".
                SetOutput(outputPort, outputGrid, true);
                this.ProgressInfo(1, 0, 1);
            }

            // Set recently used settings as last used.
            lastSettings = moduleSetings;

            // Start the measurement of the module run time.
            end = DateTime.Now;


            // Log the module run time.
            duration = end - start;
            WriteLog("reading time", duration.TotalMilliseconds);
        }



        /// <summary>
        /// Method that make the output data simply accessible for a testing project.
        /// </summary>
        /// <returns>The loaded data.</returns>
        public RegGridND GetOutput()
        {
            return multipleAttributeFileReader.OutputGrid;
        }
         
        /// <summary>
        /// Read the settings data from the module properties.
        /// </summary>
        private void ReadDataDescFromProperties()
        {
            moduleSetings.DataFilePath = dataFilePath;
            moduleSetings.DataDim = dataDim;
            //moduleSetings.HeaderSize = headerSize;
            // Assemble the data describing string from several properties.
            if (attrName != null)
            {
                moduleSetings.DataDesc = attrName + ",";
            }
            if (attrType != null)
            {
                moduleSetings.DataDesc = moduleSetings.DataDesc + attrType + "[" + attrDim.ToString() + "]";
            }
            // Set the array specifying the endian from the property.
            moduleSetings.InvertEndian = new bool[] { invertEndian };
            moduleSetings.Mins = mins;
            moduleSetings.Maxs = maxs;
            moduleSetings.Sizes = sampleNums;
            moduleSetings.MinIndeces = minIndeces;
            moduleSetings.MaxIndeces = maxIndeces;
            moduleSetings.PointOrCells = pointOrCells;
        }

        /// <summary>
        /// Checks if there are any connections to the unmandatory input ports.
        /// If there are, some settings will be readed from the ports rather than from module properties.
        /// </summary>
        private void OverridePropSettingsByPorts()
        {
            // If the port specifying file name is connected.
            if (GetInPort(fileNamePort).Link != null)
            {
                // Set the file name from the port.
                MveString temp = (MveString)GetInput(fileNamePort);
                moduleSetings.DataFilePath = temp.Value;
            }

            // If the port specifying data subset is connected.
            #region code
            if (GetInPort(dataSubsetDefinitionPort).Link != null)
            {
                // Get the input value.
                VectorND tempVector = (VectorND)GetInput(dataSubsetDefinitionPort);

                // Check if the vector is of the right dimension.
                if (tempVector.Length != (dataDim * 2))
                {
                    throw new MveException(this, "Incorrect dimension of data subset definition.");
                }
                else
                {
                    int i;

                    // Create the min indeces describing string.
                    moduleSetings.MinIndeces = "";
                    for (i = 0; i < dataDim; i++)
                    {
                        moduleSetings.MinIndeces += tempVector.Array[i].ToString() + ";";
                    }
                    moduleSetings.MinIndeces = moduleSetings.MinIndeces.Substring(0,
                        moduleSetings.MinIndeces.Length - 1);

                    // Create the max indeces describing string.
                    moduleSetings.MaxIndeces = "";
                    for (i = dataDim; i < dataDim * 2; i++)
                    {
                        moduleSetings.MaxIndeces += tempVector.Array[i].ToString() + ";";
                    }
                    moduleSetings.MaxIndeces = moduleSetings.MaxIndeces.Substring(0,
                        moduleSetings.MaxIndeces.Length - 1);
                }
            }
            #endregion

            // If the port specifying .vif header file is connected.
            if (GetInPort(vifHeaderFilePort).Link != null)
            {
                MveString temp = (MveString)GetInput(vifHeaderFilePort);
                headerFilePath = temp.Value;
            }
        }

        private void OverrideSettingsByVif()
        {
            if (headerFilePath != null && headerFilePath != "")
            {
                vifReader.ReadFromVif(headerFilePath, moduleSetings);
            }
        }
        /// <summary>
        /// Sets the filetimes for the specified file, if it exists.
        /// </summary>
        private void SetFileTime()
        {
            if (File.Exists(moduleSetings.DataFilePath) == true)
            {
                moduleSetings.DataFileCreationTime = File.GetCreationTime(moduleSetings.DataFilePath);
                moduleSetings.DataFileLastWriteTime = File.GetLastWriteTime(moduleSetings.DataFilePath);
            }
        }

        /// <summary>
        /// Method that make ProgressInfo accessible to other classes. It sets it.
        /// </summary>
        /// <param name="progress">the set progress (0.0 - 1.0)</param>
        /// <param name="stage">the stage whichs progress shall be set</param>
        /// <param name="stageCount">progress stage count</param>
        public void SetProgressInfo(double progress, int stage, int stageCount)
        {
            ProgressInfo(progress, stage, stageCount);
        }
    }
}
