using System;
using System.Collections.Generic;
using System.Text;

namespace Zcu.Mve.Visualization.Loaders
{
    /// <summary>
    /// Contains a validated and parsed description of the loaded data.
    /// </summary>
    public class BL_DataDescription
    {

        #region grid settings properties

        private int dataDim;
        /// <summary>
        /// Dimension of the data grid.
        /// </summary>
        public int DataDim
        {
            get { return dataDim; }
        }


        private int[] dataSizes;
        /// <summary>
        /// Numbers of samples per each axis.
        /// </summary>
        public int[] DataSizes
        {
            get { return dataSizes; }
        }


        private int[] extractionSizes;
        /// <summary>
        /// Sizes of the selected data subset. If whole data selected, it is sam as the sizes field.
        /// </summary>
        public int[] ExtractionSizes
        {
            get { return extractionSizes; }
        }


        private double[] mins;
        /// <summary>
        /// Minimal coordinate values per each axis.
        /// </summary>
        public double[] Mins
        {
            get { return mins; }
        }


        private double[] maxs;
        /// <summary>
        /// Maximal coordinate values per each axis.
        /// </summary>
        public double[] Maxs
        {
            get { return maxs; }
        }

        #endregion

        #region subsampling properties

        int[] minIndeces;
        /// <summary>
        /// Minimal indeces of the data sub set.
        /// </summary>
        public int[] MinIndeces
        {
            get { return minIndeces; }
        }

        int[] maxIndeces;
        /// <summary>
        /// Maximal indeces of the data sub set.
        /// </summary>
        public int[] MaxIndeces
        {
            get { return maxIndeces; }
        }

        #endregion

        #region attribute settings properties

        private string[] attributesNames;
        /// <summary>
        /// Names of consequent attributes.
        /// </summary>
        public string[] AttributesNames
        {
            get { return attributesNames; }
        }


        private string[] attributeDataTypeNames;
        /// <summary>
        /// Names of consequent attribute data types.
        /// </summary>
        public string[] AttributeDataTypeNames
        {
            get { return attributeDataTypeNames; }
        }


        private int[] attributesDimensions;
        /// <summary>
        /// Dimensions of consequent attributes.
        /// </summary>
        public int[] AttributesDimensions
        {
            get { return attributesDimensions; }
        }

        #endregion

        #region sample size property

        private int sampleSizeInBytes;
        /// <summary>
        /// Size in bytests of one attribute.
        /// </summary>
        public int SampleSizeInBytes
        {
            get { return sampleSizeInBytes; }
        }

        #endregion

        #region dataFilePath
        
        private string dataFilePath;
        /// <summary>
        /// Path to the file from which the data will be readed.
        /// </summary>
        public string DataFilePath
        {
            get { return dataFilePath; }
            set { dataFilePath = value; }
        }

        #endregion

        #region points/cells

        private BinaryLoaderGlobalConstants.PointsOrCells pointOrCells;
        /// <summary>
        /// Specifies wether the attributes are of point or cell type.
        /// </summary>
        public BinaryLoaderGlobalConstants.PointsOrCells PointOrCells
        {
            get { return pointOrCells; }
        }


        #endregion

        #region invert endian
        private bool[] invertEndian;
        /// <summary>
        /// Specifies for each attribut wheter its endian shall be inverted.
        /// </summary>
        public bool[] InvertEndian
        {
            get { return invertEndian; }
        }
        #endregion

        #region header size

        private int headerSize;
        /// <summary>
        /// Size in bytes of a skipped header in the data.
        /// </summary>
        public int HeaderSize
        {
            get { return headerSize; }
        }
        #endregion

        /// <summary>
        /// Create a new instance of DataDescription.
        /// </summary>
        /// <param name="dataFilePath">Path to the file containg the readed data.</param>
        /// <param name="dataDim">Dimension of the data grid.</param>
        /// <param name="sizes">Numbers of data samples per each axis.</param>
        /// <param name="mins">Minimal space coordinates of the data per each axis.</param>
        /// <param name="maxs">Maximal space coordinates of the data per each axis.</param>
        /// <param name="minIndeces">Minimal indeces of the data subset to be readed. (non-mandataory)</param>
        /// <param name="maxIndeces">Minimal indeces of the data subset to be readed. (non-mandataory)</param>
        /// <param name="attributeNames">Names of the attributes.</param>
        /// <param name="attributeDataTypeNames">Data types of the attributes.</param>
        /// <param name="attributeDimensions">Dimensions of the attributes.</param>
        /// <param name="invertEndians">Array specifiing wether to invert endian of each attribute on reading.</param>
        /// <param name="pointOrCells">Field specifiing if the attributes are point attributes or cell attributes.</param>
        /// <param name="headerSize">Size in bytes of the header in the data file to skip.</param>
        public BL_DataDescription(string dataFilePath, int dataDim, int headerSize, int[] sizes, double[] mins, double[] maxs,
            int[] minIndeces, int[] maxIndeces, string[] attributeNames, string[] attributeDataTypeNames,
            int[] attributeDimensions, bool[] invertEndians, BinaryLoaderGlobalConstants.PointsOrCells pointOrCells)
        {
            this.dataFilePath = dataFilePath;
            this.headerSize = headerSize;

            this.dataDim = dataDim;
            this.dataSizes = sizes;
            this.mins = mins;
            this.maxs = maxs;

            this.minIndeces = minIndeces;
            this.maxIndeces = maxIndeces;

            this.attributesNames = attributeNames;
            this.attributeDataTypeNames = attributeDataTypeNames;
            this.attributesDimensions = attributeDimensions;

            this.invertEndian = invertEndians;
            this.pointOrCells = pointOrCells;

            // If there is no subset specified, the extraction is a whole data set.
            if (this.minIndeces == null)
            {
                extractionSizes = sizes;
            }
            // Else set the extraction according to min and max indeces (including those indeces).
            else
            {
                extractionSizes = new int[sizes.Length];

                for (int i = 0; i < sizes.Length; i++)
                {
                    extractionSizes[i] = maxIndeces[i] - minIndeces[i] + 1;
                }
            }

            // Compute the byte size of a data sample.
            ComputeSampleSize();
        }

        /// <summary>
        /// Computes size in bytes of one data sample (adds dataSizes of all attributes)
        /// </summary>
        private void ComputeSampleSize()
        {
            int i, j;
            sampleSizeInBytes = 0;

            // For each attribute.
            for (i = 0; i < attributesNames.Length; i++)
            {
                // Find its index in the GlobalConstants.dataTypeNames
                for (j = 0; j < BinaryLoaderGlobalConstants.dataTypeNames.Length; j++)
                {
                    // If we have a match, at the size of the attribute to the whole size.
                    if (attributeDataTypeNames[i].CompareTo(BinaryLoaderGlobalConstants.dataTypeNames[j]) == 0)
                    {
                        sampleSizeInBytes += BinaryLoaderGlobalConstants.typesSizesInBytes[j] * attributesDimensions[i];
                    }
                }
            }
        }
    }
}
