using System;
using System.Collections.Generic;
using System.IO;
using System.Text;
using System.Threading;

using Zcu.Mve.Core;
using Zcu.Mve.Numerics;
using Zcu.Mve.Visualization;

namespace Zcu.Mve.Visualization.Loaders
{

    /// <summary>
    /// Class for reading data from file.
    /// </summary>
    public class BL_FileReader
    {

        #region variables and properties

        // Event of reading attributes.
        private event TypeReadHandler ReadAttributes;

        // Delegate for the reading event.
        private delegate void TypeReadHandler();



        private RegGridND outputGrid;
        /// <summary>
        /// The readed data.
        /// </summary>
        public RegGridND OutputGrid
        {
            get { return outputGrid; }
        }


        // Array of type reader instances - one for each attribute.
        private BL_TypeReader[] typeReaders;

        // Data description instancce reference describing the readed data
        private BL_DataDescription dataDescription;


        // Filestream to read the data from.
        private FileStream fs;

        // Reader of the filestream.
        private BinaryReader streamReader;

        // Size of the filestream buffer.
        private int readerBufferSize;

        // Reference to the module owning the instance of GenericAttributeFileReader.
        private BL_IProgressInfo ownerModule;

        #endregion


        /// <summary>
        /// Sets the reader before reading.
        /// </summary>
        /// <param name="dataDescription">Description of the readed data.</param>
        /// <param name="readerBufferSize">Size in bytes of used filestream buffer.</param>
        /// <param name="ownerModule">Module that owns this reader.</param>
        public void SetTheReader(BL_DataDescription dataDescription, int readerBufferSize, BL_IProgressInfo ownerModule)
        {

            // Propagate the method parameters to instance variables.
            this.dataDescription = dataDescription;
            this.readerBufferSize = readerBufferSize;
            this.ownerModule = ownerModule;

            // Create cell attributes accordin to recieved data description.
            if (dataDescription.PointOrCells == BinaryLoaderGlobalConstants.PointsOrCells.cell)
            {
                outputGrid = new RegGridND(dataDescription.Mins, dataDescription.Maxs, dataDescription.ExtractionSizes);
                AddCellAttributes(dataDescription);
            }
            else
            {
                int[] temp = new int[dataDescription.ExtractionSizes.Length];

                for (int i = 0; i < temp.Length; i++)
                {
                    temp[i] = dataDescription.ExtractionSizes[i] - 1;
                }

                outputGrid = new RegGridND(dataDescription.Mins, dataDescription.Maxs, temp);
                AddPointAttributes(dataDescription);
            }

            // Create appropriate type readers.
            CreateTypeReaders(dataDescription);
        }



        /////////////////////////////////////////////
        // methods for reading the whole data file //
        /////////////////////////////////////////////

        /// <summary>
        /// Reads 2D data from the file.
        /// </summary>
        public void ReadWhole2DdataFromFile()
        {

            // Itreration variables.
            int i, xIter, yIter, xIterMax, yIterMax;

            //Create an index array that will contain indeces to write to.
            int[] indeces = new int[2];

            // Set the maximal iteration counts in each axis.
            xIterMax = dataDescription.DataSizes[0];
            yIterMax = dataDescription.DataSizes[1];


            // Try to read the data.
            try
            {
                // Create the filestream and its binary reader.
                fs = new FileStream(dataDescription.DataFilePath, FileMode.Open, FileAccess.Read,
                    FileShare.Read, BinaryLoaderGlobalConstants.readerBufferSize);

                streamReader = new BinaryReader(fs);


                // Set references to the stream and indeces for each reader.
                for (i = 0; i < dataDescription.AttributesNames.Length; i++)
                {
                    typeReaders[i].SetTheStreamReader(streamReader);
                    typeReaders[i].IndecesToWriteTo = indeces;
                }

                // Skip the header.
                fs.Seek(dataDescription.HeaderSize, SeekOrigin.Begin);

                // For each row paraelel with the x axis.
                for (yIter = 0; yIter < yIterMax; yIter++)
                {

                    // Report the readers progress.
                    ownerModule.SetProgressInfo(yIter / (float)yIterMax, 0, 1);


                    // Set the y index.
                    indeces[1] = yIter;


                    // For each data cell in the row.
                    for (xIter = 0; xIter < xIterMax; xIter++)
                    {
                        // Set the x index.
                        indeces[0] = xIter;

                        // Read all attributes for the data cell.
                        ReadAttributes();
                    }
                }


                // Close the stream and its reader.
                streamReader.Close();
                fs.Close();
            }

            // If any exception occured during the data reading, propagate it as an MveException.
            catch (Exception e)
            {
                throw new MveException(this, "An error occured during reading the file.", e);
            }
        }

        /// <summary>
        /// Reads 3D data from the file.
        /// </summary>
        public void ReadWhole3DdataFromFile()
        {

            // Itreration variables.
            int i, xIter, yIter, zIter, xIterMax, yIterMax, zIterMax;

            //Create an index array that will contain indeces to write to.
            int[] indeces = new int[3];

            // Set the maximal iteration counts in each axis.
            xIterMax = dataDescription.DataSizes[0];
            yIterMax = dataDescription.DataSizes[1];
            zIterMax = dataDescription.DataSizes[2];

            int dataChunkSize = dataDescription.SampleSizeInBytes * xIterMax;

            // Try to read the data.
            try
            {
                // Create the filestream and its binary reader.
                fs = new FileStream(dataDescription.DataFilePath, FileMode.Open, FileAccess.Read,
                    FileShare.Read, BinaryLoaderGlobalConstants.readerBufferSize);

                streamReader = new BinaryReader(fs);


                // Set references to the stream and indeces for each reader.
                for (i = 0; i < dataDescription.AttributesNames.Length; i++)
                {
                    typeReaders[i].SetTheStreamReader(streamReader);
                    typeReaders[i].IndecesToWriteTo = indeces;
                }

                // Skip the header.
                fs.Seek(dataDescription.HeaderSize, SeekOrigin.Begin);

                // For each data plane paralel with the xy plane in the data volume.
                for (zIter = 0; zIter < zIterMax; zIter++)
                {
                    // Report the readers progress.
                    ownerModule.SetProgressInfo(zIter / (float)zIterMax, 0, 1);

                    // Set the z index.
                    indeces[2] = zIter;


                    // For each row paraelel with the x axis.
                    for (yIter = 0; yIter < yIterMax; yIter++)
                    {
                        // Set the y index.
                        indeces[1] = yIter;


                        // For each data cell in the row.
                        for (xIter = 0; xIter < xIterMax; xIter++)
                        {
                            // Set the x index.
                            indeces[0] = xIter;

                            // Read all attributes for the data cell.
                            ReadAttributes();
                        }
                    }
                }


                // Close the stream and its reader.
                streamReader.Close();
                fs.Close();

            }

            // If any exception occured during the data reading, propagate it as an MveException.
            catch (Exception e)
            {
                throw new MveException(this, "An error occured during reading the file.", e);
            }
        }

        /// <summary>
        /// Reads 4D data from the file.
        /// </summary>
        public void ReadWhole4DdataFromFile()
        {
            // Itreration variables.
            int i, xIter, yIter, zIter, tIter, xIterMax, yIterMax, zIterMax, tIterMax;

            //Create an index array that will contain indeces to write to.
            int[] indeces = new int[4];

            // Set the maximal iteration counts in each axis.
            xIterMax = dataDescription.DataSizes[0];
            yIterMax = dataDescription.DataSizes[1];
            zIterMax = dataDescription.DataSizes[2];
            tIterMax = dataDescription.DataSizes[3];


            // Try to read the data.
            try
            {
                // Create the filestream and its binary reader.
                fs = new FileStream(dataDescription.DataFilePath, FileMode.Open, FileAccess.Read,
                    FileShare.Read, BinaryLoaderGlobalConstants.readerBufferSize);

                streamReader = new BinaryReader(fs);


                // Set references to the stream and indeces for each reader.
                for (i = 0; i < dataDescription.AttributesNames.Length; i++)
                {
                    typeReaders[i].SetTheStreamReader(streamReader);
                    typeReaders[i].IndecesToWriteTo = indeces;
                }

                // Skip the header.
                fs.Seek(dataDescription.HeaderSize, SeekOrigin.Begin);

                // For each timeframe of volume data.
                for (tIter = 0; tIter < tIterMax; tIter++)
                {

                    // Report the readers progress.
                    ownerModule.SetProgressInfo(tIter / (float)tIterMax, 0, 1);

                    // Set the t index.
                    indeces[3] = tIter;

                    // For each data plane paralel with the xy plane in the data volume.
                    for (zIter = 0; zIter < zIterMax; zIter++)
                    {
                        // Set the z index.
                        indeces[2] = zIter;


                        // For each row paraelel with the x axis.
                        for (yIter = 0; yIter < yIterMax; yIter++)
                        {
                            // Set the y index.
                            indeces[1] = yIter;


                            // For each data cell in the row.
                            for (xIter = 0; xIter < xIterMax; xIter++)
                            {
                                // Set the x index.
                                indeces[0] = xIter;

                                // Read all attributes for the data cell.
                                ReadAttributes();
                            }
                        }
                    }
                }

                // Close the stream and its reader.
                streamReader.Close();
                fs.Close();
            }

            // If any exception occured during the data reading, propagate it as an MveException.
            catch (Exception e)
            {
                throw new MveException(this, "An error occured during reading the file.", e);
            }
        }




        /////////////////////////////////////////////////////
        // methods for reading a data subset from the file //
        /////////////////////////////////////////////////////

        /// <summary>
        /// Reads part of 2D data from the file.
        /// </summary>
        public void ReadPartial2DdataFromFile()
        {
            // Itreration variables.
            int i, xIter, yIter, xIterMax, yIterMax;

            // Create an index array that will contain indeces to write to.
            int[] indeces = new int[2];

            // Seek offsets.
            long bytesToDataStart, bytesToNextRow;

            // Set the maximal iteration counts in each axis.
            xIterMax = dataDescription.MaxIndeces[0] - dataDescription.MinIndeces[0] + 1;
            yIterMax = dataDescription.MaxIndeces[1] - dataDescription.MinIndeces[1] + 1;

            // Set the seek offsets.
            bytesToDataStart = (dataDescription.MinIndeces[0] +
                dataDescription.MinIndeces[1] * dataDescription.DataSizes[0]) *
                dataDescription.SampleSizeInBytes;
            // Add the header size.
            bytesToDataStart = bytesToDataStart + dataDescription.HeaderSize;

            bytesToNextRow = (dataDescription.DataSizes[0]) * dataDescription.SampleSizeInBytes
                - ((dataDescription.MaxIndeces[0] - dataDescription.MinIndeces[0] + 1) * dataDescription.SampleSizeInBytes);



            // Try to read the data.
            try
            {
                // Create the filestream and its binary reader.
                fs = new FileStream(dataDescription.DataFilePath, FileMode.Open, FileAccess.Read,
                    FileShare.Read, BinaryLoaderGlobalConstants.readerBufferSize);

                streamReader = new BinaryReader(fs);


                // Set references to the stream and indeces for each reader.
                for (i = 0; i < dataDescription.AttributesNames.Length; i++)
                {
                    typeReaders[i].SetTheStreamReader(streamReader);
                    typeReaders[i].IndecesToWriteTo = indeces;
                }

                // Seek to the subset begining.
                fs.Seek(bytesToDataStart, SeekOrigin.Begin);




                // For each row paraelel with the x axis.
                for (yIter = 0; yIter < yIterMax; yIter++)
                {

                    // Report the readers progress.
                    ownerModule.SetProgressInfo(yIter / (float)yIterMax, 0, 1);

                    // Set the y index.
                    indeces[1] = yIter;


                    // For each data cell in the row.
                    for (xIter = 0; xIter < xIterMax; xIter++)
                    {
                        // Set the x index.
                        indeces[0] = xIter;

                        // Read all attributes for the data cell.
                        ReadAttributes();
                    }

                    // If we didn't read the last row, seek to the next row.
                    if (yIter != yIterMax - 1)
                    {
                        fs.Seek(bytesToNextRow, SeekOrigin.Current);
                    }
                }


                // Close the stream and its reader.
                streamReader.Close();
                fs.Close();
            }

            // If any exception occured during the data reading, propagate it as an MveException.
            catch (Exception e)
            {
                throw new MveException(this, "An error occured during reading the file.", e);
            }
        }

        /// <summary>
        /// Reads part of 3D data from the file.
        /// </summary>
        public void ReadPartial3DdataFromFile()
        {
            // Itreration variables.
            int i, xIter, yIter, zIter, xIterMax, yIterMax, zIterMax;

            // Create an index array that will contain indeces to write to.
            int[] indeces = new int[3];

            // Seek offsets.
            long bytesToDataStart, bytesToNextRow, bytesToNextPlane;



            // Set the maximal iteration counts in each axis.
            xIterMax = dataDescription.MaxIndeces[0] - dataDescription.MinIndeces[0] + 1;
            yIterMax = dataDescription.MaxIndeces[1] - dataDescription.MinIndeces[1] + 1;
            zIterMax = dataDescription.MaxIndeces[2] - dataDescription.MinIndeces[2] + 1;

            // Set the seek offsets.
            bytesToDataStart = (dataDescription.MinIndeces[0] +
                dataDescription.MinIndeces[1] * dataDescription.DataSizes[0] +
                dataDescription.MinIndeces[2] * dataDescription.DataSizes[0] * dataDescription.DataSizes[1]) *
                dataDescription.SampleSizeInBytes;
            // Add the header size.
            bytesToDataStart = bytesToDataStart + dataDescription.HeaderSize;

            // Set the byte distance from the end of a row to the start of the next one.
            bytesToNextRow = (dataDescription.DataSizes[0]) * dataDescription.SampleSizeInBytes
                - ((dataDescription.MaxIndeces[0] - dataDescription.MinIndeces[0] + 1) * dataDescription.SampleSizeInBytes);

            // Set the byte distance from the end of a plane to the start of the next one.
            bytesToNextPlane = ((dataDescription.DataSizes[0] * dataDescription.DataSizes[1]) *
                dataDescription.SampleSizeInBytes) -
                ((dataDescription.MaxIndeces[1] - dataDescription.MinIndeces[1]) * dataDescription.DataSizes[0] * dataDescription.SampleSizeInBytes) -
                ((dataDescription.MaxIndeces[0] - dataDescription.MinIndeces[0] + 1) * dataDescription.SampleSizeInBytes);



            // Try to read the data.
            try
            {
                // Create the filestream and its binary reader.
                fs = new FileStream(dataDescription.DataFilePath, FileMode.Open, FileAccess.Read,
                    FileShare.Read, BinaryLoaderGlobalConstants.readerBufferSize);

                streamReader = new BinaryReader(fs);


                // Set references to the stream and indeces for each reader.
                for (i = 0; i < dataDescription.AttributesNames.Length; i++)
                {
                    typeReaders[i].SetTheStreamReader(streamReader);
                    typeReaders[i].IndecesToWriteTo = indeces;
                }


                // Seek to the subset begining.
                fs.Seek(bytesToDataStart, SeekOrigin.Begin);



                // For each data plane paralel with the xy plane in the data volume.
                for (zIter = 0; zIter < zIterMax; zIter++)
                {

                    // Report the readers progress.
                    ownerModule.SetProgressInfo(zIter / (float)zIterMax, 0, 1);

                    // Set the z index.
                    indeces[2] = zIter;


                    // For each row paraelel with the x axis.
                    for (yIter = 0; yIter < yIterMax; yIter++)
                    {
                        // Set the y index.
                        indeces[1] = yIter;


                        // For each data cell in the row.
                        for (xIter = 0; xIter < xIterMax; xIter++)
                        {
                            // Set the x index.
                            indeces[0] = xIter;

                            // Read all attributes for the data cell.
                            ReadAttributes();
                        }

                        // If we didn't read the last row, seek to the next row.
                        if (yIter != yIterMax - 1)
                        {
                            fs.Seek(bytesToNextRow, SeekOrigin.Current);
                        }
                    }

                    // If we didn't read the last plane, seek to the next plane.
                    if (zIter != zIterMax - 1)
                    {
                        fs.Seek(bytesToNextPlane, SeekOrigin.Current);
                    }
                }


                // Close the stream and its reader.
                streamReader.Close();
                fs.Close();
            }

            // If any exception occured during the data reading, propagate it as an MveException.
            catch (Exception e)
            {
                throw new MveException(this, "An error occured during reading the file.", e);
            }
        }

        /// <summary>
        /// Reads part of 4D data from the file.
        /// </summary>
        public void ReadPartial4DdataFromFile()
        {
            // Itreration variables.
            int i, xIter, yIter, zIter, tIter, xIterMax, yIterMax, zIterMax, tIterMax;

            // Create an index array that will contain indeces to write to.
            int[] indeces = new int[4];

            // Seek offsets.
            long bytesToDataStart, bytesToNextRow, bytesToNextPlane, byteSizeOfAVolume;



            // Set the maximal iteration counts in each axis.
            xIterMax = dataDescription.MaxIndeces[0] - dataDescription.MinIndeces[0] + 1;
            yIterMax = dataDescription.MaxIndeces[1] - dataDescription.MinIndeces[1] + 1;
            zIterMax = dataDescription.MaxIndeces[2] - dataDescription.MinIndeces[2] + 1;
            tIterMax = dataDescription.MaxIndeces[3] - dataDescription.MinIndeces[3] + 1;

            // Set the seek offsets.
            bytesToDataStart = (dataDescription.MinIndeces[0] +
                dataDescription.MinIndeces[1] * dataDescription.DataSizes[0] +
                dataDescription.MinIndeces[2] * dataDescription.DataSizes[1] * dataDescription.DataSizes[0] +
                dataDescription.MinIndeces[3] * dataDescription.DataSizes[2] * dataDescription.DataSizes[1] * dataDescription.DataSizes[0]) *
                dataDescription.SampleSizeInBytes;
            // Add the header size.
            bytesToDataStart = bytesToDataStart + dataDescription.HeaderSize;

            // Set the byte distance from the end of a row to the start of the next one.
            bytesToNextRow = (dataDescription.DataSizes[0]) * dataDescription.SampleSizeInBytes
                - ((dataDescription.MaxIndeces[0] - dataDescription.MinIndeces[0] + 1) * dataDescription.SampleSizeInBytes);

            // Set the byte distance from the end of a plane to the start of the next one.
            bytesToNextPlane = ((dataDescription.DataSizes[0] * dataDescription.DataSizes[1]) *
                dataDescription.SampleSizeInBytes) -
                ((dataDescription.MaxIndeces[1] - dataDescription.MinIndeces[1]) * dataDescription.DataSizes[0] * dataDescription.SampleSizeInBytes) -
                ((dataDescription.MaxIndeces[0] - dataDescription.MinIndeces[0] + 1) * dataDescription.SampleSizeInBytes);



            byteSizeOfAVolume = (dataDescription.DataSizes[0] * dataDescription.DataSizes[1] * dataDescription.DataSizes[2])
                * dataDescription.SampleSizeInBytes;



            // Try to read the data.
            try
            {
                // Create the filestream and its binary reader.
                fs = new FileStream(dataDescription.DataFilePath, FileMode.Open, FileAccess.Read,
                    FileShare.Read, BinaryLoaderGlobalConstants.readerBufferSize);

                streamReader = new BinaryReader(fs);


                // Set references to the stream and indeces for each reader.
                for (i = 0; i < dataDescription.AttributesNames.Length; i++)
                {
                    typeReaders[i].SetTheStreamReader(streamReader);
                    typeReaders[i].IndecesToWriteTo = indeces;
                }


                // Seek to the subset begining.
                fs.Seek(bytesToDataStart, SeekOrigin.Begin);

                // For each timeframe of volume data.
                for (tIter = 0; tIter < tIterMax; tIter++)
                {

                    // Report the readers progress.
                    ownerModule.SetProgressInfo(tIter / (float)tIterMax, 0, 1);

                    // Set the t index.
                    indeces[3] = tIter;

                    // For each data plane paralel with the xy plane in the data volume.
                    for (zIter = 0; zIter < zIterMax; zIter++)
                    {
                        // Set the z index.
                        indeces[2] = zIter;


                        // For each row paraelel with the x axis.
                        for (yIter = 0; yIter < yIterMax; yIter++)
                        {
                            // Set the y index.
                            indeces[1] = yIter;


                            // For each data cell in the row.
                            for (xIter = 0; xIter < xIterMax; xIter++)
                            {
                                // Set the x index.
                                indeces[0] = xIter;

                                // Read all attributes for the data cell.
                                ReadAttributes();
                            }

                            // If we didn't read the last row, seek to the next row.
                            if (yIter != yIterMax - 1)
                            {
                                fs.Seek(bytesToNextRow, SeekOrigin.Current);
                            }
                        }

                        // If we didn't read the last plane, seek to the next plane.
                        if (zIter != zIterMax - 1)
                        {
                            fs.Seek(bytesToNextPlane, SeekOrigin.Current);
                        }
                    }

                    // If we didn't read the last timeframe, seek to the next one.
                    if (tIter != tIterMax - 1)
                    {
                        fs.Seek(byteSizeOfAVolume * (tIter + 1) + bytesToDataStart, SeekOrigin.Begin);
                    }
                }

                // Close the stream and its reader.
                streamReader.Close();
                fs.Close();
            }

            // If any exception occured during the data reading, propagate it as an MveException.
            catch (Exception e)
            {
                throw new MveException(this, "An error occured during reading the file.", e);
            }
        }



        //////////////////////
        // helpfull methods //
        //////////////////////

        /// <summary>
        /// Adds cell attributes to the grid according to attributes stored in dataDescription.
        /// </summary>
        /// <param name="dataDescription">DataDescription instance describing the loaded data.</param>
        private void AddCellAttributes(BL_DataDescription dataDescription)
        {
            // For each attribute.
            for (int i = 0; i < dataDescription.AttributesNames.Length; i++)
            {

                // If it is a scalar attribute.
                if (dataDescription.AttributesDimensions[i] == 1)
                {
                    if (dataDescription.AttributeDataTypeNames[i].Contains("int") == true)
                    {
                        OutputGrid.AddPointAttrs(dataDescription.AttributesNames[i],
                         new MveArray(typeof(Integer), dataDescription.ExtractionSizes));
                    }
                    else
                    {
                        OutputGrid.AddPointAttrs(dataDescription.AttributesNames[i],
                            new MveArray(typeof(Scalar), dataDescription.ExtractionSizes));
                    }
                }

                // If it is a 2D vector attribute.
                else if (dataDescription.AttributesDimensions[i] == 2)
                {
                    OutputGrid.AddCellAttrs(dataDescription.AttributesNames[i],
                        new MveArray(typeof(Vector2D), dataDescription.ExtractionSizes));
                }

                // If it is a 3D vector attribute.
                else if (dataDescription.AttributesDimensions[i] == 3)
                {
                    OutputGrid.AddCellAttrs(dataDescription.AttributesNames[i],
                        new MveArray(typeof(Vector3D), dataDescription.ExtractionSizes));
                }

                // If it is a ND vector attribute.
                else
                {
                    OutputGrid.AddCellAttrs(dataDescription.AttributesNames[i],
                        new MveArray(typeof(VectorND), dataDescription.ExtractionSizes));
                }
            }
        }

        /// <summary>
        /// Adds point attributes to the grid according to attributes stored in dataDescription.
        /// </summary>
        /// <param name="dataDescription">DataDescription instance describing the loaded data.</param>
        private void AddPointAttributes(BL_DataDescription dataDescription)
        {
            // For each attribute.
            for (int i = 0; i < dataDescription.AttributesNames.Length; i++)
            {

                // If it is a scalar attribute.
                if (dataDescription.AttributesDimensions[i] == 1)
                {
                    if (dataDescription.AttributeDataTypeNames[i].Contains("int") == true)
                    {
                        OutputGrid.AddPointAttrs(dataDescription.AttributesNames[i],
                         new MveArray(typeof(Integer), dataDescription.ExtractionSizes));
                    }
                    else
                    {
                        OutputGrid.AddPointAttrs(dataDescription.AttributesNames[i],
                            new MveArray(typeof(Scalar), dataDescription.ExtractionSizes));
                    }
                }

                // If it is a 2D vector attribute.
                else if (dataDescription.AttributesDimensions[i] == 2)
                {
                    OutputGrid.AddPointAttrs(dataDescription.AttributesNames[i],
                        new MveArray(typeof(Vector2D), dataDescription.ExtractionSizes));
                }

                // If it is a 3D vector attribute.
                else if (dataDescription.AttributesDimensions[i] == 3)
                {
                    OutputGrid.AddPointAttrs(dataDescription.AttributesNames[i],
                        new MveArray(typeof(Vector3D), dataDescription.ExtractionSizes));
                }

                // If it is a ND vector attribute.
                else
                {
                    OutputGrid.AddPointAttrs(dataDescription.AttributesNames[i],
                        new MveArray(typeof(VectorND), dataDescription.ExtractionSizes));
                }
            }
        }

        /// <summary>
        /// Creates typereaders according to readed attributes and adds appropriate event handlers.
        /// </summary>
        /// <param name="dataDescription">DataDescription instance describing the loaded data.</param>
        private void CreateTypeReaders(BL_DataDescription dataDescription)
        {

            // Create a new array of type readers.
            typeReaders = new BL_TypeReader[dataDescription.AttributesNames.Length];


            // Create each reder and set an event handler according to it.
            for (int i = 0; i < dataDescription.AttributesNames.Length; i++)
            {
                if (dataDescription.PointOrCells == BinaryLoaderGlobalConstants.PointsOrCells.cell)
                {
                    typeReaders[i] = new BL_TypeReader(dataDescription.AttributesNames[i],
                        dataDescription.AttributeDataTypeNames[i], dataDescription.AttributesDimensions[i],
                        OutputGrid.GetCellAttrs(dataDescription.AttributesNames[i]),
                        dataDescription.InvertEndian[i]);

                }
                else
                {
                    typeReaders[i] = new BL_TypeReader(dataDescription.AttributesNames[i],
                     dataDescription.AttributeDataTypeNames[i], dataDescription.AttributesDimensions[i],
                     OutputGrid.GetPointAttrs(dataDescription.AttributesNames[i]),
                     dataDescription.InvertEndian[i]);
                }
                // Set the handler.
                AddTypeReadHandler(typeReaders[i]);
            }
        }

        /// <summary>
        /// Sets event handler according to specified type reader.
        /// </summary>
        /// <param name="reader">The type reader.</param>
        private void AddTypeReadHandler(BL_TypeReader reader)
        {
            // If the reader will read uint8 type.
            if (reader.AttrType.CompareTo("uint8") == 0)
            {
                SetUint8Reader(reader);
            }

            // Else if the reader will read int8 type.
            else if (reader.AttrType.CompareTo("int8") == 0)
            {
                SetInt8Reader(reader);
            }

            // Else if the reader will read uint16 type.
            else if (reader.AttrType.CompareTo("uint16") == 0)
            {
                if (reader.InvertEndian == true)
                {
                    SetUint16ReaderInv(reader);
                }
                else
                {
                    SetUint16Reader(reader);
                }
            }

            // Else if the reader will read int16 type.
            else if (reader.AttrType.CompareTo("int16") == 0)
            {
                if (reader.InvertEndian == true)
                {
                    SetInt16ReaderInv(reader);
                }
                else
                {
                    SetInt16Reader(reader);
                }
            }

            // Else if the reader will read uint32 type.
            else if (reader.AttrType.CompareTo("uint32") == 0)
            {
                if (reader.InvertEndian == true)
                {
                    SetUint32ReaderInv(reader);
                }
                else
                {
                    SetUint32Reader(reader);
                }
            }

            // Else if the reader will read int32 type.
            else if (reader.AttrType.CompareTo("int32") == 0)
            {
                if (reader.InvertEndian == true)
                {
                    SetInt32ReaderInv(reader);
                }
                else
                {
                    SetInt32Reader(reader);
                }
            }

            // Else if the reader will read uint64 type.
            else if (reader.AttrType.CompareTo("uint64") == 0)
            {
                if (reader.InvertEndian == true)
                {
                    SetUint64ReaderInv(reader);
                }
                else
                {
                    SetUint64Reader(reader);
                }
            }

            // Else if the reader will read int64 type.
            else if (reader.AttrType.CompareTo("int64") == 0)
            {
                if (reader.InvertEndian == true)
                {
                    SetInt64ReaderInv(reader);
                }
                else
                {
                    SetInt64Reader(reader);
                }
            }

            // Else if the reader will read float32 type.
            else if (reader.AttrType.CompareTo("float32") == 0)
            {
                if (reader.InvertEndian == true)
                {
                    SetFloat32ReaderInv(reader);
                }
                else
                {
                    SetFloat32Reader(reader);
                }
            }

            // Else if the reader will read float64 type.
            else if (reader.AttrType.CompareTo("float64") == 0)
            {
                if (reader.InvertEndian == true)
                {
                    SetFloat64ReaderInv(reader);
                }
                else
                {
                    SetFloat64Reader(reader);
                }
            }
        }




        ////////////////////////////
        // reader setting methods //
        ////////////////////////////


        //////////////////////////
        // 1 byte types readers //
        //////////////////////////

        /// <summary>
        /// Adds an uint8 reading method of a reader to respond to the event Read.
        /// The own method is sellected depending on data dimension.
        /// </summary>
        /// <param name="reader">The reader.</param>
        private void SetUint8Reader(BL_TypeReader reader)
        {
            // Select a specific method according to attribute type and dimension.
            if (reader.AttrTypeDimension == 1)
            {
                ReadAttributes += new TypeReadHandler(reader.ReadUInt8scal);
            }
            else if (reader.AttrTypeDimension == 2)
            {
                ReadAttributes += new TypeReadHandler(reader.ReadUInt8vec2);
            }
            else if (reader.AttrTypeDimension == 3)
            {
                ReadAttributes += new TypeReadHandler(reader.ReadUInt8vec3);
            }
            else
            {
                ReadAttributes += new TypeReadHandler(reader.ReadUInt8vecN);
            }
        }

        /// <summary>
        /// Adds an int8 reading method of a reader to respond to the event Read.
        /// The own method is sellected depending on data dimension.
        /// </summary>
        /// <param name="reader">The reader.</param>
        private void SetInt8Reader(BL_TypeReader reader)
        {
            // Select a specific method according to attribute type and dimension.
            if (reader.AttrTypeDimension == 1)
            {
                ReadAttributes += new TypeReadHandler(reader.ReadInt8scal);
            }
            else if (reader.AttrTypeDimension == 2)
            {
                ReadAttributes += new TypeReadHandler(reader.ReadInt8vec2);
            }
            else if (reader.AttrTypeDimension == 3)
            {
                ReadAttributes += new TypeReadHandler(reader.ReadInt8vec3);
            }
            else
            {
                ReadAttributes += new TypeReadHandler(reader.readInt8vecN);
            }
        }



        //////////////////////////
        // 2 byte types readers //
        //////////////////////////

        /// <summary>
        /// Adds an uint16 reading method of a reader to respond to the event Read.
        /// The own method is sellected depending on data dimension.
        /// </summary>
        /// <param name="reader">The reader.</param>
        private void SetUint16Reader(BL_TypeReader reader)
        {
            // Select a specific method according to attribute type and dimension.
            if (reader.AttrTypeDimension == 1)
            {
                ReadAttributes += new TypeReadHandler(reader.ReadUInt16scal);
            }
            else if (reader.AttrTypeDimension == 2)
            {
                ReadAttributes += new TypeReadHandler(reader.ReadUInt16vec2);
            }
            else if (reader.AttrTypeDimension == 3)
            {
                ReadAttributes += new TypeReadHandler(reader.ReadUInt16vec3);
            }
            else
            {
                ReadAttributes += new TypeReadHandler(reader.ReadUInt16vecN);
            }
        }

        /// <summary>
        /// Adds an uint16 reading method of a reader to respond to the event Read.
        /// The method changes the data endian on reading.
        /// The own method is sellected depending on data dimension.
        /// </summary>
        /// <param name="reader">The reader.</param>
        private void SetUint16ReaderInv(BL_TypeReader reader)
        {
            // Select a specific method according to attribute type and dimension.
            if (reader.AttrTypeDimension == 1)
            {
                ReadAttributes += new TypeReadHandler(reader.ReadUInt16scalInv);
            }
            else if (reader.AttrTypeDimension == 2)
            {
                ReadAttributes += new TypeReadHandler(reader.ReadUInt16vec2Inv);
            }
            else if (reader.AttrTypeDimension == 3)
            {
                ReadAttributes += new TypeReadHandler(reader.ReadUInt16vec3Inv);
            }
            else
            {
                ReadAttributes += new TypeReadHandler(reader.ReadUInt16vecNInv);
            }
        }



        /// <summary>
        /// Adds an int16 reading method of a reader to respond to the event Read.
        /// The own method is sellected depending on data dimension.
        /// </summary>
        /// <param name="reader">The reader.</param>
        private void SetInt16Reader(BL_TypeReader reader)
        {
            // Select a specific method according to attribute type and dimension.
            if (reader.AttrTypeDimension == 1)
            {
                ReadAttributes += new TypeReadHandler(reader.ReadInt16scal);
            }
            else if (reader.AttrTypeDimension == 2)
            {
                ReadAttributes += new TypeReadHandler(reader.ReadInt16vec2);
            }
            else if (reader.AttrTypeDimension == 3)
            {
                ReadAttributes += new TypeReadHandler(reader.ReadInt16vec3);
            }
            else
            {
                ReadAttributes += new TypeReadHandler(reader.ReadInt16vecN);
            }
        }

        /// <summary>
        /// Adds an int16 reading method of a reader to respond to the event Read.
        /// The method changes the data endian on reading.
        /// The own method is sellected depending on data dimension.
        /// </summary>
        /// <param name="reader">The reader.</param>
        private void SetInt16ReaderInv(BL_TypeReader reader)
        {
            // Select a specific method according to attribute type and dimension.
            if (reader.AttrTypeDimension == 1)
            {
                ReadAttributes += new TypeReadHandler(reader.ReadInt16scalInv);
            }
            else if (reader.AttrTypeDimension == 2)
            {
                ReadAttributes += new TypeReadHandler(reader.ReadInt16vec2Inv);
            }
            else if (reader.AttrTypeDimension == 3)
            {
                ReadAttributes += new TypeReadHandler(reader.ReadInt16vec3Inv);
            }
            else
            {
                ReadAttributes += new TypeReadHandler(reader.ReadInt16vecNInv);
            }
        }



        //////////////////////////
        // 4 byte types readers //
        //////////////////////////

        /// <summary>
        /// Adds an uint32 reading method of a reader to respond to the event Read.
        /// The own method is sellected depending on data dimension.
        /// </summary>
        /// <param name="reader">The reader.</param>
        private void SetUint32Reader(BL_TypeReader reader)
        {
            // Select a specific method according to attribute type and dimension.
            if (reader.AttrTypeDimension == 1)
            {
                ReadAttributes += new TypeReadHandler(reader.readUInt32scal);
            }
            else if (reader.AttrTypeDimension == 2)
            {
                ReadAttributes += new TypeReadHandler(reader.readUInt32vec2);
            }
            else if (reader.AttrTypeDimension == 3)
            {
                ReadAttributes += new TypeReadHandler(reader.readUInt32vec3);
            }
            else
            {
                ReadAttributes += new TypeReadHandler(reader.readUInt32vecN);
            }
        }

        /// <summary>
        /// Adds an uint32 reading method of a reader to respond to the event Read.
        /// The method changes the data endian on reading.
        /// The own method is sellected depending on data dimension.
        /// </summary>
        /// <param name="reader">The reader.</param>
        private void SetUint32ReaderInv(BL_TypeReader reader)
        {
            // Select a specific method according to attribute type and dimension.
            if (reader.AttrTypeDimension == 1)
            {
                ReadAttributes += new TypeReadHandler(reader.ReadUInt32scalInv);
            }
            else if (reader.AttrTypeDimension == 2)
            {
                ReadAttributes += new TypeReadHandler(reader.ReadUInt32vec2Inv);
            }
            else if (reader.AttrTypeDimension == 3)
            {
                ReadAttributes += new TypeReadHandler(reader.ReadUInt32vec3Inv);
            }
            else
            {
                ReadAttributes += new TypeReadHandler(reader.ReadUInt32vecNInv);
            }
        }



        /// <summary>
        /// Adds an int32 reading method of a reader to respond to the event Read.
        /// The own method is sellected depending on data dimension.
        /// </summary>
        /// <param name="reader">The reader.</param>
        private void SetInt32Reader(BL_TypeReader reader)
        {
            // Select a specific method according to attribute type and dimension.
            if (reader.AttrTypeDimension == 1)
            {
                ReadAttributes += new TypeReadHandler(reader.ReadInt32scal);
            }
            else if (reader.AttrTypeDimension == 2)
            {
                ReadAttributes += new TypeReadHandler(reader.ReadInt32vec2);
            }
            else if (reader.AttrTypeDimension == 3)
            {
                ReadAttributes += new TypeReadHandler(reader.ReadInt32vec3);
            }
            else
            {
                ReadAttributes += new TypeReadHandler(reader.ReadInt32vecN);
            }
        }

        /// <summary>
        /// Adds an int32 reading method of a reader to respond to the event Read.
        /// The method changes the data endian on reading.
        /// The own method is sellected depending on data dimension.
        /// </summary>
        /// <param name="reader">The reader.</param>
        private void SetInt32ReaderInv(BL_TypeReader reader)
        {
            // Select a specific method according to attribute type and dimension.
            if (reader.AttrTypeDimension == 1)
            {
                ReadAttributes += new TypeReadHandler(reader.ReadInt32scalInv);
            }
            else if (reader.AttrTypeDimension == 2)
            {
                ReadAttributes += new TypeReadHandler(reader.ReadInt32vec2Inv);
            }
            else if (reader.AttrTypeDimension == 3)
            {
                ReadAttributes += new TypeReadHandler(reader.ReadInt32vec3Inv);
            }
            else
            {
                ReadAttributes += new TypeReadHandler(reader.ReadInt32vecNInv);
            }
        }



        /// <summary>
        /// Adds an float32 reading method of a reader to respond to the event Read.
        /// The own method is sellected depending on data dimension.
        /// </summary>
        /// <param name="reader">The reader.</param>
        private void SetFloat32Reader(BL_TypeReader reader)
        {
            // Select a specific method according to attribute type and dimension.
            if (reader.AttrTypeDimension == 1)
            {
                ReadAttributes += new TypeReadHandler(reader.ReadFloat32scal);
            }
            else if (reader.AttrTypeDimension == 2)
            {
                ReadAttributes += new TypeReadHandler(reader.ReadFloat32vec2);
            }
            else if (reader.AttrTypeDimension == 3)
            {
                ReadAttributes += new TypeReadHandler(reader.ReadFloat32vec3);
            }
            else
            {
                ReadAttributes += new TypeReadHandler(reader.readFloat32vecN);
            }
        }

        /// <summary>
        /// Adds an float32 reading method of a reader to respond to the event Read.
        /// The method changes the data endian on reading.
        /// The own method is sellected depending on data dimension.
        /// </summary>
        /// <param name="reader">The reader.</param>
        private void SetFloat32ReaderInv(BL_TypeReader reader)
        {
            // Select a specific method according to attribute type and dimension.
            if (reader.AttrTypeDimension == 1)
            {
                ReadAttributes += new TypeReadHandler(reader.ReadFloat32scalInv);
            }
            else if (reader.AttrTypeDimension == 2)
            {
                ReadAttributes += new TypeReadHandler(reader.ReadFloat32vec2Inv);
            }
            else if (reader.AttrTypeDimension == 3)
            {
                ReadAttributes += new TypeReadHandler(reader.ReadFloat32vec3Inv);
            }
            else
            {
                ReadAttributes += new TypeReadHandler(reader.ReadFloat32vecNInv);
            }
        }




        //////////////////////////
        // 8 byte types readers //
        //////////////////////////

        /// <summary>
        /// Adds an uint64 reading method of a reader to respond to the event Read.
        /// The own method is sellected depending on data dimension.
        /// </summary>
        /// <param name="reader">The reader.</param>
        private void SetUint64Reader(BL_TypeReader reader)
        {
            // Select a specific method according to attribute type and dimension.
            if (reader.AttrTypeDimension == 1)
            {
                ReadAttributes += new TypeReadHandler(reader.readUInt64scal);
            }
            else if (reader.AttrTypeDimension == 2)
            {
                ReadAttributes += new TypeReadHandler(reader.readUInt64vec2);
            }
            else if (reader.AttrTypeDimension == 3)
            {
                ReadAttributes += new TypeReadHandler(reader.readUInt64vec3);
            }
            else
            {
                ReadAttributes += new TypeReadHandler(reader.readUInt64vecN);
            }
        }

        /// <summary>
        /// Adds an uint64 reading method of a reader to respond to the event Read.
        /// The method changes the data endian on reading.
        /// The own method is sellected depending on data dimension.
        /// </summary>
        /// <param name="reader">The reader.</param>
        private void SetUint64ReaderInv(BL_TypeReader reader)
        {
            // Select a specific method according to attribute type and dimension.
            if (reader.AttrTypeDimension == 1)
            {
                ReadAttributes += new TypeReadHandler(reader.ReadUInt64scalInv);
            }
            else if (reader.AttrTypeDimension == 2)
            {
                ReadAttributes += new TypeReadHandler(reader.ReadUInt64vec2Inv);
            }
            else if (reader.AttrTypeDimension == 3)
            {
                ReadAttributes += new TypeReadHandler(reader.ReadUInt64vec3Inv);
            }
            else
            {
                ReadAttributes += new TypeReadHandler(reader.ReadUInt64vecNInv);
            }
        }



        /// <summary>
        /// Adds an int64 reading method of a reader to respond to the event Read.
        /// The own method is sellected depending on data dimension.
        /// </summary>
        /// <param name="reader">The reader.</param>
        private void SetInt64Reader(BL_TypeReader reader)
        {
            // Select a specific method according to attribute type and dimension.
            if (reader.AttrTypeDimension == 1)
            {
                ReadAttributes += new TypeReadHandler(reader.readInt64scal);
            }
            else if (reader.AttrTypeDimension == 2)
            {
                ReadAttributes += new TypeReadHandler(reader.readInt64vec2);
            }
            else if (reader.AttrTypeDimension == 3)
            {
                ReadAttributes += new TypeReadHandler(reader.readInt64vec3);
            }
            else
            {
                ReadAttributes += new TypeReadHandler(reader.readInt64vecN);
            }
        }

        /// <summary>
        /// Adds an int64 reading method of a reader to respond to the event Read.
        /// The method changes the data endian on reading.
        /// The own method is sellected depending on data dimension.
        /// </summary>
        /// <param name="reader">The reader.</param>
        private void SetInt64ReaderInv(BL_TypeReader reader)
        {
            // Select a specific method according to attribute type and dimension.
            if (reader.AttrTypeDimension == 1)
            {
                ReadAttributes += new TypeReadHandler(reader.ReadInt64scalInv);
            }
            else if (reader.AttrTypeDimension == 2)
            {
                ReadAttributes += new TypeReadHandler(reader.ReadInt64vec2Inv);
            }
            else if (reader.AttrTypeDimension == 3)
            {
                ReadAttributes += new TypeReadHandler(reader.ReadInt64vec3Inv);
            }
            else
            {
                ReadAttributes += new TypeReadHandler(reader.ReadInt64vecNInv);
            }
        }



        /// <summary>
        /// Adds an float64 reading method of a reader to respond to the event Read.
        /// The own method is sellected depending on data dimension.
        /// </summary>
        /// <param name="reader">The reader.</param>
        private void SetFloat64Reader(BL_TypeReader reader)
        {
            // Select a specific method according to attribute type and dimension.
            if (reader.AttrTypeDimension == 1)
            {
                ReadAttributes += new TypeReadHandler(reader.ReadFloat64scal);
            }
            else if (reader.AttrTypeDimension == 2)
            {
                ReadAttributes += new TypeReadHandler(reader.ReadFloat64vec2);
            }
            else if (reader.AttrTypeDimension == 3)
            {
                ReadAttributes += new TypeReadHandler(reader.ReadFloat64vec3);
            }
            else
            {
                ReadAttributes += new TypeReadHandler(reader.ReadFloat64vecN);
            }
        }

        /// <summary>
        /// Adds an float64 reading method of a reader to respond to the event Read.
        /// The method changes the data endian on reading.
        /// The own method is sellected depending on data dimension.
        /// </summary>
        /// <param name="reader">The reader.</param>
        private void SetFloat64ReaderInv(BL_TypeReader reader)
        {
            // Select a specific method according to attribute type and dimension.
            if (reader.AttrTypeDimension == 1)
            {
                ReadAttributes += new TypeReadHandler(reader.ReadFloat64scalInv);
            }
            else if (reader.AttrTypeDimension == 2)
            {
                ReadAttributes += new TypeReadHandler(reader.ReadFloat64vec2Inv);
            }
            else if (reader.AttrTypeDimension == 3)
            {
                ReadAttributes += new TypeReadHandler(reader.ReadFloat64vec3Inv);
            }
            else
            {
                ReadAttributes += new TypeReadHandler(reader.ReadFloat64vecNInv);
            }
        }
    }
}
