using System;
using System.Collections.Generic;
using System.IO;
using System.Text;

namespace Zcu.Mve.Visualization.Loaders
{

    /// <summary>
    /// Contains all needed data attributes (minsS, maxsS, endian, etc.)
    /// </summary>
    public class BL_ModuleSettings
    {

        #region dataFilePath
        /// <summary>
        /// Path to the source file.
        /// </summary>
        private string dataFilePath;
        /// <summary>
        /// Path to the source file.
        /// </summary>
        public string DataFilePath
        {
            get { return dataFilePath; }
            set { dataFilePath = value; }
        }

        /// <summary>
        /// Path to the header file.
        /// </summary>
        private string headerFilePath;
        /// <summary>
        /// Path to the header file.
        /// </summary>
        public string HeaderFilePath
        {
            get { return headerFilePath; }
            set { headerFilePath = value; }
        }

        #endregion

        #region headerSize;
        /// <summary>
        /// Size in bytes if a header in the data file - number of bytes to skip.
        /// </summary>
        private int headerSize;
        /// <summary>
        /// Size in bytes if a header in the data file - number of bytes to skip.
        /// </summary>
        public int HeaderSize
        {
            get { return headerSize; }
            set { headerSize = value; }
        }
        #endregion

        #region dataDim
        /// <summary>
        /// Dimension of the data.
        /// </summary>
        private int dataDim;
        /// <summary>
        /// Dimension of the data.
        /// </summary>
        public int DataDim
        {
            get { return dataDim; }
            set { dataDim = value; }
        }
        #endregion

        #region attrType
        /// <summary>
        /// String describing the data attributes. See manul for more info. 
        /// </summary>
        private string dataDesc;
        /// <summary>
        /// String describing the data attributes. See manaul for more info. 
        /// </summary>
        public string DataDesc
        {
            get { return dataDesc; }
            set { dataDesc = value; }
        }
        #endregion

        #region invertEndian
        /// <summary>
        /// If the data bits shall be reversed after reading.
        /// </summary>
        private bool[] invertEndian;
        /// <summary>
        /// If the data bits shall be reversed after reading.
        /// </summary>
        public bool[] InvertEndian
        {
            get { return invertEndian; }
            set { invertEndian = value; }
        }
        #endregion

        #region minsS
        /// <summary>
        /// Minimum data space coordinates in each axis. The data samples spans from minsS to maxsS.
        /// </summary>
        private string mins;
        /// <summary>
        /// Minimum data space coordinates in each axis. The data samples spans from minsS to maxsS.
        /// </summary>
        public string Mins
        {
            get { return mins; }
            set { mins = value; }
        }
        #endregion

        #region maxsS
        /// <summary>
        /// Maximum data space coordinates in each axis. The data samples spans from minsS to maxsS.
        /// </summary>
        private string maxs;
        /// <summary>
        /// Maximum data space coordinates in each axis. The data samples spans from minsS to maxsS.
        /// </summary>
        public string Maxs
        {
            get { return maxs; }
            set { maxs = value; }
        }
        #endregion

        #region dataSizes
        /// <summary>
        /// Grid dataSizes in each axis. (nuber of data samples)
        /// </summary>
        private string sizes;
        /// <summary>
        /// Grid dataSizes in each axis. (nuber of data samples)
        /// </summary>
        public string Sizes
        {
            get { return sizes; }
            set { sizes = value; }
        }
        #endregion

        #region minIndeces
        /// <summary>
        /// Minimal sample indeces of the desired data subset.
        /// </summary>
        private string minIndeces;
        /// <summary>
        /// Minimal sample indeces of the desired data subset.
        /// </summary>
        public string MinIndeces
        {
            get { return minIndeces; }
            set { minIndeces = value; }
        }
        #endregion

        #region maxIndeces
        /// <summary>
        /// Maximal sample indeces of the desired data subset.
        /// </summary>
        private string maxIndeces;
        /// <summary>
        /// Maximal sample indeces of the desired data subset.
        /// </summary>
        public string MaxIndeces
        {
            get { return maxIndeces; }
            set { maxIndeces = value; }
        }
        #endregion

        #region points/cells

        private BinaryLoaderGlobalConstants.PointsOrCells pointOrCells;
        /// <summary>
        /// Specifies wether the properties shall be point or cell type.
        /// </summary>
        public BinaryLoaderGlobalConstants.PointsOrCells PointOrCells
        {
            get { return pointOrCells; }
            set { pointOrCells = value; }
        }


        #endregion

        #region fileTimes
        private DateTime dataFileCreationTime;
        /// <summary>
        /// The time when the specified file was created.
        /// </summary>
        public DateTime DataFileCreationTime
        {
            get { return dataFileCreationTime; }
            set { dataFileCreationTime = value; }
        }

        private DateTime dataFileLastWriteTime;
        /// <summary>
        /// The time the specified file was last written to.
        /// </summary>
        public DateTime DataFileLastWriteTime
        {
            get { return dataFileLastWriteTime; }
            set { dataFileLastWriteTime = value; }
        }


        private DateTime headerFileCreationTime;
        /// <summary>
        /// The time when the specified file was created.
        /// </summary>
        public DateTime HeaderFileCreationTime
        {
            get { return headerFileCreationTime; }
            set { headerFileCreationTime = value; }
        }

        private DateTime headerFileLastWriteTime;
        /// <summary>
        /// The time the specified file was last written to.
        /// </summary>
        public DateTime HeaderFileLastWriteTime
        {
            get { return headerFileLastWriteTime; }
            set { headerFileLastWriteTime = value; }
        }
        #endregion

        private static int i;

        /// <summary>
        /// Compares two module settnigs. Both settings have to be valid.
        /// </summary>
        /// <param name="first">first settings</param>
        /// <param name="second">second settings</param>
        /// <returns>TRUE if both settings refer to the same data, otherwise FALSE</returns>
        public static bool AreSettingsSame(BL_ModuleSettings first, BL_ModuleSettings second)
        {
            try
            {
                // Check header sizes.
                if (first.headerSize != second.headerSize)
                {
                    return false;
                }

                // Check the filepaths.
                if (first.dataFilePath.CompareTo(second.dataFilePath) != 0)
                {
                    return false;
                }

                // Check the headerFilePaths.
                if ((first.headerFilePath == null && second.headerFilePath == null) == false)
                {
                    if (first.headerFilePath.CompareTo(second.headerFilePath) != 0)
                    {
                        return false;
                    }
                }

                // Check the data file creation times and last write times.
                if (File.Exists(first.dataFilePath) == true && File.Exists(second.dataFilePath) == true)
                {
                    if (first.dataFileCreationTime.CompareTo(second.dataFileCreationTime) != 0)
                    {
                        return false;
                    }

                    if (first.dataFileLastWriteTime.CompareTo(second.dataFileLastWriteTime) != 0)
                    {
                        return false;
                    }
                }

                // Check the header file creation times and last write times.
                if (first.headerFilePath != null && second.headerFilePath != null)
                {
                    if (File.Exists(first.headerFilePath) == true && File.Exists(second.headerFilePath) == true)
                    {
                        if (first.headerFileCreationTime.CompareTo(second.headerFileCreationTime) != 0)
                        {
                            return false;
                        }

                        if (first.headerFileLastWriteTime.CompareTo(second.headerFileLastWriteTime) != 0)
                        {
                            return false;
                        }
                    }
                }


                // Check the data dimension.
                if (first.dataDim != second.dataDim)
                {
                    return false;
                }

                // Check the data describing strings.
                if (first.dataDesc.CompareTo(second.dataDesc) != 0)
                {
                    return false;
                }

                // Check the endian settings lenghts.
                if (first.invertEndian.Length != second.invertEndian.Length)
                {
                    return false;
                }

                // Check the endian settings.
                for (i = 0; i < first.invertEndian.Length; i++)
                {
                    if (first.invertEndian[i] != second.invertEndian[i])
                    {
                        return false;
                    }
                }

                // Check the mins.
                if (first.mins.CompareTo(second.mins) != 0)
                {
                    return false;
                }

                // Check the maxes.
                if (first.maxs.CompareTo(second.maxs) != 0)
                {
                    return false;
                }

                // Check the sizes.
                if (first.sizes.CompareTo(second.sizes) != 0)
                {
                    return false;
                }

                // Check the attribute types.
                if (first.pointOrCells != second.pointOrCells)
                {
                    return false;
                }

                // Check if there is any reference to min indeces and
                // if in both settings the data subset settings are same.
                if ((first.minIndeces != null && first.maxIndeces != null &&
                    second.minIndeces != null && second.maxIndeces != null))
                {
                    if (first.minIndeces.CompareTo(second.minIndeces) != 0)
                    {
                        return false;
                    }

                    if (first.maxIndeces.CompareTo(second.maxIndeces) != 0)
                    {
                        return false;
                    }
                }
                else
                {
                    if (first.minIndeces == null && second.minIndeces == null &&
                        first.maxIndeces == null && second.maxIndeces == null)
                    {
                        return true;
                    }
                    else
                    {
                        return false;
                    }
                }
            }
            catch (NullReferenceException)
            {
                return false;
            }

            return true;
        }
    }
}
