using System;
using Zcu.Mve.Visualization;
using Zcu.Mve.Core;

namespace Zcu.Mve.Visualization
{
	/// <summary>
	/// Represent min-max box.
	/// It is automatically maintained by given point array.
	/// </summary>
	public sealed class BoundingBox
	{
		private bool isMinMaxVaild;

		/// <summary>
		/// Number of points with maximum dimension
		/// </summary>
		private int maxDimCnt;

		/// <summary>
		/// Current maximum dimension
		/// </summary>
		private int maxDim;

		private IPoint maxPoint;
		private IPoint minPoint;
		private IPoint midPoint;
		private double maxDiag;

		private UniformDataArray points;

		/// <summary>
		/// Return point with maximal values.
		/// Do not change its values!
		/// </summary>
		public IPoint MaxPoint 
		{ 
			get 
			{ 
				if (this.maxPoint == null)
					return null;

				if (!isMinMaxVaild) 
					this.ValidateMinMax(); 

				return (IPoint) maxPoint.DeepCopy(); 
			} 
		}

		/// <summary>
		/// Return point with minimal values.
		/// Do not change its values!
		/// </summary>
		public IPoint MinPoint 
		{ 
			get 
			{ 
				if (this.minPoint == null)
					return null;

				if (!isMinMaxVaild) 
					this.ValidateMinMax(); 

				return (IPoint) minPoint.DeepCopy(); 
			} 
		}

		/// <summary>
		/// Return center of min-max box.
		/// </summary>
		public IPoint MidPoint 
		{ 
			get 
			{ 
				if (this.midPoint == null)
					return null;

				if (!isMinMaxVaild) 
					this.ValidateMinMax(); 

				return (IPoint) midPoint.DeepCopy(); 
			} 
		}

		/// <summary>
		/// Creates bounding box that is automatically maintained with respect to given array.
		/// </summary>
		/// <param name="points"></param>
		public BoundingBox(UniformDataArray points)
		{
			this.points = points;

			isMinMaxVaild = false;

			if (typeof(IPoint).IsAssignableFrom(points.ElemType))
			{
				if (points.Count != 0)
					ValidateDimension();
				else
					System.Diagnostics.Debug.WriteLine("BoundingBox warning: there are no points in the DataSet, which the BoundingBox is created for.");
			}
			else 
			{
				throw new MveException(this, "BoundingBox: Unknown point type " + points.ElemType.GetType().FullName);
			}
			/*
						minPoint = new PointND(maxDim);
						maxPoint = new PointND(maxDim);
						midPoint = new PointND(maxDim);
			*/
			this.points.OnElementSetEvent += new MveArray.ElementSetCallBack(PoinSetEventHandler);
		}        

        /// <summary>
        /// Returns point array that is managed by curent bounding box.
        /// </summary>
        public UniformDataArray Points
		{
			get
			{
				return this.points;
			}
		}

		/// <summary>
		/// Convert Point2D, Point3D, or PointND to PointND.
		/// </summary>
		/// <param name="point"></param>
		/// <returns></returns>
		private static PointND ConvertFrom(IPoint point)
		{
			if (point.GetType() == typeof(Point2D))
			{
				return (PointND)((Point2D)point);
			}

			if (point.GetType() == typeof(Point3D))
			{
				return (PointND)((Point3D)point);        
			}

			if (point.GetType() == typeof(PointND))
			{
				return (PointND)point;
			}

			throw new MveException(typeof(BoundingBox), point.GetType().FullName + " is unsupported point type.");
		}

		/// <summary>
		/// Event checks whether the new inserted point influences 
		/// the bounding box. If the new point is outside the current
		/// bounding box, the status of boundingbox is not changed and
		/// the box is changed. If the old point lies on the boundingbox
		/// the status and the new point lies inside the boudingbox, the 
		/// status is changed to false.
		/// </summary>
		/// <param name="newPoint">new point</param>
		/// <param name="index">index of changed point</param>
		/// <returns>true</returns>
		private bool PoinSetEventHandler(object newPoint, params int [] index)
		{
			PointND point;

			// IPoint is removed
			if (newPoint == null)
			{
				point = ConvertFrom((IPoint) points[index]);
				for (int i = 0; i < point.Length; i++)
				{
					if (point[i] >= maxPoint[i] || point[i] <= minPoint[i])
					{
						isMinMaxVaild = false;
					}
					if (points.ElemType == typeof(PointND) && point.Length == maxDim)
					{
						maxDimCnt--;
						if (maxDimCnt == 0)
						{
							ValidateDimension();
						}
					}
				}

				// TODO: Bacha na odebrani posledniho bodu s nejvetsi dimenzi ... musime pak snizit maxDim ...
				return true;
			}

			// IPoint is added
			point = ConvertFrom((IPoint) newPoint);

			if (point.Length > this.maxDim)
			{
				this.ResetDimension(point.Length);
			}

			ValideOnePointAdd(point);

			return true;
		}

		/// <summary>
		/// Recomputes the BoundingBox from all points
		/// </summary>
		private void ValidateMinMax()
		{
			for (int i = 0; i < maxDim; i++)
			{
				minPoint[i] = Double.MaxValue;
				maxPoint[i] = Double.MinValue;
			}

			for (int i = 0; i < points.Count; i++) 
			{
				PointND point = ConvertFrom((IPoint)points[i]);

				for(int dim = 0; dim < point.Length; dim++) 
				{
					minPoint[dim] = Math.Min(minPoint[dim], point[dim]);
					maxPoint[dim] = Math.Max(maxPoint[dim], point[dim]);
				}
			}
			isMinMaxVaild = true;
			this.ValidateMidDiag();
		}


		/// <summary>
		/// Find maxDim and sim maxDimCnt
		/// </summary>
		private void ValidateDimension()
		{
			maxDim = 0;
			maxDimCnt = 0;

			for (int i = 0; i < points.Count; i++)
			{
				if (points[i] != null) 
				{
					if ((maxDim < ((IPoint)points[i]).Length))
					{
						maxDim = ((IPoint)points[i]).Length;
						maxDimCnt = 0;
					}
						// Calculate number of points with maximal dimension.
					else if ((maxDim == ((IPoint)points[i]).Length))
					{ 
						maxDimCnt++;
					}
				}
			}

			ResetDimension(maxDim);
		}

		/// <summary>
		/// Reset dimension of min-mid-max points
		/// </summary>
		/// <param name="newDimension"></param>
		private void ResetDimension(int newDimension)
		{
			this.minPoint = PointND.ChangeDimension(this.minPoint, newDimension);
			this.maxPoint = PointND.ChangeDimension(this.maxPoint, newDimension);
			this.midPoint = PointND.ChangeDimension(this.midPoint, newDimension);
			this.maxDim = newDimension;
		}

		/// <summary>
		/// Check whether addaed point change bounding box and recalculate it.
		/// </summary>
		/// <param name="newPoint"></param>
		private void ValideOnePointAdd(PointND newPoint)
		{
			bool bbChanged = false;

			for (int i = 0; i < newPoint.Length; i++)
			{
				if (newPoint[i] > maxPoint[i])
				{
					maxPoint[i] = newPoint[i];    
					bbChanged = true;
				}
				if (newPoint[i] < minPoint[i])
				{
					minPoint[i] = newPoint[i];
					bbChanged = true;
				}
			}

			if (bbChanged)
			{
				ValidateMidDiag();
			}
		}

		/// <summary>
		/// Recompute mid-point and max diagonal length.
		/// </summary>
		private void ValidateMidDiag()
		{
			double sum = 0;
			double dif;

			for (int i = 0; i < maxDim; i++)
			{
				midPoint[i] = (minPoint[i] + maxPoint[i]) / 2.0;
				dif = maxPoint[i] - minPoint[i];
				sum += dif * dif;
			}

			this.maxDiag = Math.Sqrt(sum);
		}

		/// <summary> 
		/// Return volume diagonal of the bounding box.
		/// </summary> 
		public double MaxDiag  
		{ 
			get 
			{ 
				if(!isMinMaxVaild)  
					this.ValidateMinMax();  

				return maxDiag; 
			} 
		}

		/// <summary>
		/// Returns description of BoundignBox.
		/// </summary>
		public override string ToString()
		{
			string str = "BoundingBox: ";

			if (this.MinPoint == null || this.MaxPoint == null || this.MidPoint == null)
				return str += "undefined (there are no points)";

			str += "\n";
			str += "\t" + "- min point : " + this.MinPoint + "\n";
			str += "\t" + "- max point : " + this.MaxPoint + "\n";
			str += "\t" + "- mid point : " + this.MidPoint + "\n";
			str += "\t" + "- max diag  : " + this.MaxDiag + "\n";
			return str;
		}

	} // class BoundingBox
} // namespace
