using System;
using Zcu.Mve.Core;
using Zcu.Mve.Numerics;
using System.Drawing;
using System.Xml;

namespace Zcu.Mve.Visualization.Processing
{
    /// <summary>
    /// A module that transforms a Scalar value into a ColorRGBA. Values lower or equal to zero will be computed as equal to zero, values greater or equal to one will be computed as equal to one , values in between will become linear interpolations among control points.*
    /// </summary>
    [ModuleInfo("Jan tefny", "A module that transforms a Scalar value into a ColorRGBA. Values lower or equal to zero will be computed as equal to zero, values greater or equal to one will be computed as equal to one , values in between will become linear interpolation or cosine interpolation among control points.", CategoryType.Filter, CategoryData.Number, IconName = "PaletteLinear.ico")]
    [PortInfo("Value", "Scalar value to be transformed into ColorRGBA")]
    [PortInfo("Color", "Resulting ColorRGBA value")]   
    public class EditablePalette:Module
    {
        private PaletteGUI gui;//take palette and allows usage of its variables
        private ColorPointList list;
        bool stop = false, actual = false;

        /// <summary>
        /// Default constructor, adds input and output ports.
        /// </summary> 
        public EditablePalette()
        {
            AddInPort("Value", typeof(Scalar)); //input port
            AddOutPort("Color", typeof(ColorRGBA));//output port
            gui = new PaletteGUI();
        }

        /// <summary>
        /// WriteConfig method for saving of positions and values of CP on palette.
        /// </summary>
        public override void WriteConfig(System.Xml.XmlElement config)
        {
            ColorPoint pom = gui.colorPointList.First;

            while (pom != null)
            {
                XmlElement ec = config.OwnerDocument.CreateElement("Point");
                ec.InnerText = pom.x + ", " + pom.y + ", " + pom.color.R + ", " + pom.color.G + ", " + pom.color.B;
                config.AppendChild(ec);
                pom = pom.right;
            }
        }

        /// <summary>
        /// ReadConfig method for loading of positions and values of CP which were saved by the user last time.
        /// </summary>
        public override void ReadConfig(System.Xml.XmlElement config)
        {
            XmlNodeList c = config.GetElementsByTagName("Point");

            if (c.Count >= 2)
            {
                ColorPoint b = null, first = null;

                for (int i = 0; i < c.Count; i++)
                {
                    XmlNode cn = c[i];

                    string[] line = cn.InnerText.Split(',');

                    ColorPoint n = new ColorPoint(int.Parse(line[0]), int.Parse(line[1]), Color.FromArgb(int.Parse(line[2]), int.Parse(line[3]), int.Parse(line[4])), null, null);

                    if (i != 0)
                    {
                        n.left = b;
                        b.right = n;
                    }
                    else
                    {
                        first = n;
                    }

                    b = n;
                }

                gui.colorPointList = new ColorPointList(first, b);
            }
        }

        /// <summary>
        /// Execution of modul. It reloads palette if palette isn't actual and calls method which transfers scalar to color.
        /// </summary>
        public override void Execute()
        {
                if (!actual)//testing if color palette is actual, if not, palette is reloaded
                    GetPalet();

                ExecuteScalarToColor();//calls method which transfers scalar to color
            
        }

        /// <summary>
        /// Method, which gets control points from EditablePalette.
        /// </summary>
        private void GetPalet()
        {
            list = gui.colorPointList;
            actual = true;

        }

        /// <summary>
        /// Method, which computes ColorRGBA value from input Scalar value.
        /// </summary>
        private void ExecuteScalarToColor()
        {
            Scalar input = (Scalar)GetInput("Value");//gets scalar input
            float v = (float)input.Value;
            ColorPoint pom;
            pom = list.First;
            int x;

            if (v < 0)//adjusting of forbidden values
                v = 0.0f;
            if (v > 1)
                v = 1.0f;

            x = (int)(v * (gui.width - 1));//conversion of input value to palette value

            if (gui.function == 0)//linear transition
            {
                while (!stop)
                {

                    if (x <= pom.right.x)//if value is between two command points
                    {
                        float pomocna = 1.0f / (pom.right.x - pom.x);//help value for calculation of RGB 
                        float pomocnay = -1.0f * ((float)(pom.right.y - pom.y) / (float)(pom.right.x - pom.x));//help value for calculation of Alpha value
                        int alpha;
                        alpha = (int)((float)(255.0f - ((float)pom.y / 59.0f * 255.0f)) + ((float)(x - pom.x) * pomocnay));
                        ColorRGBA output;//output
                        output = Color.FromArgb(alpha, (int)(pomocna * (pom.right.x - x) * pom.color.R + pomocna * (x - pom.x) * pom.right.color.R), (int)(pomocna * (pom.right.x - x) * pom.color.G + pomocna * (x - pom.x) * pom.right.color.G), (int)(pomocna * (pom.right.x - x) * pom.color.B + pomocna * (x - pom.x) * pom.right.color.B));//calculation of colorRGBA value for output
                        SetOutput("Color", output);//setting output
                        stop = true;
                    }
                    else pom = pom.right;//if value isn't between control points get to new control point

                }
                stop = false;
            }
            else if (gui.function == 1)//cosine transition
            {
                while (!stop)
                {
                    if (x <= pom.right.x)//if value is between two command points
                    {
                        double w1, w2, vzd;
                        Color c;
                        vzd = pom.right.x - pom.x;
                        float pomocnay = -1.0f * ((float)(pom.right.y - pom.y) / (float)(pom.right.x - pom.x));//help value for calculation of Alpha value

                        int alpha;
                        alpha = (int)((float)(255.0f - ((float)pom.y / 59.0f * 255.0f)) + ((float)(x - pom.x) * pomocnay));

                        w1 = 0.5 + 0.5 * Math.Cos(Math.PI * (x - pom.x) / vzd);
                        w2 = 1 - w1;

                        c = Color.FromArgb(alpha,(int)(pom.color.R * w1 + pom.right.color.R * w2), (int)(pom.color.G * w1 + pom.right.color.G * w2), (int)(pom.color.B * w1 + pom.right.color.B * w2));

                        ColorRGBA output;
                        output = c;
                        SetOutput("Color", output);//setting output
                        stop = true;
                    }
                    else pom = pom.right;//if value isn't between control points get to new control point

                }
                stop = false;
            }
            

        }

        /// <summary>
        /// Method, which invokes setup
        /// </summary>
        public override ModuleSetup InvokeSetup() //method for invoking setup
        {
            actual = false;
            return gui;
        }
    }    
}
