﻿using System;
using System.Collections.Generic;
using System.Text;
using Zcu.Mve.Core;

namespace Zcu.Mve.Visualization.Processing
{
    /// <summary>
    /// Module constructs a dual graph for a triangular mesh. The module uses the first attribute of the Triangle type. The vertices of the output are located at centres of original triangles, neighbouring triangles are connected by edges.
    /// </summary>
    [ModuleInfo("Libor Váša", "Module constructs a dual graph for a triangular mesh. The module uses the first attribute of the Triangle type. The vertices of the output are located at centres of original triangles, neighbouring triangles are connected by edges.", CategoryType.Filter, CategoryData.Mesh)]
    [PortInfo("InMesh", "Input mesh, should contain at least one attribute of the Triangle type. Vertices should be of the Point3D type.")]
    [PortInfo("DualGraph", "Output mesh, one vertex for each original triangle, neighbouring triangles connected by an edge.")]
    class MeshDualGraph:Module
    {
        public MeshDualGraph()
        {
            AddInPort("InMesh", typeof(UnstrGrid));
            AddOutPort("DualGraph", typeof(UnstrGrid));
        }

        UnstrGrid result = null;

        public override void Execute()
        {
            if (IsAllDataSame() && (result != null))
            {
                SetOutput("DualGraph", result, true);
            }
            else
            {
                UnstrGrid input = (UnstrGrid)GetInput("InMesh");
                result = new UnstrGrid();
                UniformDataArray triangles = (UniformDataArray)input.GetCells(typeof(Triangle))[0];
                UniformDataArray newPoints = new UniformDataArray(typeof(Point3D), triangles.Count);
                Dictionary<Line, int> edges = new Dictionary<Line, int>();
                List<Line> newEdges = new List<Line>();
                int index = 0;
                foreach (Triangle t in triangles)
                {
                    Point3D p = ((Point3D)input.Points[t.V1] + (Point3D)input.Points[t.V2] + (Point3D)input.Points[t.V3]) * (1.0 / 3);
                    newPoints[index] = p;
                    Line l = new Line(t.V1, t.V2, false);
                    if (edges.ContainsKey(l))
                    {
                        newEdges.Add(new Line(index, edges[l]));
                    }
                    else
                    {
                        edges.Add(l, index);
                    }

                    l = new Line(t.V1, t.V3, false);
                    if (edges.ContainsKey(l))
                    {
                        newEdges.Add(new Line(index, edges[l]));
                    }
                    else
                    {
                        edges.Add(l, index);
                    }

                    l = new Line(t.V3, t.V2, false);
                    if (edges.ContainsKey(l))
                    {
                        newEdges.Add(new Line(index, edges[l]));
                    }
                    else
                    {
                        edges.Add(l, index);
                    }
                    if (index % 1000 == 0)
                        ProgressInfo((double)index / triangles.Count);
                    index++;                    
                }
                result.Points = newPoints;
                result.AddCells("Edges", UniformDataArray.FromList(newEdges));
                SetOutput("DualGraph", result);
            }
        }
    }
}
