using System;
using System.IO;

namespace Zcu.Mve.Visualization.Savers
{
	/// <summary>
	/// Class for saving text STL file format.
	/// </summary>
	public class SaverStl
	{
		private int numberOfCells = 0;
		private int numberOfVertices  = 0;

		private Zcu.Mve.Visualization.UniformDataArray verticesArray = null;

		private Zcu.Mve.Visualization.UniformDataArray[] cellsArrays;

		private Zcu.Mve.Visualization.UniformDataArray cellNormalsArray = null;

		/// <summary>
		/// Saves mesh data to output file.
		/// </summary>
		/// <param name="mesh">Input mesh.</param>
		/// <param name="nameOfFile">Name of output file.</param>
		public void SaveFile(Zcu.Mve.Visualization.Mesh mesh, string nameOfFile)
		{
			this.verticesArray = mesh.Points;              // Gets vertices array.
			this.numberOfVertices = mesh.Points.Count;     // Gets number of vertices.

			this.cellsArrays = mesh.GetCells(typeof(Zcu.Mve.Visualization.Triangle));
			                                               // Gets cells arrays.
			this.numberOfCells = this.SumNumberOfCells();
			                                               // Gets number of cells.

			if ((this.numberOfCells < 1) || (this.numberOfVertices < 3))
			{
                throw new Zcu.Mve.Core.MveException(typeof(MeshSaver), "No triangles or not enough vertices to store.");
			}

			this.ReadCellsNormals(mesh);

			try
			{
				this.WriteDataToFile(nameOfFile);
			}
			catch (System.Exception)
			{
                throw new Zcu.Mve.Core.MveException(typeof(MeshSaver), "Error writing output file.");
			}
		}

		/// <summary>
		/// Counts number of cells.
		/// </summary>
		/// <returns>Number of triangles.</returns>
		private int SumNumberOfCells()
		{
			int sum = 0;

			for (int index = 0; index < this.cellsArrays.Length; index++)
			{
				sum += this.cellsArrays[index].Count;
			}
			return sum;
		}

		/// <summary>
		/// Reads cell normals from mesh and store it to temporal array.
		/// </summary>
		/// <param name="mesh">Input mesh.</param>
		private void ReadCellsNormals(Zcu.Mve.Visualization.Mesh mesh)
		{
			string[] cellsIds;
			string[] attributesIds;
			int      indexOfNormal = 0;

			cellsIds = new string[mesh.GetCellIds().Count];
			mesh.GetCellIds().CopyTo(cellsIds, 0);

			this.cellNormalsArray = new Zcu.Mve.Visualization.UniformDataArray(typeof(Zcu.Mve.Visualization.Normal3D), this.numberOfCells);

			for (int cellIndex = 0; cellIndex < cellsIds.Length; cellIndex++)
			{
				attributesIds = new string[mesh.GetCellAttrsIds(cellsIds[cellIndex]).Count];
				mesh.GetCellAttrsIds(cellsIds[cellIndex]).CopyTo(attributesIds, 0);

				for (int attIndex = 0; attIndex < attributesIds.Length; attIndex++)
				{
					if (mesh.GetCellAttrs(cellsIds[cellIndex], attributesIds[attIndex]).ElemType == typeof(Zcu.Mve.Visualization.Normal3D))
					{
						for (int index = 0; index < mesh.GetCellAttrs(cellsIds[cellIndex], attributesIds[attIndex]).Count; index++)
						{
							this.cellNormalsArray[indexOfNormal] = mesh.GetCellAttrs(cellsIds[cellIndex], attributesIds[attIndex])[index];
							indexOfNormal++;
						}
					}
				}
			}

			if (indexOfNormal > this.numberOfCells)
			{
                throw new Zcu.Mve.Core.MveException(typeof(MeshSaver), "Number of triangle normals and number of triangles is not same.");
			}

			if (indexOfNormal == 0)
			{
                throw new Zcu.Mve.Core.MveException(typeof(MeshSaver), "Stl format needs triangle normals.");
			}
		}

		/// <summary>
		/// Writes complete data to output file.
		/// </summary>
		/// <param name="nameOfFile">Name of output file.</param>
		private void WriteDataToFile(string nameOfFile)
		{
			Zcu.Mve.Visualization.Point3D vertex;
			Zcu.Mve.Visualization.Triangle triangle;
			Zcu.Mve.Visualization.Normal3D normal3d;
			int normalIndex = 0;
			
			StreamWriter streamWriter = new StreamWriter(nameOfFile);

			System.Globalization.NumberFormatInfo nfi = new System.Globalization.NumberFormatInfo();

			nfi.CurrencyDecimalSeparator = ".";

			streamWriter.WriteLine("solid generated by MVE2 module MeshSaver");

			for (int arrayIndex = 0; arrayIndex < this.cellsArrays.Length; arrayIndex++)
			{
				for (int cellIndex = 0; cellIndex < this.cellsArrays[arrayIndex].Count; cellIndex++)
				{
					triangle = (Zcu.Mve.Visualization.Triangle)this.cellsArrays[arrayIndex][cellIndex];	
					normal3d = (Zcu.Mve.Visualization.Normal3D)this.cellNormalsArray[normalIndex];
					normalIndex++;

					streamWriter.WriteLine("  facet normal {0} {1} {2}", normal3d.Nx.ToString("0.######0.0", nfi), normal3d.Ny.ToString("0.######0.0", nfi), normal3d.Nz.ToString("0.######0.0", nfi));

					streamWriter.WriteLine("    outer loop");

					vertex = (Zcu.Mve.Visualization.Point3D)this.verticesArray[triangle.V1];
					streamWriter.WriteLine("      vertex {0} {1} {2}", vertex.X.ToString("0.######0.0", nfi), vertex.Y.ToString("0.######0.0", nfi), vertex.Z.ToString("0.######0.0", nfi));

					vertex = (Zcu.Mve.Visualization.Point3D)this.verticesArray[triangle.V2];
					streamWriter.WriteLine("      vertex {0} {1} {2}", vertex.X.ToString("0.######0.0", nfi), vertex.Y.ToString("0.######0.0", nfi), vertex.Z.ToString("0.######0.0", nfi));
					
					vertex = (Zcu.Mve.Visualization.Point3D)this.verticesArray[triangle.V3];
					streamWriter.WriteLine("      vertex {0} {1} {2}", vertex.X.ToString("0.######0.0", nfi), vertex.Y.ToString("0.######0.0", nfi), vertex.Z.ToString("0.######0.0", nfi));

					streamWriter.WriteLine("    endloop");

					streamWriter.WriteLine("  endfacet");
				}
			}
			
			streamWriter.WriteLine("endsolid generated by MVE2 module MeshSaver");
			streamWriter.Close();
		}
	}
}
