using System;
using System.ComponentModel;
using System.Collections.Generic;
using Zcu.Mve.Core;
using Zcu.Mve.Visualization;


namespace Zcu.Mve.Visualization.Processing
{
    /// <summary>
    /// Summary description for NormalComputer.
    /// </summary>  
    [ModuleInfo("Libor Va", "Module that computes vertex normals for a TriangleMesh. Normals are placed in a new point attribute of name specified in the AttrName property.", CategoryType.Filter, CategoryData.Mesh, IconName = "NormalComputer.ico")]
    [PortInfo("TriangleMesh", "Input triangle mesh to be processed")]
    [PortInfo("Computed TriangleMesh", "Shallow copy of the input mesh with added attribute that represents vertex normals.")]
    public class NormalComputer : Module
    {
        private TriangleMesh output = null;

        private string attrName = "Normal3D";

        private bool invertNormals = false;

        private bool overwrite = true;


        /// <summary>
        /// Gets or sets the attribute name that will contain the normals. Cannot be set to an empty string.
        /// </summary>
        [Description("Gets or sets the attribute name that will contain the normals. Cannot be set to an empty string.")]
        public string AttrName
        {
            get { return (attrName); }
            set
            {
                if (value != "") attrName = value;
                else throw new MveException(this, "Attribute name for normals cannot be an empty strng.");
            }
        }

        /// <summary>
        /// Determines whether or not normals should be inverted. Important for specular reflection computation.
        /// </summary>
        [Description("Use this option if you are importing files originally made for right-handed coordinate systems.")]
        public bool InvertNormals
        {
            get { return invertNormals; }
            set { invertNormals = value; }
        }

        /// <summary>
        /// Specifies whether or not normals should be overwritten if they are already present in the mesh.
        /// </summary>
        [Description("Specifies whether or not normals should be overwritten if they are already present in the mesh.")]
        public bool OverwriteNormals
        {
            get { return overwrite; }
            set { overwrite = value; }
        }


        /// <summary>
        /// Default constructor.
        /// </summary>
        public NormalComputer()
        {
            AddInPort("TriangleMesh", typeof(TriangleMesh));
            AddOutPort("Computed TriangleMesh", typeof(TriangleMesh));
        }


        /// <summary>
        /// Computes the normals of the mesh. If the mesh is unchanged, and no properties were changed since the last run, then the result of the last run is returned with dataSame set to true.
        /// </summary>
        public override void Execute()
        {
            if ((!IsAllUserSettingsSame()) || (!IsAllDataSame()) || (output == null))
            {
                TriangleMesh input = (TriangleMesh)GetInput("TriangleMesh");
                output = (TriangleMesh)input.ShallowCopy();

                ICollection<string> pointAttrs = output.GetPointAttrsIds();
                List<string> normalAttrs = new List<string>();
                foreach (string attr in pointAttrs)
                {
                    if (output.GetPointAttrs(attr).ElemType == typeof(Normal3D)) normalAttrs.Add(attr);
                }
                if (normalAttrs.Count > 0)
                {
                    if (overwrite)
                    {
                        Console.WriteLine("{0}: Normals already present, replacing...", Name);
                        foreach (string attr in normalAttrs) output.RemovePointAttrs(attr);
                    }
                    else
                    {
                        Console.WriteLine("{0}: Normals already present, ending execution...", Name);
                        SetOutput("Computed TriangleMesh", output, false);
                        ProgressInfo(1.0);
                        return;
                    }
                }



                // LVA 27.11.2008 - I don't understand this thing, however it screws up my stuff, so I remove it
                /*if ((output != null) && (!paramsChanged) && (output.Points.ModificationTime == input.Points.ModificationTime))
                {
                  if (output.GetCellIds().Count == input.GetCellIds().Count)
                  {
                    foreach (String s in output.GetCellIds())
                    {
                      if (input.GetCellIds().Contains(s))
                      {
                        if (((UniformDataArray) output.GetCells(s)).ModificationTime
                           ==
                           ((UniformDataArray) input.GetCells(s)).ModificationTime)
                        {
                          SetOutput("Computed TriangleMesh", output, false);
                          return;
                        }
                      }
                    }
                  }
                }*/

                UniformDataArray triangles = (input.GetCells(typeof(Triangle)))[0];

                UniformDataArray normals = new UniformDataArray(typeof(Normal3D), input.Points.Count);

                Point3D[] pArray = (Point3D[])input.Points.Array;
                Normal3D[] nArray = (Normal3D[])normals.Array;

                Point3D v1, v2;
                Normal3D normal;

                for (int i = 0; i < nArray.Length; i++) nArray[i] = new Normal3D(0, 0, 0);

                int pos = 1;
                foreach (Triangle t in triangles)
                {
                    v1 = pArray[t.V2] - pArray[t.V1];
                    v2 = pArray[t.V3] - pArray[t.V1];
                    normal = (Normal3D)(v1.CrossProduct(v2));
                    normal.Normalize();

                    if (!(double.IsNaN(normal[0]) || double.IsNaN(normal[1]) || double.IsNaN(normal[2])))
                    {
                        nArray[t.V1] += normal;
                        nArray[t.V2] += normal;
                        nArray[t.V3] += normal;
                    }

                    pos++;
                    if ((pos % 1000) == 0) ProgressInfo((double)pos / triangles.Count);
                }

                for (int i = 0; i < nArray.Length; i++)
                {
                    normal = nArray[i];

                    if (double.IsNaN(normal[0]) || double.IsNaN(normal[1]) || double.IsNaN(normal[2]))
                    {
                        normal = new Normal3D(1, 0, 0);
                    }

                    normal.Normalize();

                    if (invertNormals) nArray[i] = -normal;
                    else nArray[i] = normal;
                }

                output.AddPointAttrs(attrName, normals);
                SetOutput("Computed TriangleMesh", output, false);
                ProgressInfo(1.0);
            }
            else
            {
                SetOutput("Computed TriangleMesh", output, true);
            }
        }
    }
}
