using System;
using Zcu.Mve.Core;
using Zcu.Mve.Numerics;
using Zcu.Mve.Visualization;
using Zcu.Mve.Strings;
using System.IO;
using System.Globalization;
using System.ComponentModel;
using System.Windows.Forms.Design;

namespace Zcu.Mve.Visualization.Loaders
{
    /// <summary>
    /// Obj triangle mesh file loader module.
    /// </summary>
    [ModuleInfo("Libor Va", "Module for .obj format import. Produces Mesh.", CategoryType.Loader, CategoryData.Mesh, IconName = "objLoader.ico")]
	[PortInfo("Filename","MveString that contains name of file to load. If nothing is connected, then the value of the FileName property is used")]
	[PortInfo("TriangleMesh","Mesh instance that contains vertices and triangles loaded from the file")]
	public class ObjLoader:Module
	{

		private bool paramsChanged = true;

		private bool reload = false;

        /// <summary>
        /// Determines whether or not to reload data each time the modul is asked for result. Set to true when the contens of the input files change, otherwise keep false.
        /// </summary>
		[Description("Determines whether or not to reload data each time the modul is asked for result. Set to true when the contens of the input files change, otherwise keep false.")]
		public bool Reload
		{
			get{return(reload);}
			set{reload = value;}
		}

		private string fileName = "mesh.obj";

		private TriangleMesh output;

		private string trianglesCellName = "triangles";

        /// <summary>
        /// Sets the identifier for the triangle cells in the output TriangleMesh.
        /// </summary>
		[Description("Sets the identifier for the triangle cells in the output TriangleMesh")]
		public string TrianglesCellName
		{
			get{return(trianglesCellName);}
			set
			{
				if (value=="")
				{
					throw new MveException(this, "Cell identifier of triangle in the ObjLoader module cannot be an empty string.");
				}
				else
				{
					this.trianglesCellName = value;
					paramsChanged = true;
				}
			}
		}

        /// <summary>
        /// File to load data from. Should be in the .obj format. The property has no effect when there's a connection at the Filename port
        /// </summary>
		[Editor(typeof(System.Windows.Forms.Design.FileNameEditor),typeof(System.Drawing.Design.UITypeEditor))]
		[Description("File to load data from. Should be in the .obj format. The property has no effect when there's a connection at the Filename port.")]
		public string FileName
		{
			get{return(fileName);}
			set
			{
				if (value != "")
				{
					fileName = value;						 
					paramsChanged = true;
				}
				else
				{
					throw new MveException(this, "FileName property cannot be set to an empty string!");
				}
			}
		}

        /// <summary>
        /// Default constructor.
        /// </summary>
		public ObjLoader()
		{
			AddOutPort("TriangleMesh", typeof(TriangleMesh));
			AddInPort("Filename",typeof(MveString),false);
		}

        /// <summary>
        /// Loads the input file.
        /// I property values remain unchanged since the last run, then the file is not reloaded, unless the Realod property is set to true.
        /// </summary>
		public override void Execute()
		{
			if ((paramsChanged)||(reload)||(output == null)||(!IsAllDataSame()))
			{
				// data is loaded only when needed
				string fn;

				if (GetInPort("Filename").Link!=null)
				{
					MveString inFileName = (MveString)GetInput("Filename");
					fn = inFileName.Value;
				}
				else
				{
					fn = fileName;
				}
                WriteLog("FileName", fn);

				// first pass determines the number of vertices and the number of triangles
                StreamReader sr;
                try
                {
                    sr = new StreamReader(fn);
                }
                catch (FileNotFoundException fnfe)
                {
                    throw new MveException(this, fnfe.Message);
                }

				int triangleCount = 0;
				int vertexCount = 0;
				int lineCount = 0;
				string line = sr.ReadLine();

                bool normalsFound = false;

				while (line != null)
				{
                    line = line.Trim();

					if (line.StartsWith("v ")) vertexCount++;
                    if (line.StartsWith("v ")) normalsFound = true;
                    if ((line.StartsWith("f "))||(line.StartsWith("fo ")))
                    {
                        triangleCount++;
                        if (line.Split(' ').Length == 5)
                            triangleCount++;                            
                    }
					line = sr.ReadLine();
					lineCount++;
				}

				sr.Close();

				// second pass performs the actual parsing
				sr = new StreamReader(fn);
				UniformDataArray vertices = new UniformDataArray(typeof(Point3D),vertexCount);
                UniformDataArray normals = null;
                if (normalsFound)
                    normals = new UniformDataArray(typeof(Normal3D), vertexCount);
				UniformDataArray triangles = new UniformDataArray(typeof(Triangle),triangleCount);

				int vi = 0;
				int ti = 0;
                int ni = 0;
				Point3D point;
                Normal3D normal;
				Triangle triangle;
				NumberFormatInfo nfi = new NumberFormatInfo();
				nfi.NumberDecimalSeparator = ".";
				nfi.NumberGroupSeparator = ",";                

				int linePos = 0;
				line = sr.ReadLine();
				while (line!=null)
				{
                    line = line.Trim();

					//parsing of a vertex
					if (line.StartsWith("v "))
					{
						string[] coords = line.Split(new char[] {' '},4,StringSplitOptions.RemoveEmptyEntries);

						Scalar c1 = double.Parse(coords[1],nfi);
						Scalar c2 = double.Parse(coords[2],nfi);
						Scalar c3 = double.Parse(coords[3],nfi);

						point = new Point3D(c1,c2,c3);
						vertices[vi] = point;
						vi++;
					}

                    //parsing of a vertex
                    if (line.StartsWith("vn "))
                    {
                        string[] coords = line.Split(new char[] { ' ' }, 4, StringSplitOptions.RemoveEmptyEntries);
                        Scalar c1,c2,c3;
                        try
                        {
                            c1 = double.Parse(coords[1], nfi);
                            c2 = double.Parse(coords[2], nfi);
                            c3 = double.Parse(coords[3], nfi);
                        }
                        catch (FormatException)
                        {
                            c1 = 0;
                            c2 = 0;
                            c3 = 1;
                        }

                        normal = new Normal3D(c1, c2, c3);
                        normals[ni] = normal;
                        ni++;
                    }

					// parsing of a triangle
                    if ((line.StartsWith("f ")) || (line.StartsWith("fo ")))
					{
                        string[] indices = line.Split(new char[] { ' ' }, 5,StringSplitOptions.RemoveEmptyEntries);

                        string part = indices[1].Split('/')[0];
						int v1 = int.Parse(part)-1;
                        part = indices[2].Split('/')[0];
						int v2 = int.Parse(part)-1;
                        part = indices[3].Split('/')[0];
						int v3 = int.Parse(part)-1;

						triangle = new Triangle(v1,v2,v3);
						triangles[ti] = triangle;
						ti++;

                        if (indices.Length == 5)
                        {
                            part = indices[4].Split('/')[0];
                            v2 = int.Parse(part) - 1;

                            triangle = new Triangle(v1, v3, v2);
                            triangles[ti] = triangle;
                            ti++;
                        }
					}
					line = sr.ReadLine();
					linePos++;
					if ((linePos%1000) == 0)
						ProgressInfo((float)linePos/lineCount);
				}

				// creating an output Mesh instance
				output = new TriangleMesh();
				output.Points = vertices;
				output.AddCells("triangles",triangles);
                if (normalsFound)
                    output.AddPointAttrs("Normal3D", normals);
				SetOutput("TriangleMesh",output);
				paramsChanged = false;
			}
			else
			{
				// nothing's changed, we just pass current data
                SetOutput("TriangleMesh", output, true);
			}
		}// Execute

	}//ObjLoader
}
