using System;
using System.Xml;
using Zcu.Mve.Core;

namespace Zcu.Mve.Visualization
{
	/// <summary>
	/// This structure represents a Pixel - elementary element of Grid.
	/// Pixel determinates the topological information. Pixel is given by
	/// four points. Real points coordinates are irrelevant. We work with
	/// the indices [i,j] of points in the Grid. We hold in the memory only
	/// the point no. 0 (left bottom corner of Pixel). Other points are
	/// computed automaticaly.
	/// </summary>
	/// <remarks>
	/// Indexing of points in the Pixel is given:
	/// [0,0] - point no. 0 - left bottom corner of Pixel;
	/// [0,1] - point no. 1 - right bottom corner of Pixel; 
	/// [1,0] - point no. 2 - left top corner of Pixel;
	/// [1,1] - point no. 3 - right top corner of Pixel;
	/// </remarks>
	public struct Pixel : IGridCell
	{
		/// <summary>
		/// Index of left bottom point of Pixel (in reference to Grid in the x-axes direction).
		/// </summary>
		private int i;

		/// <summary>
		/// Index of left bottom point of Pixel (in reference to Grid in the y-axes direction).
		/// </summary>
		private int j;

		/// <summary>
		/// Create Pixel object and fill it with data.
		/// </summary>
		/// <param name="i">Index of left bottom point of Pixel (in reference to Grid in the x-axes direction).</param>
		/// <param name="j">Index of left bottom point of Pixel (in reference to Grid in the y-axes direction).</param>
		public Pixel(int i, int j)
		{
			this.i = i;
			this.j = j;
		} // Pixel()

		/// <summary>
		/// Get the index of left bottom point of Pixel (in reference to Grid in the x-axes direction).
		/// </summary>
		public int I
		{
			get
			{
				return this.i;
			}
		} // I

		/// <summary>
		/// Get the index of left bottom point of Pixel (in reference to Grid in the y-axes direction).
		/// </summary>
		public int J
		{
			get
			{
				return this.j;
			}
		} // J

		/// <summary>
		/// Returns indices [i,j] of point in the Grid according to the index of point in the Pixel.
		/// Indices [i,j] refer to concrete Grid.
		/// </summary>
		/// <param name="index">Index of the point in the Pixel - can be only {0, 1, 2, 3}.</param>
		/// <returns>One-dimensional array 1 x 2 that represents indices [i,j] of the point in the Grid.</returns>
		public int[] GetPoint(int index)
		{
			return this[index];
		} // GetPoint()

		/// <summary>
		/// Returns indices [i,j] of point in the Grid according to indices [m,n] of point in the Pixel.
		/// Indices [i,j] refer to concrete Grid.
		/// </summary>
		/// <remarks>
		/// Indices [m, n] refer to the Pixel: 
		/// [0,0] - point no. 0 - left bottom corner of Pixel;
		/// [0,1] - point no. 1 - right bottom corner of Pixel; 
		/// [1,0] - point no. 2 - left top corner of Pixel;
		/// [1,1] - point no. 3 - right top corner of Pixel;
		/// </remarks>
		/// <param name="m">Index of the point in the Pixel (in x-axes direction) - can be only {0, 1}.</param>
		/// <param name="n">Index of the point in the Pixel (in y-axes direction) - can be only {0, 1}.</param>
		/// <returns>One-dimensional array 1 x 2 that represents indices [i,j] of the point in the Grid.</returns>
		public int[] GetPoint(int m, int n)
		{
			return this[m, n];
		} // GetPoint()
        
		/// <summary>
		/// Get one-dimensional array 1 x 2 that represents indices [i,j] of the point in the Grid.
		/// The index determines the point of the Pixel and can be {0, 1, 2, 3} only.
		/// </summary>
		public int[] this[int index]
		{
			get
			{
				if (index <= 3 && index >= 0)
					switch (index) 
					{
						case 0:		// bod 0 pixelu = levy spodni bod 
							return new int[] {this.i, this.j};
						case 1:		// bod 1 pixelu = pravy spodni bod
							return new int[] {this.i, this.j + 1};
						case 2: 	// bod 2 pixelu = levy horni bod
							return new int[] {this.i - 1, this.j};
						case 3:		// bod 3 pixelu = pravy horni bod
							return new int[] {this.i - 1, this.j + 1};
						default:
							return null;
					}
				else
					throw new MveException(this, "Index " + index + " is out of range {0, 1, 2, 3}");
			}
		} // this[int]

		/// <summary>
		/// Get one-dimensional array 1 x 2 that represents indices [i,j] of the point in the Grid.
		/// Indices [i,j] refer to the concrete Grid, while the indices [m,n] refer to the Pixel.
		/// </summary>
		/// <remarks>
		/// Indices [m, n] refer to the Pixel: 
		/// [0,0] - point no. 0 - left bottom corner of Pixel;
		/// [0,1] - point no. 1 - right bottom corner of Pixel; 
		/// [1,0] - point no. 2 - left top corner of Pixel;
		/// [1,1] - point no. 3 - right top corner of Pixel;
		/// </remarks>
		public int[] this[int m, int n]
		{
			get
			{
				if ((m == 0 || m == 1) && (n == 0 || n == 1))
				{
					if (m == 0 && n == 0)		// bod [0,0] pixelu = levy spodni bod
						return new int[] {this.i, this.j};
					else if (m == 0 && n == 1)	// bod [0,1] pixelu = pravy spodni bod
						return new int[] {this.i, this.j + 1};
					else if (m == 1 && n == 0)	// bod [1,0] pixelu = levy horni bod
						return new int[] {this.i - 1, this.j};
					else if (m == 1 && n == 1)	// bod [1,1] pixelu = pravy horni bod
						return new int[] {this.i - 1, this.j + 1};
					else
						return null;
				}
				else
					throw new MveException(this, "Index " + m + " or " + n + " is out of range {0, 1}");
			}
		} // this[int, int]

		/// <summary>
		/// Read data from Xml file to current structure.
		/// </summary>
		/// <param name="xmlTextReader">Stream to be read.</param>
		public void ReadData(System.Xml.XmlTextReader xmlTextReader)
		{
			string[] points = xmlTextReader.ReadString().Trim().Split(' ');
			this.i = Int32.Parse(points[0]);
			this.j = Int32.Parse(points[1]);
		} // ReadData()

		/// <summary>
		/// Write current data to Xml file.
		/// </summary>
		/// <param name="xmlTextWriter">Stream to be write.</param>
		public void WriteData(System.Xml.XmlTextWriter xmlTextWriter)
		{
			xmlTextWriter.WriteString(this.i + " " + this.j);
		} // WriteData()

		/// <summary>
		/// Deep copy of data object and its subelements.
		/// </summary>
		/// <returns>Same data, different memory space.</returns>
		public IDataObject DeepCopy()
		{
			return new Pixel(this.i, this.j);
		} // DeepCopy()

		/// <summary>
		/// Check the consistency of current data structure.
		/// Returns false if the Pixel is out of bounds of the Grid.
		/// Possible minimum is: [i,j] = [0,1].
		/// </summary>
		/// <returns>True if structure is consistent.</returns>
		public bool CheckConsistence()
		{
			bool isOK = true;
			if (this.i < 0 || this.j < 1)
			{
				Console.WriteLine("Error! " + this.GetType().Name + " is out of grid range.");
				isOK = false;
			}
			return isOK;
		} // CheckConsistence()

		/// <summary>
		/// Returns a String that represents the current Object.
		/// </summary>
		/// <returns>String that represents the current Object.</returns>
		public override string ToString()
		{
			return "Pixel: \n\tPoint 0: [" +  this.i      + ", " +  this.j      + "]"
				+ "\n\tPoint 1: [" +  this.i      + ", " + (this.j + 1) + "]"
				+ "\n\tPoint 2: [" + (this.i - 1) + ", " +  this.j      + "]"
				+ "\n\tPoint 3: [" + (this.i - 1) + ", " + (this.j + 1) + "]";
		} // ToString()

		/// <summary>
		/// Returns array of assembly name.
		/// </summary>
		public string [] LibNames
		{
			get
			{
				return new string [] {Globals.GetLibraryName(this.GetType())};
			}
		}

	} // class Pixel
} // namespace
