using System;
using System.Globalization;
using Zcu.Mve.Core;
using System.ComponentModel;

namespace Zcu.Mve.Visualization
{
	/// <summary>
	/// Represents a 3D point defined by three coordinates - X, Y and Z.
	/// </summary>
	[Zcu.Mve.Core.DataObject("Represents a 3D point defined by three coordinates - X, Y and Z.")]
  [TypeConverter(typeof(Point3DConverter))]
  public struct Point3D : IPoint
	{
    /// <summary>
    /// x-vector value.
    /// </summary>
    private double x;

    /// <summary>
    /// Get/Set x-vector value.
    /// </summary>
    public double X
    {
      get
      {
        return this.x;
      }
      set
      {
        this.x = value;
      }
    }

    /// <summary>
    /// y-vector value.
    /// </summary>
    private double y;

    /// <summary>
    /// Get/Set y-vector value.
    /// </summary>
    public double Y
    {
      get
      {
        return this.y;
      }
      set
      {
        this.y = value;
      }
    }

    /// <summary>
    /// z-vector value.
    /// </summary>
    private double z;

    /// <summary>
    /// Get/Set z-vector value.
    /// </summary>
    public double Z
    {
      get
      {
        return this.z;
      }
      set
      {
        this.z = value;
      }
    }

		/// <summary>
		/// The toleration for equality operator.
		/// </summary>
		public static double Epsilon = Globals.EpsilonDoubleImplicit;
		
		/// <summary>
		/// Create a Point3D object.
		/// </summary>
		/// <param name="x">x-coordinate of point.</param>
		/// <param name="y">y-coordinate of point.</param>
		/// <param name="z">z-coordinate of point.</param>
		public Point3D(double x, double y, double z)
		{
			this.x = x;
			this.y = y;
			this.z = z;
		} // Point3D()

        /// <summary>
        /// default constructor, constructs a new instance from an array of three integers. If smaller array is passed then an exception is thrown.
        /// </summary>
        /// <param name="position">Three element array of XYZ coordinates.</param>
      public Point3D(params double[] position)
      {
          if (position.Length < 3)
          {
              throw new MveException(typeof(Point3D), "Less than three parameters passed to Point3D constructor");
          }
          else
          {
              this.x = position[0];
              this.y = position[1];
              this.z = position[2];
          }
      }

		/// <summary>
		/// Read data from Xml file to current structure.
		/// </summary>
		/// <param name="xmlTextReader">Stream to be read.</param>
		public void ReadData(System.Xml.XmlTextReader xmlTextReader)
		{
			NumberFormatInfo nfi = new CultureInfo("en-US", false).NumberFormat;
			string[] pos = xmlTextReader.ReadString().Trim().Split(' ');
			this.X = Double.Parse(pos[0], nfi);
			this.Y = Double.Parse(pos[1], nfi);
			this.Z = Double.Parse(pos[2], nfi);
		} // ReadData()

		/// <summary>
		/// Write current data to Xml file.
		/// </summary>
		/// <param name="xmlTextWriter">Stream to be write.</param>
		public void WriteData(System.Xml.XmlTextWriter xmlTextWriter)
		{
			NumberFormatInfo nfi = new CultureInfo("en-US", false).NumberFormat;
			xmlTextWriter.WriteString(this.X.ToString(nfi) + " " 
				+ this.Y.ToString(nfi) + " "
				+ this.Z.ToString(nfi));
		} // WriteData()

		/// <summary>
		/// Deep copy of data object and its subelements.
		/// </summary>
		/// <returns>Same data, different memory space.</returns>
		public IDataObject DeepCopy()
		{
			return new Point3D(this.X, this.Y, this.Z);
		} // DeepCopy()

		/// <summary>
		/// Check the consistency of current data structure.
		/// </summary>
		/// <returns>True if structure is consistent.</returns>
		public bool CheckConsistence()
		{
			return true;
		} // CheckConsistence()

		/// <summary>
		/// Returns the distance of point and coordinate origin.
		/// </summary>
		/// <returns>The distance of point from and coordinate origin.</returns>
		public double Abs()
		{
			return Math.Sqrt(this.X * this.X + this.Y * this.Y + this.Z * this.Z);
		} // Abs()

		/// <summary>
		/// Get/Set element of point.
		/// </summary>
		public double this [int index]
		{
			get
			{
				switch(index)
				{
					case 0:
						return this.X;
					case 1:
						return this.Y;
					case 2:
						return this.Z;
					default:
						throw new IndexOutOfRangeException(this.GetType().FullName + ": Dimension exceeded.");
				}
			}
			set
			{
				switch(index)
				{
					case 0:
						this.X = value;
						break;
					case 1:
						this.Y = value;
						break;
					case 2:
						this.Z = value;
						break;
					default:
						throw new IndexOutOfRangeException(this.GetType().FullName + ": Dimension exceeded.");
				}
			}
		}

		/// <summary>
		/// Get number of elements of point.
		/// </summary>
    [Browsable(false)]
		public int Length
		{
			get
			{
				return 3;
			}
		}

		/// <summary>
		/// Returns array of assembly name.
		/// </summary>
    [Browsable(false)]
		public string [] LibNames
		{
			get
			{
				return new string [] {Globals.GetLibraryName(this.GetType())};
			}
		}

		// ===== Methods of the class Object ==============================================
		#region Overrided methods of the class Object

		/// <summary>
		/// Returns a String that represents the current Object.
		/// </summary>
		/// <returns>String that represents the current Object.</returns>
		override public string ToString()
		{
			return "Point3D: [" + this.X + "; " + this.Y + "; " + this.Z + "]";
		} // ToString()

		/// <summary>
		/// Determines whether the specified Objects are equal.
		/// </summary>
		/// <param name="a">First Object.</param>
		/// <param name="b">Second Object.</param>
		/// <returns>True if objects are equal. Otherwise, false.</returns>
		public static new bool Equals(object a, object b)
		{
			if (a.GetType() != typeof(Point3D) || b.GetType() != typeof(Point3D))
				return false;

			return Math.Abs(((Point3D)a).X - ((Point3D)b).X) <= Point3D.Epsilon
				&& Math.Abs(((Point3D)a).Y - ((Point3D)b).Y) <= Point3D.Epsilon
				&& Math.Abs(((Point3D)a).Z - ((Point3D)b).Z) <= Point3D.Epsilon;
		}
		
		/// <summary>
		/// Determines whether the specified Object is equal to the current Object.
		/// </summary>
		/// <param name="obj">Specified Object to compare with the current Object.</param>
		/// <returns>True if the specified Object is equal to the current Object. Otherwise, false.</returns>
		public override bool Equals(object obj)
		{
			return Point3D.Equals(this, obj);
		}
		
		/// <summary>
		/// Returns a hash code for the current Object.
		/// </summary>
		/// <returns>A hash code for the current Object.</returns>
		public override int GetHashCode()
		{
			return this.X.GetHashCode() ^ this.Y.GetHashCode() ^ this.Z.GetHashCode();
		}

		#endregion

		// ===== Conversions ==============================================================
		#region Conversion

		// ===== Point3D -> Point =================
		/// <summary>
		/// Explicit conversion Point3D -> Point2D
		/// </summary>
		/// <param name="p">Point3D</param>
		/// <returns>Point2D</returns>
		public static explicit operator Point2D(Point3D p) 
		{
			return new Point2D(p.X, p.Y);
		} // Point3D -> Point2D

		/// <summary>
		/// Implicit conversion Point3D -> PointND
		/// </summary>
		/// <param name="p">Point3D</param>
		/// <returns>PointND</returns>
		public static implicit operator PointND(Point3D p) 
		{
			return new PointND(new double[] {p.X, p.Y, p.Z});
		} // Point3D -> PointND

		// ===== Point3D -> Vector ================
		/// <summary>
		/// Explicit conversion Point3D -> Vector2D
		/// </summary>
		/// <param name="p">Point3D</param>
		/// <returns>Vector2D</returns>
		public static explicit operator Zcu.Mve.Numerics.Vector2D(Point3D p) 
		{
			return new Zcu.Mve.Numerics.Vector2D(p.X, p.Y);
		} // Point3D -> Vector2D

		/// <summary>
		/// Implicit conversion Point3D -> Vector3D
		/// </summary>
		/// <param name="p">Point3D</param>
		/// <returns>Vector3D</returns>
		public static implicit operator Zcu.Mve.Numerics.Vector3D(Point3D p) 
		{
			return new Zcu.Mve.Numerics.Vector3D(p.X, p.Y, p.Z);
		} // Point3D -> Vector3D

		/// <summary>
		/// Implicit conversion Point3D -> VectorND
		/// </summary>
		/// <param name="p">Point3D</param>
		/// <returns>VectorND</returns>
		public static implicit operator Zcu.Mve.Numerics.VectorND(Point3D p) 
		{
			return new Zcu.Mve.Numerics.VectorND(new double[] {p.X, p.Y, p.Z});
		} // Point3D -> VectorND

		// ===== Vector -> Point3D ================
		/// <summary>
		/// Implicit conversion Vector2D -> Point3D
		/// </summary>
		/// <param name="v">Vector2D</param>
		/// <returns>Point3D</returns>
		public static implicit operator Point3D(Zcu.Mve.Numerics.Vector2D v) 
		{
			return new Point3D(v.X, v.Y, 0);
		} // Vector2D -> Point3D

		/// <summary>
		/// Implicit conversion Vector3D -> Point3D
		/// </summary>
		/// <param name="v">Vector3D</param>
		/// <returns>Point3D</returns>
		public static implicit operator Point3D(Zcu.Mve.Numerics.Vector3D v) 
		{
			return new Point3D(v.X, v.Y, v.Z);
		} // Vector3D -> Point3D

		/// <summary>
		/// Explicit conversion VectorND -> Point3D
		/// </summary>
		/// <param name="v">VectorND</param>
		/// <returns>Point3D</returns>
		public static explicit operator Point3D(Zcu.Mve.Numerics.VectorND v) 
		{
			if (v.Values.Length == 1)
				return new Point3D(v.Values[0], 0, 0);
			else if (v.Values.Length == 2)
				return new Point3D(v.Values[0], v.Values[1], 0);
			else
				return new Point3D(v.Values[0], v.Values[1], v.Values[2]);
		} // VectorND -> Point3D

		#endregion

		// ===== Operators ================================================================
		#region Operators

		/// <summary>
		/// The equality operator (==) returns true if the values of its operands are equal, false otherwise.
		/// </summary>
		/// <param name="a">First operand.</param>
		/// <param name="b">Second operand.</param>
		/// <returns>True if the values of operands are equal, false otherwise.</returns>
		public static bool operator ==(Point3D a, Point3D b)
		{
			return Point3D.Equals(a, b);
		}
		
		/// <summary>
		/// The inequality operator (!=) returns false if its operands are equal, true otherwise.
		/// </summary>
		/// <param name="a">First operand.</param>
		/// <param name="b">Second operand.</param>
		/// <returns>False if the values of operands are equal, true otherwise.</returns>
		public static bool operator !=(Point3D a, Point3D b) 
		{
			return !(a == b);
		}

		/// <summary>
		/// Addition of two points.
		/// </summary>
		/// <param name="a">First operand.</param>
		/// <param name="b">Second operand.</param>
		/// <returns>Resulting sum (a + b).</returns>
		public static Point3D operator +(Point3D a, Point3D b)
		{
			return(new Point3D(a.X+b.X,a.Y+b.Y,a.Z+b.Z));
		}

		/// <summary>
		/// Subtraction of two points.
		/// </summary>
		/// <param name="a">First operand.</param>
		/// <param name="b">Second operand.</param>
		/// <returns>Resulting difference (a - b).</returns>
		public static Point3D operator -(Point3D a, Point3D b)
		{
			return(new Point3D(a.X-b.X, a.Y-b.Y, a.Z-b.Z));
        }

        /// <summary>
        /// Sign inverse of a point.
        /// </summary>
        /// <param name="p">Operand.</param>
        /// <returns>Resulting point.</returns>
        public static Point3D operator -(Point3D p)
        {
            return (new Point3D(-p.X, -p.Y, -p.Z));
        }

		/// <summary>
		/// Multiplication by constant number.
		/// </summary>
		/// <param name="p">Point to be multiplied.</param>
		/// <param name="m">multiplicator.</param>
		/// <returns>Resulting point (m*p.x, m*p.y, m*p.z)</returns>
		public static Point3D operator *(Point3D p, double m)
		{
			return(new Point3D(p.X*m, p.Y*m, p.Z*m));
		}

		/// <summary>
		/// Multiplication by a scalar.
		/// </summary>
		/// <param name="p">Point to be multiplied.</param>
		/// <param name="m">Multiplicator.</param>
		/// <returns>Resulting point (m*p.x, m*p.y, m*p.z).</returns>
		public static Point3D operator *(double m, Point3D p)
		{
			return(new Point3D(p.X*m, p.Y*m, p.Z*m));
		}

		/// <summary>
		/// Division by a scalar.
		/// </summary>
		/// <param name="p">Point to be divided.</param>
		/// <param name="m">Divider.</param>
		/// <returns>Resulting point (p.x/m, p.y/m, p.z/m).</returns>
		public static Point3D operator /(Point3D p, double m)
		{
			return(new Point3D(p.X/m, p.Y/m, p.Z/m));
		}

		#endregion

		// ===== Operations ===============================================================
		#region Operations

		/// <summary>
		/// Returns euclidean distance between two points.
		/// </summary>
		/// <param name="p">Argument point to which distance will be computed.</param>
		/// <returns>Euclidean distance between the points.</returns>
		public double DistanceTo(Point3D p)
		{
			return(Math.Sqrt((p.X-this.X)*(p.X-this.X)+(p.Y-this.Y)*(p.Y-this.Y)+(p.Z-this.Z)*(p.Z-this.Z)));
		}

		/// <summary>
		/// Returns squared euclidean distance between two points
		/// (is slightly faster than the DistanceTo function as no square root is computed),
		/// </summary>
		/// <param name="p">Argument point to which distance will be computed.</param>
		/// <returns>Square euclidean distance between the points.</returns>
		public double SquareDistanceTo(Point3D p)
		{
			return((p.X-this.X)*(p.X-this.X)+(p.Y-this.Y)*(p.Y-this.Y)+(p.Z-this.Z)*(p.Z-this.Z));
		}

		/// <summary>
		/// Returns the dot product of two points.
		/// </summary>
		/// <param name="arg">Second argument of the dot product operation.</param>
		/// <returns>Scalar value of this.arg.</returns>
		public double DotProduct(Point3D arg)
		{
			return(this.X*arg.X+this.Y*arg.Y+this.Z*arg.Z);
		}

		/// <summary>
		/// Returns the cross product of two points.
		/// </summary>
		/// <param name="arg">Second argument of the cross product operation.</param>
		/// <returns>Point3D representing the resulting vector.</returns>
		public Point3D CrossProduct(Point3D arg)
		{
			return(new Point3D(this.Y*arg.Z-this.Z*arg.Y,this.Z*arg.X-this.X*arg.Z,this.X*arg.Y-this.Y*arg.X));
		}

		#endregion

	} // class Point3D
} // namespace
