using System;
using System.Collections.Generic;
using System.Text;
using System.ComponentModel;
using System.Globalization;
using System.ComponentModel.Design.Serialization;
using System.Reflection;

namespace Zcu.Mve.Visualization
{
  /// <summary>
  /// Converts Point3D from one data type to another. Access this class through the TypeDescriptor.
  /// </summary>
  public class Point3DConverter : TypeConverter
  {
    /// <summary>
    /// Returns whether this object supports properties, using the specified context.
    /// </summary>
    /// <param name="context">An ITypeDescriptorContext that provides a format context.</param>
    /// <returns>true if GetProperties should be called to find the properties of this object; otherwise, false.</returns>
    public override bool GetPropertiesSupported(ITypeDescriptorContext context)
    {
      return true;
    }

    /// <summary>
    /// Returns whether changing a value on this object requires a call to CreateInstance to create a new value, using the specified context.
    /// </summary>
    /// <param name="context">An ITypeDescriptorContext that provides a format context.</param>
    /// <returns>true if changing a property on this object requires a call to CreateInstance to create a new value; otherwise, false.</returns>
    public override bool GetCreateInstanceSupported(ITypeDescriptorContext context)
    {
      return true;
    }

    /// <summary>
    /// Creates an instance of the type that this TypeConverter is associated with, using the specified context, given a set of property values for the object.
    /// </summary>
    /// <param name="context">An ITypeDescriptorContext that provides a format context.</param>
    /// <param name="propertyValues">An IDictionary of new property values.</param>
    /// <returns>
    /// An Object representing the given IDictionary, or a null reference if the object cannot be created.
    /// </returns>
    public override object CreateInstance(ITypeDescriptorContext context, System.Collections.IDictionary propertyValues)
    {
      return new Point3D((double) propertyValues["X"], (double) propertyValues["Y"], (double) propertyValues["Z"]);
    }

    /// <summary>
    /// Retrieves the set of properties for this type.
    /// </summary>
    /// <param name="context">A type descriptor through which additional context can be provided.</param>
    /// <param name="value">The value of the object to get the properties for.</param>
    /// <param name="attributes">An array of Attribute objects that describe the properties.</param>
    /// <returns>
    /// The set of properties that are exposed for this data type. If no properties are exposed, this method might return a null reference.
    /// </returns>
    public override PropertyDescriptorCollection GetProperties(ITypeDescriptorContext context, object value, Attribute[] attributes)
    {
      PropertyDescriptorCollection collection = TypeDescriptor.GetProperties(typeof(Point3D), attributes);

      
      return collection.Sort(new string[] { "X", "Y", "Z" });
    }

    /// <summary>
    /// Determines if this converter can convert an object in the given source type to the native type of the converter.
    /// </summary>
    /// <param name="context">
    /// A formatter context. This object can be used to get additional information about the environment this converter is being called from.
    /// This may be a null reference, so you should always check. Also, properties on the context object may also return a null reference. 
    /// </param>
    /// <param name="sourceType">The type you want to convert from.</param>
    /// <returns>true if this object can perform the conversion; otherwise, false.</returns>
    public override bool CanConvertFrom(ITypeDescriptorContext context, Type sourceType)
    {
      if (sourceType == typeof(string))
      {
        return true;
      }
      return base.CanConvertFrom(context, sourceType);
    }

    /// <summary>
    /// Gets a value indicating whether this converter can convert an object to the given destination type using the context.
    /// </summary>
    /// <param name="context">An ITypeDescriptorContext object that provides a format context.</param>
    /// <param name="destinationType">A Type object that represents the type you want to convert to.</param>
    /// <returns>true if this converter can perform the conversion; otherwise, false.</returns>
    public override bool CanConvertTo(ITypeDescriptorContext context, Type destinationType)
    {
      if (destinationType == typeof(InstanceDescriptor))
      {
        return true;
      }
      return base.CanConvertTo(context, destinationType);
    }

    /// <summary>
    /// Converts the specified object to the specified type.
    /// </summary>
    /// <param name="context">
    /// A formatter context. This object can be used to get additional information about the environment this converter is being called from.
    /// This may be a null reference, so you should always check.
    /// Also, properties on the context object may also return a null reference.
    /// </param>
    /// <param name="culture">
    /// An object that contains culture specific information, such as the language, calendar, and cultural conventions associated with a specific culture.
    /// It is based on the RFC 1766 standard. 
    /// </param>
    /// <param name="value">The object to convert.</param>
    /// <param name="destinationType">The type to convert the object to.</param>
    /// <returns>The converted object.</returns>
    public override object ConvertTo(ITypeDescriptorContext context, System.Globalization.CultureInfo culture, object value, Type destinationType)
    {
      if (destinationType == null)
      {
        throw new ArgumentNullException("destinationType");
      }

      if ((destinationType == typeof(string)) && (value is Point3D))
      {
        Point3D point = (Point3D) value;

        if (culture == null)
        {
          culture = CultureInfo.CurrentCulture;
        }

        string text = culture.TextInfo.ListSeparator + " ";

        TypeConverter converter = TypeDescriptor.GetConverter(typeof(float));

        string[] textArray = new string[3];

        int num = 0;
        textArray[num++] = converter.ConvertToString(context, culture, point.X);
        textArray[num++] = converter.ConvertToString(context, culture, point.Y);
        textArray[num++] = converter.ConvertToString(context, culture, point.Z);

        return string.Join(text, textArray);
      }

      if ((destinationType == typeof(InstanceDescriptor)) && (value is Point3D))
      {
        Point3D point = (Point3D) value;

        ConstructorInfo info = typeof(Point3D).GetConstructor(new Type[] { typeof(double), typeof(double), typeof(double) });
        if (info != null)
        {
          return new InstanceDescriptor(info, new object[] { point.X, point.Y, point.Z });
        }
      }

      return base.ConvertTo(context, culture, value, destinationType);
    }

    /// <summary>
    /// Converts the specified object to a Point3D object. 
    /// </summary>
    /// <param name="context">
    /// A formatter context. This object can be used to get additional information about the environment this converter is being called from.
    /// This may be a null reference, so you should always check. Also, properties on the context object may also return a null reference. 
    /// </param>
    /// <param name="culture">
    /// An object that contains culture specific information, such as the language, calendar, and cultural conventions associated with a specific culture.
    /// It is based on the RFC 1766 standard.
    /// </param>
    /// <param name="value">The object to convert.</param>
    /// <returns>The converted object.</returns>
    public override object ConvertFrom(ITypeDescriptorContext context, System.Globalization.CultureInfo culture, object value)
    {
      if (!(value is string))
      {
        return base.ConvertFrom(context, culture, value);
      }

      string text = ((string) value).Trim();

      if (text.Length == 0)
      {
        return null;
      }

      if (culture == null)
      {
        culture = CultureInfo.CurrentCulture;
      }

      char ch = culture.TextInfo.ListSeparator[0];
      string[] textArray = text.Split(new char[] { ch });
      double[] numArray = new double[textArray.Length];
      TypeConverter converter = TypeDescriptor.GetConverter(typeof(double));
      for (int num = 0; num < numArray.Length; num++)
      {
        numArray[num] = (double) converter.ConvertFromString(context, culture, textArray[num]);
      }

      if (numArray.Length != 3)
      {
        throw new ArgumentException("TextParseFailedFormat");
      }

      return new Point3D(numArray[0], numArray[1], numArray[2]);
    }
  }
}
