using System;
using System.Drawing;
using System.ComponentModel;
using Zcu.Mve.Core;
using Zcu.Mve.Numerics;

namespace Zcu.Mve.Visualization.Renderers
{
	/// <summary>
	/// Module for rendering of RegGrid2D (essentially picture).
	/// </summary>
	[ModuleInfo("Miroslav Vavruka, Libor Va",
		 "Module for rendering of RegGrid2D (essentially picture). The user may choose which cell attribute to display (the AttrName property). Attributes of Scalar and ColorRGBA type may be used. When Scalar values are drawn, then values are normalized so that the whole white->black scale is used in the image.",
		 CategoryType.Output,
         CategoryData.Number,
         Assemblied = "2004-09-07",
		 IconName = "RegGrid2DRenderer.ico")]
	[PortInfo("Input", "RegGrid2D to render.")]
	public class RegGrid2DRenderer : Zcu.Mve.Core.Module
	{
		/// <summary>
		/// Panel for picture rendering.
		/// </summary>
		private RegGrid2DRendererPanel panel;
		/// <summary>
		/// Picture to render.
		/// </summary>
		private Bitmap bmp;

		string attrName = "values";
		bool paramsChanged = true;

		/// <summary>
		/// Property that encapsulates name fo the input attribute to be displayed
		/// </summary>
		/// 
		[Description("Name of the cell attribute to be displayed. Must be present in input RegGrid2D and must be of Scalar or ColorRGBA type.")]
		public string AttrName
		{
			get
			{
				return(attrName);
			}
			set
			{
				if (value!="")
				{
					attrName = value;
				}
				else
				{
					throw new MveException(this, "Input attribute name cannot be an empty string.");
				}
			}
		}
		
		/// <summary>
		/// Cretes module and input port.
		/// </summary>
		public RegGrid2DRenderer()
		{
			this.AddInPort("Input", typeof (Visualization.RegGrid2D));
		} // RegGrid2DRenderer()
		
		/// <summary>
		/// Executes Renderer functionality.
		/// </summary>
		public override void Execute()
		{
			if ((paramsChanged)||(bmp==null)||(!IsDataSame("Input")))
			{
				RegGrid2D data = (RegGrid2D) this.GetInput("Input");
				UniformDataArray2D colors = data.GetCellAttrs(attrName);

				if (colors.ElemType == typeof(ColorRGBA))
				{
					bmp = new Bitmap(data.Width, data.Height);
					for(int i = 0; i < data.Width; i++)
						for(int j = 0; j < data.Height; j++)
							bmp.SetPixel(i, j, (ColorRGBA) colors[j, i]);
					paramsChanged = false;
				}
				else if (colors.ElemType == typeof(Scalar))
				{
					double min = double.MaxValue;
					double max = double.MinValue;
					double v;

					for(int i = 0; i < data.Width; i++)
						for(int j = 0; j < data.Height; j++)
						{
							v = (Scalar)colors[j,i];
							if (v<min) min = v;
							if (v>max) max = v;
						}

					bmp = new Bitmap(data.Width, data.Height);
					for(int i = 0; i < data.Width; i++)
						for(int j = 0; j < data.Height; j++)
						{
							int val = (int)((((Scalar)colors[j,i]-min)/(max-min))*255);
							bmp.SetPixel(i, j, Color.FromArgb(val,val,val));
						}
					paramsChanged = false;
				}
				else
				{
					throw new MveException(this, "Input attribute must be UniformDataArray2D of either ColorRGBA or Scalar values.");
				}
			}


			System.Threading.Thread regGridRen = new System.Threading.Thread(new System.Threading.ThreadStart(this.RunRenderer));
			regGridRen.Start();
		} // Execute()

		/// <summary>
		/// Show the Renderer panel and render the picture.
		/// </summary>
		public void RunRenderer()
		{
			panel = new RegGrid2DRendererPanel(bmp);
			panel.ShowDialog();
		}
	} // class RegGrid2DRenderer
} // namespace
