﻿using System;
using Zcu.Mve.Core;
using Zcu.Mve.Numerics;
using Zcu.Mve.Visualization;
using System.Collections;
using System.Collections.Generic;
using System.ComponentModel;

namespace Zcu.Mve.Visualization.Processing
{
    /// <summary>
    /// Module for slicing volume data
    /// </summary>
    /// 
    [ModuleInfo("Vladimír Geršl", "Module for slicing volume data.", CategoryType.Filter, CategoryData.Volume, IconName = "Slicer.ico")]
    [PortInfo("Input array", "RegGridND of any kind of data and any dimension to be sliced.")]
    [PortInfo("Output triangle array", "TriangleMesh that represent a slice of the input data. Points have attributes: Normal3D and ColorRGBA (default name).")]
    [PortInfo("Output data array", "UniformDataArray2D that represent a slice of the input data.")]
    public class Slicer : Module
    {
        string coordinates = "10,5,1,0";

        bool filtrate = true;

        String outAttrs = "ColorRGBA";

        UniformDataArray2D output = null;

        MveArray cube = new MveArray(typeof(Scalar), 2, 2, 2);
        double offsetX, offsetY, offsetZ;
        double[] approxX = new double[4];   //front down, front up, rear down, rear up
        double[] approxY = new double[2];   //front, rear
        double approxZ;

        private List<Point3D> pts = null;
        private List<Triangle> indices = null;

        private TriangleMesh mesh = null; // output data
        private UniformDataArray points, triangles; // arrays of primitives for output

        private Point3D[,] body = null;

        private Point3D positionOrigin;

        bool paramsChanged = true;

        string attrName = "Values";

        /// <summary>
        /// Input attribut name.
        /// </summary>
        [Category("Input")]
        [Description("Attribut name.")]
        public string InAttrName
        {
            get { return (attrName); }
            set
            {
                if (value == "")
                {
                    throw new MveException(this, "Attribute name cannot be an empty string.");
                }
                else
                {
                    attrName = value;
                    paramsChanged = true;
                }
            }
        }


        /// <summary>
        /// Output attribut name.
        /// </summary>
        [Category("Output")]
        [Description("Output attribut name.")]
        public string OutAttrName
        {
            get { return (outAttrs); }
            set
            {
                if (value == "")
                {
                    throw new MveException(this, "Attribute name cannot be an empty string.");
                }
                else
                {
                    outAttrs = value;
                    paramsChanged = true;
                }
            }
        }

        /// <summary>
        /// Comma separated double values that represent coefficients a, b, c, d which defines plane of slice: ax + by + cz + d = 0.
        /// </summary>
        [Category("Slicer")]
        [Description("Comma separated double values that represent coefficients a, b, c, d which defines plane of slice: ax + by + cz + d = 0. Values a, b, c defines normal vector and d is an offset.")]
        public string Parameters
        {
            get { return (coordinates); }
            set
            {
                coordinates = value;
                paramsChanged = true;
            }
        }



        int widthRes = 400;

        double[] coords = null;
        RegGridND inputTmp = null;
        MveArray input = null;

        /// <summary>
        /// Index of the dimension that will be used as width and as height of the slice. Must be >=1.
        /// </summary>
        [Category("Slicer")]
        [Description("Index of the dimension that will be used as width and as height of the slice. Must be >=1.")]
        public int Resolution
        {
            get { return (widthRes); }
            set
            {
                if (value > 0)
                {
                    widthRes = value;
                    paramsChanged = true;
                }
                else
                {
                    throw new MveException(this, "Value of width dimension cannot be lower than one.");
                }
            }
        }

        /// <summary>
        /// Trillinear interpolation. If is it 'true', then interpolate values.
        /// </summary>
        [Category("Slicer")]
        [Description("Trillinear interpolation. If is it 'true', then interpolate values.")]
        public bool Interpolation
        {
            get { return (filtrate); }
            set
            {
                if (value != filtrate)
                {
                    filtrate = value;
                    paramsChanged = true;
                }
            }
        }

        /// <summary>
        /// Define one input and two outputs.
        /// </summary>
        public Slicer()
        {
            AddInPort("Input array", typeof(RegGridND));
            AddOutPort("Output triangle array", typeof(TriangleMesh));
            AddOutPort("Output data array", typeof(UniformDataArray2D));
        }

        /// <summary>
        /// Provide function of all modul.
        /// </summary>
        public override void Execute()
        {
            if ((paramsChanged) || (!IsDataSame("Input array")) || (output == null))
            {

                if (widthRes < 1)
                {
                    throw new MveException(this, "Width dimension cannot be less then 1.");
                }
                int ticks = Environment.TickCount;

                string[] coordinateStrings = coordinates.Split(new char[] { ',' });

                inputTmp = (RegGridND)GetInput("Input array");

                if (inputTmp.UserTags.ContainsKey("Position") && !inputTmp.UserTags["Position"].ToString().Equals(""))
                {
                    positionOrigin = (Point3D)inputTmp.UserTags["Position"];
                }
                else
                {
                    positionOrigin = new Point3D(0, 0, 0);
                }

               // 13.10.2009 Michal Novak - No used
               // ProgressInfoChangedEventHandler bar = new ProgressInfoChangedEventHandler(UpdateProgress);

                indices = new List<Triangle>();
                pts = new List<Point3D>();

                input = inputTmp.GetPointAttrs(attrName);

                if (input.ElemType != typeof(Scalar))
                {
                    throw new MveException(this, "Input array element type must be Scalar.");
                }

                if (coordinateStrings.Length != 4)
                {
                    throw new MveException(this, "Wrong number of coordinates set, must be equal to 4.");
                }

                coords = new double[coordinateStrings.Length];

                for (int i = 0; i < coords.Length; i++)
                {
                    coords[i] = Double.Parse(coordinateStrings[i]);
                }

                output = new UniformDataArray2D(typeof(Scalar), widthRes, widthRes);
                body = new Point3D[widthRes, widthRes];

                System.Console.WriteLine(175);
                
                double absA = Math.Abs(coords[0]);
                double absB = Math.Abs(coords[1]);
                double absC = Math.Abs(coords[2]);

                double max = Math.Max(Math.Max(absA, absB), absC);

                if (max == absA)
                {
                    plane(1);
                }
                else
                {
                    if (max == absB)
                    {
                        plane(2);
                    }
                    else
                    {
                        if (max == absC)
                        {
                            plane(3);
                        }
                        else
                            plane(1);
                    }
                }

                mesh = new TriangleMesh();

                points = new UniformDataArray(typeof(Point3D), pts.Count);

                for (int i = 0; i < pts.Count; i++)
                    points[i] = pts[i];

                System.Console.WriteLine(pts.Count);

                mesh.Points = points;

                UniformDataArray normals = new UniformDataArray(typeof(Normal3D), widthRes * widthRes);
                for (int i = 0; i < widthRes * widthRes; i++)
                    normals[i] = new Normal3D((double)coords[0], (double)coords[1], (double)coords[2]);
                mesh.AddPointAttrs("Normal3D", normals);

                int index = 0;

                int delkaHrany = output.Lengths[0];

                for (int i = 0; i < output.Lengths[0] - 1; i++)
                    for (int j = 0; j < output.Lengths[1] - 1; j++)
                    {
                        indices.Add(new Triangle(i * delkaHrany + j, i * delkaHrany + j + 1, i * delkaHrany + j + delkaHrany));
                        index++;
                        indices.Add(new Triangle(i * delkaHrany + j + delkaHrany + 1, i * delkaHrany + j + 1, i * delkaHrany + j + delkaHrany));
                        index++;
                    }

                System.Console.WriteLine((output.Lengths[0] - 1) + ", " + (output.Lengths[1] - 1) + " index: " + index);

                triangles = new UniformDataArray(typeof(Triangle), indices.Count);
                for (int i = 0; i < indices.Count; i++)
                    triangles[i] = indices[i];

                mesh.AddCells("Triangles", triangles);

                UniformDataArray pointsAttrs = new UniformDataArray(typeof(ColorRGBA), output.Array.Length);

                for (int i = 0; i < output.Lengths[1]; i++)
                    for (int j = 0; j < output.Lengths[0]; j++)
                    {
                        pointsAttrs[j * output.Lengths[1] + i] = new ColorRGBA((byte)(Scalar)output[j, i], (byte)(Scalar)output[j, i], (byte)(Scalar)output[j, i]);
                    }

                mesh.AddPointAttrs(outAttrs, pointsAttrs);

                triangles = null;
                points = null;
                indices = null;
                pts = null;

                if (/*ratio >= 2.0f && */filtrate == true)
                {
                    SetOutput("Output triangle array", mesh);
                    SetOutput("Output data array", output);
                }
                else
                {
                    SetOutput("Output triangle array", mesh);
                    SetOutput("Output data array", output);
                }
                System.Console.WriteLine(string.Format("Loading time: {0} ms", Environment.TickCount - ticks));

                paramsChanged = false;
            }
            else
            {
                SetOutput("Output triangle array", mesh, true);
                SetOutput("Output data array", output, true);
            }
        }

        /// <summary>
        /// Map slice size on size of volumetric data and finde the point on slicing plane.
        /// </summary>
        private void plane(int whichPlane)
        {
            for (int i = 0; i < output.Lengths[1]; i++)
                for (int j = 0; j < output.Lengths[0]; j++)
                {
                    double xv, yv, zv;
                    if (whichPlane == 1)
                    {
                        yv = (double)input.Lengths[1] * ((double)i / (double)output.Lengths[0]);
                        zv = (double)input.Lengths[2] * ((double)j / (double)output.Lengths[1]);
                        xv = (((double)coords[1] * yv + (double)coords[2] * zv + (double)coords[3])) / (double)coords[0];
                    }
                    else
                    {
                        if (whichPlane == 2)
                        {
                            xv = (double)input.Lengths[0] * ((double)i / (double)output.Lengths[0]);
                            zv = (double)input.Lengths[2] * ((double)j / (double)output.Lengths[1]);
                            yv = (((double)coords[0] * xv + (double)coords[2] * zv + (double)coords[3])) / (double)coords[1];
                        }
                        else
                        {
                            xv = (double)input.Lengths[0] * ((double)i / (double)output.Lengths[0]);
                            yv = (double)input.Lengths[1] * ((double)j / (double)output.Lengths[1]);
                            zv = (((double)coords[0] * xv + (double)coords[1] * yv + (double)coords[3])) / (double)coords[2];
                        }
                    }


                    if (!filtrate)
                    {
                        int x, y, z;

                        x = (int)Math.Floor(xv + positionOrigin.X);
                        y = (int)Math.Floor(yv + positionOrigin.Y);
                        z = (int)Math.Floor(zv + positionOrigin.Z);

                        pts.Add(new Point3D(xv, yv, zv));

                        int[] coor = { x, y, z };

                        if (input.Lengths[0] > x && input.Lengths[1] > y && input.Lengths[2] > z)
                        {

                            Scalar inp = (Scalar)(/*(Integer)*/input[coor]);

                            if (inp > (Scalar)(/*(Integer)*/255))
                                inp = (Scalar)(/*(Integer)*/255);

                            if (inp < (Scalar)(/*(Integer)*/0))
                                inp = (Scalar)(/*(Integer)*/0);

                            if (j > 0 || i > 0)
                            {
                                output[j, i] = inp;
                            }
                            else
                                output[j, i] = new Scalar(0);// input[0, 0, 0];
                            output[0, 0] = new Scalar(255);
                        }
                        else
                        {
                            if (j > 0 || i > 0)
                            {
                                output[j, i] = new Scalar(255);
                            }
                            else
                                output[j, i] = new Scalar(0);
                        }
                    }
                    else
                    {
                        output[j, i] = Interpolate(xv, yv, zv);
                        pts.Add(new Point3D(xv, yv, zv));
                    }
                }
            System.Console.WriteLine("plane3");
        }

        /* 13.10.2009 Michal Novak - No used
        /// <summary>
        /// Change progress bar value.
        /// </summary>
        private void UpdateProgress(object sender, double percent)
        {
            ProgressInfo(percent);
        }
        */

        /// <summary>
        /// Interpolate value from eight clothes points (trillinear interpolation).
        /// </summary>
        private Scalar Interpolate(double xv, double yv, double zv)
        {
            Scalar value = 0;
            int x, y, z;


            x = (int)Math.Floor(xv + positionOrigin.X);
            y = (int)Math.Floor(yv + positionOrigin.Y);
            z = (int)Math.Floor(zv + positionOrigin.Z);



            offsetX = xv - (double)x;
            offsetY = yv - (double)y;
            offsetZ = zv - (double)z;

            if (input.Lengths[0] > x + 1 && input.Lengths[1] > y + 1 && input.Lengths[2] > z + 1)
            {
                cube[0, 0, 0] = input[x, y, z];
                cube[1, 0, 0] = input[x + 1, y, z];
                cube[0, 1, 0] = input[x, y + 1, z];
                cube[0, 0, 1] = input[x, y, z + 1];
                cube[1, 1, 0] = input[x + 1, y + 1, z];
                cube[1, 0, 1] = input[x + 1, y, z + 1];
                cube[0, 1, 1] = input[x, y + 1, z + 1];
                cube[1, 1, 1] = input[x + 1, y + 1, z + 1];
            }
            else
            {
                return (Scalar)0;
            }

            approxX[0] = (double)((Scalar)cube[0, 0, 0]) * (double)(1.0 - offsetX) + (double)((Scalar)cube[1, 0, 0]) * offsetX;
            approxX[1] = (double)((Scalar)cube[0, 1, 0]) * (double)(1.0 - offsetX) + (double)((Scalar)cube[1, 1, 0]) * offsetX;
            approxX[2] = (double)((Scalar)cube[0, 0, 1]) * (double)(1.0 - offsetX) + (double)((Scalar)cube[1, 0, 1]) * offsetX;
            approxX[3] = (double)((Scalar)cube[0, 1, 1]) * (double)(1.0 - offsetX) + (double)((Scalar)cube[1, 1, 1]) * offsetX;

            approxY[0] = approxX[0] * (double)(1.0 - offsetY) + approxX[1] * offsetY;
            approxY[1] = approxX[2] * (double)(1.0 - offsetY) + approxX[3] * offsetY;

            approxZ = approxY[0] * (double)(1.0 - offsetZ) + approxY[1] * offsetZ;

            value = (Scalar)/*Math.Floor*/(approxZ);

            return value;
        }
    }
}