﻿using System;
using System.Collections.Generic;
using System.Text;
using Zcu.Mve.Visualization;
using Zcu.Mve.Core;
using Zcu.Mve.Numerics;

namespace Zcu.Mve.Visualization.Processing
{
	/// <summary>
	/// Abstract class for interpolation of values stored in grid
	/// </summary>
	abstract class Interpolator {
		protected delegate IDataObject Lerp(IDataObject first, IDataObject second, double d);
		protected delegate IDataObject Cubic(IDataObject a, IDataObject b, IDataObject c, IDataObject d, double dif);
		protected delegate IDataObject InterpolateAt(Point3D point);

		/// <summary>
		/// Indicates if a point is out of bounds
		/// </summary>
		/// <param name="point">point</param>
		/// <returns>true if point is out of bounds</returns>
		abstract public bool IsOut(Point3D point);

		/// <summary>
		/// Interpolates value at a given point
		/// </summary>
		/// <param name="point">point</param>
		/// <returns>interpolated value</returns>
		abstract public IDataObject InterpolateAtPoint(Point3D point);

		/// <summary>
		/// Interpolates values at all points in array
		/// </summary>
		/// <param name="points">point array</param>
		/// <returns>interpolated values array</returns>
		abstract public UniformDataArray Compute(UniformDataArray points);


		#region linear interpolations of different value types

		protected IDataObject LerpScalar(IDataObject first, IDataObject second, double d) {
			return (Scalar)first + ((Scalar)second - (Scalar)first) * d;
		}

		protected IDataObject LerpInteger(IDataObject first, IDataObject second, double d) {
			return (Integer)first + (Integer)(((Integer)second - (Integer)first) * d);
		}

		protected IDataObject LerpPoint2D(IDataObject first, IDataObject second, double d) {
			return (Point2D)first + ((Point2D)second - (Point2D)first) * d;
		}

		protected IDataObject LerpPoint3D(IDataObject first, IDataObject second, double d) {
			return (Point3D)first + ((Point3D)second - (Point3D)first) * d;
		}

		protected IDataObject LerpVector2D(IDataObject first, IDataObject second, double d) {
			Vector2D a = (Vector2D)first;
			Vector2D b = (Vector2D)second;
			return new Vector2D(a.X + (b.X - a.X) * d, a.Y + (b.Y - a.Y) * d);
		}

		protected IDataObject LerpVector2DTo3D(IDataObject first, IDataObject second, double d) {
			if (first is Vector3D && second is Vector3D) {
				return LerpVector3D(first, second, d);
			}
			else {
				Vector2D a = (Vector2D)first;
				Vector2D b = (Vector2D)second;
				return new Vector3D(a.X + (b.X - a.X) * d, 0, a.Y + (b.Y - a.Y) * d);
			}
		}

		protected IDataObject LerpVector3D(IDataObject first, IDataObject second, double d) {
			Vector3D a = (Vector3D)first;
			Vector3D b = (Vector3D)second;
			return new Vector3D(a.X + (b.X - a.X) * d, a.Y + (b.Y - a.Y) * d, a.Z + (b.Z - a.Z) * d);
		}

		#endregion

		#region cubic interpolations of different value types

		protected IDataObject CubicScalar(IDataObject a, IDataObject b, IDataObject c, IDataObject d, double diff) {
			double x0 = (Scalar)a;
			double x1 = (Scalar)b;
			double x2 = (Scalar)c;
			double x3 = (Scalar)d;

			double diff2 = diff * diff;
			double v0 = x3 - x2 - x0 + x1;
			double v1 = x0 - x1 - v0;
			double v2 = x2 - x0;
			double v3 = x1;

			return (Scalar)(v0 * diff * diff2 + v1 * diff2 + v2 * diff + v3);
		}

		protected IDataObject CubicInteger(IDataObject a, IDataObject b, IDataObject c, IDataObject d, double diff) {
			int x0 = (Integer)a;
			int x1 = (Integer)b;
			int x2 = (Integer)c;
			int x3 = (Integer)d;

			double diff2 = diff * diff;
			int v0 = x3 - x2 - x0 + x1;
			int v1 = x0 - x1 - v0;
			int v2 = x2 - x0;
			int v3 = x1;

			return (Integer)(v0 * diff * diff2 + v1 * diff2 + v2 * diff + v3);
		}

		protected IDataObject CubicPoint2D(IDataObject a, IDataObject b, IDataObject c, IDataObject d, double diff) {
			Point2D x0 = (Point2D)a;
			Point2D x1 = (Point2D)b;
			Point2D x2 = (Point2D)c;
			Point2D x3 = (Point2D)d;

			double diff2 = diff * diff;
			Point2D v0 = x3 - x2 - x0 + x1;
			Point2D v1 = x0 - x1 - v0;
			Point2D v2 = x2 - x0;
			Point2D v3 = x1;

			return (Point2D)(v0 * diff * diff2 + v1 * diff2 + v2 * diff + v3);
		}

		protected IDataObject CubicPoint3D(IDataObject a, IDataObject b, IDataObject c, IDataObject d, double diff) {
			Point3D x0 = (Point3D)a;
			Point3D x1 = (Point3D)b;
			Point3D x2 = (Point3D)c;
			Point3D x3 = (Point3D)d;

			double diff2 = diff * diff;
			Point3D v0 = x3 - x2 - x0 + x1;
			Point3D v1 = x0 - x1 - v0;
			Point3D v2 = x2 - x0;
			Point3D v3 = x1;

			return (Point3D)(v0 * diff * diff2 + v1 * diff2 + v2 * diff + v3);
		}

		protected IDataObject CubicVector2D(IDataObject a, IDataObject b, IDataObject c, IDataObject d, double diff) {
			Point2D x0 = (Point2D)(Vector2D)a;
			Point2D x1 = (Point2D)(Vector2D)b;
			Point2D x2 = (Point2D)(Vector2D)c;
			Point2D x3 = (Point2D)(Vector2D)d;

			double diff2 = diff * diff;
			Point2D v0 = x3 - x2 - x0 + x1;
			Point2D v1 = x0 - x1 - v0;
			Point2D v2 = x2 - x0;
			Point2D v3 = x1;

			return (Vector2D)(v0 * diff * diff2 + v1 * diff2 + v2 * diff + v3);
		}

		protected IDataObject CubicVector2DTo3D(IDataObject a, IDataObject b, IDataObject c, IDataObject d, double diff) {
			if (a is Vector3D) {
				return CubicVector3D(a, b, c, d, diff);
			}
			else {
				Vector2D v = (Vector2D)CubicVector2D(a, b, c, d, diff);
				return new Vector3D(v.X, 0, v.Y);
			}
		}

		protected IDataObject CubicVector3D(IDataObject a, IDataObject b, IDataObject c, IDataObject d, double diff) {
			Point3D x0 = (Point3D)(Vector3D)a;
			Point3D x1 = (Point3D)(Vector3D)b;
			Point3D x2 = (Point3D)(Vector3D)c;
			Point3D x3 = (Point3D)(Vector3D)d;

			double diff2 = diff * diff;
			Point3D v0 = x3 - x2 - x0 + x1;
			Point3D v1 = x0 - x1 - v0;
			Point3D v2 = x2 - x0;
			Point3D v3 = x1;

			return (Vector3D)(v0 * diff * diff2 + v1 * diff2 + v2 * diff + v3);
		}

		#endregion
	}
}
