﻿using System;
using System.Collections.Generic;
using System.Text;
using Zcu.Mve.Visualization;
using Zcu.Mve.Numerics;
using Zcu.Mve.Core;

namespace Zcu.Mve.Visualization.Processing
{
	/// <summary>
	/// 2D interpolator of values in grid
	/// </summary>
	class Interpolator2D : Interpolator{
		UniformDataArrayND values;
		Point3D min;
		Point3D max;
		Point2D cellSize;
		int[] dims;
		Type type;
		bool force3DOutput;
		EInterpolation interpolation;

		Lerp lerpFunction;
		Cubic cubicFunction;
		InterpolateAt interpolateFunction;

		/// <summary>
		/// Creates new interpolator in 2D
		/// </summary>
		/// <param name="values">values to interpolate (stored in regular grid)</param>
		/// <param name="min">min position (bounds)</param>
		/// <param name="max">max position (bounds)</param>
		/// <param name="cellSize">cell size</param>
		/// <param name="force3DOutput">if Vector2D should be interpreted as Vector3D</param>
		/// <param name="interpolation">interpolation type</param>
		public Interpolator2D(UniformDataArrayND values, Point3D min, Point3D max, Point2D cellSize, bool force3DOutput, EInterpolation interpolation) {
			this.values = values;
			this.min = min;
			this.max = max;
			this.cellSize = cellSize;
			this.force3DOutput = force3DOutput;
			this.interpolation = interpolation;
			type = values.ElemType;
			dims = values.Lengths;

			if (interpolation == EInterpolation.Linear) {
				interpolateFunction = InterpolateLinearAt;
			}
			else {
				interpolateFunction = InterpolateCubicAt;
			}

			if (type == typeof(Scalar)) {
				lerpFunction = LerpScalar;
				cubicFunction = CubicScalar;
			}
			else if (type == typeof(Integer)) {
				lerpFunction = LerpInteger;
				cubicFunction = CubicInteger;
			}
			else if (type == typeof(Point2D)) {
				lerpFunction = LerpPoint2D;
				cubicFunction = CubicPoint2D;
			}
			else if (type == typeof(Point3D)) {
				lerpFunction = LerpPoint3D;
				cubicFunction = CubicPoint3D;
			}
			else if (type == typeof(Vector2D)) {
				if (force3DOutput) {
					cubicFunction = CubicVector2DTo3D;
					lerpFunction = LerpVector2DTo3D;
				}
				else {
					cubicFunction = CubicVector2D;
					lerpFunction = LerpVector2D;
				}
			}
			else if (type == typeof(Vector3D)) {
				lerpFunction = LerpVector3D;
				cubicFunction = CubicVector3D;
			}
			else {
				throw new MveException(this, "Unknown data type for interpolation (" + type + ").");
			}
		}

		/// <summary>
		/// Computes interpolated values in points in paramter
		/// </summary>
		/// <param name="points">points</param>
		/// <returns>array of interpolated values</returns>
		public override UniformDataArray Compute(UniformDataArray points) {
			int count = points.Count;
			UniformDataArray interpolated = new UniformDataArray(type, count);

			for (int i = 0; i < count; i++) {
				interpolated[i] = InterpolateAtPoint((Point3D)points[i]);
			}

			return interpolated;
		}

		/// <summary>
		/// Computes cell to which point belongs.
		/// </summary>
		/// <param name="point">point in grid</param>
		/// <param name="cell">cell to which point belongs</param>
		/// <param name="ndp">normalized distances from cell origin</param>
		void GetCell(Point3D point, ref int[] cell, ref Point2D ndp) {
			point = point - min;
			cell[0] = (int)Math.Floor(point.X / cellSize.X);
			cell[1] = (int)Math.Floor(point.Z / cellSize.Y);

			ndp.X = (point.X - cell[0] * cellSize.X) / cellSize.X;
			ndp.Y = (point.Z - cell[1] * cellSize.Y) / cellSize.Y;

			if (cell[0] == dims[0] - 1) {
				cell[0] = dims[0] - 2;
				ndp.X = 1;
			}
			if (cell[1] == dims[1] - 1) {
				cell[1] = dims[1] - 2;
				ndp.Y = 1;
			}
		}

		/// <summary>
		/// Indicates if point is out of bounds
		/// </summary>
		/// <param name="point">point</param>
		/// <returns>true if point is out</returns>
		public override bool IsOut(Point3D point) {
			if (point.X < min.X || point.Y < min.Y || point.Z < min.Z ||
				point.X > max.X || point.Y > max.Y || point.Z > max.Z) {
				return true;
			}
			else {
				return false;
			}
		}

		#region linear and cubic interpolations

		IDataObject InterpolateLinearAt(Point3D point) {
			point.X = Math.Min(point.X, max.X);
			point.X = Math.Max(point.X, min.X);

			point.Z = Math.Min(point.Z, max.Z);
			point.Z = Math.Max(point.Z, min.Z);

			int[] cell = new int[2];
			Point2D ndp = new Point2D();
			GetCell(point, ref cell, ref ndp);

			// bilinear interpolation of vector field at points position
			IDataObject down1 = lerpFunction(values[cell[0], cell[1]], values[cell[0] + 1, cell[1]], ndp.X);
			IDataObject up1 = lerpFunction(values[cell[0], cell[1] + 1], values[cell[0] + 1, cell[1] + 1], ndp.X);
			return lerpFunction(down1, up1, ndp.Y);
		}

		IDataObject InterpolateCubicAt(Point3D point) {
			point.X = Math.Min(point.X, max.X);
			point.X = Math.Max(point.X, min.X);

			point.Z = Math.Min(point.Z, max.Z);
			point.Z = Math.Max(point.Z, min.Z);

			int[] cell = new int[2];
			Point2D ndp = new Point2D();
			GetCell(point, ref cell, ref ndp);

			int[] cellPlus2 = new int[2];
			cellPlus2[0] = Math.Min(cell[0] + 2, dims[0] - 1);
			cellPlus2[1] = Math.Min(cell[1] + 2, dims[1] - 1);

			int[] cellMinus1 = new int[2];
			cellMinus1[0] = Math.Max(cell[0] - 1, 0);
			cellMinus1[1] = Math.Max(cell[1] - 1, 0);

			IDataObject a, b, c, d;

			// bicubic interpolation of vector field at points position
			a = cubicFunction(
				values[cellMinus1[0], cellMinus1[1]],
				values[cell[0], cellMinus1[1]],
				values[cell[0] + 1, cellMinus1[1]],
				values[cellPlus2[0], cellMinus1[1]], ndp.X);
			b = cubicFunction(
				values[cellMinus1[0], cell[1]],
				values[cell[0], cell[1]],
				values[cell[0] + 1, cell[1]],
				values[cellPlus2[0], cell[1]], ndp.X);
			c = cubicFunction(
				values[cellMinus1[0], cell[1]+1],
				values[cell[0], cell[1] + 1],
				values[cell[0] + 1, cell[1] + 1],
				values[cellPlus2[0], cell[1] + 1], ndp.X);
			d = cubicFunction(
				values[cellMinus1[0], cellPlus2[1]+1],
				values[cell[0], cellPlus2[1] + 1],
				values[cell[0] + 1, cellPlus2[1] + 1],
				values[cellPlus2[0], cellPlus2[1] + 1], ndp.X);
			return cubicFunction(a, b, c, d, ndp.Y);
		}

		#endregion

		/// <summary>
		/// Interpolates value at a point
		/// </summary>
		/// <param name="point">point</param>
		/// <returns>interpolated value</returns>
		public override IDataObject InterpolateAtPoint(Point3D point) {
			return interpolateFunction(point);
			
		}
	}
}
