﻿using System;
using System.Collections.Generic;
using System.Text;
using Zcu.Mve.Core;
using Zcu.Mve.Visualization;
using Zcu.Mve.Numerics;
using System.Diagnostics;

namespace Zcu.Mve.Visualization.Sources
{
	/// <summary>
	/// Module for generation of 3D regular grid with vector field
	/// </summary>
	[ModuleInfo("Martin Dohnal", "Generates 3D regular grid with random vectors in every point.", CategoryType.Source, CategoryData.Volume)]
	class RandomVectorGrid : Module {
		RegGridND vectorGrid;
		public enum RandomGridType { Random, Well, Circles };
		public enum VectorType { Vector3D, Vector2D };

		public RandomVectorGrid() {
			AddOutPort("vectorGrid", typeof(RegGridND));
		}

		#region attributes

		int sizeX = 20;
		int sizeY = 20;
		int sizeZ = 20;

		double lenghtX = 40;
		double lenghtY = 40;
		double lenghtZ = 40;

		string vectorFieldName = "vectors";

		VectorType vectorType = VectorType.Vector3D;

		bool randomizeSize = true;

		RandomGridType type = RandomGridType.Well;
	
		public bool RandomizeSize {
			get { return randomizeSize; }
			set { randomizeSize = value; }
		}

		public RandomGridType Type {
			get { return type; }
			set { type = value;}
		}

		public VectorType TypeOfVectors {
			get { return vectorType; }
			set { vectorType = value; }
		}

		public string VectorFieldName {
			get { return vectorFieldName; }
			set { vectorFieldName = value; }
		}

		public double LenghtX {
			get { return lenghtX; }
			set { lenghtX = value; }
		}

		public double LenghtY {
			get { return lenghtY; }
			set { lenghtY = value;  }
		}

		public double LenghtZ {
			get { return lenghtZ; }
			set { lenghtZ = value;}
		}

		public int SizeX {
			get { return sizeX; }
			set { sizeX = value; }
		}

		public int SizeY {
			get { return sizeY; }
			set { sizeY = value;}
		}

		public int SizeZ {
			get { return sizeZ; }
			set { sizeZ = value; }
		}

		#endregion

		void GenerateGrid() {
			vectorGrid = new RegGridND(new double[] { 0, 0, 0 }, new double[] { lenghtX, lenghtY, lenghtZ },
				new int[] { sizeX, sizeY, sizeZ });
			UniformDataArrayND vectors;
			Random random = new Random();

			if (vectorType == VectorType.Vector3D) {
				vectors = new UniformDataArrayND(typeof(Vector3D), new int[] { sizeX + 1, sizeY + 1, sizeZ + 1 });

				switch (type) {
					case RandomGridType.Random: {
							for (int x = 0; x < sizeX + 1; x++) {
								for (int y = 0; y < sizeY + 1; y++) {
									for (int z = 0; z < sizeZ + 1; z++) {
										vectors[x, y, z] = new Vector3D(
											random.NextDouble() * 2 - 1,
											random.NextDouble() * 2 - 1,
											random.NextDouble() * 2 - 1);
									}
								}
							}
						}
						break;
					case RandomGridType.Well: {
							Point3D point;
							Point3D vect;
							for (int x = 0; x < sizeX + 1; x++) {
								for (int y = 0; y < sizeY + 1; y++) {
									for (int z = 0; z < sizeZ + 1; z++) {
										point = (Point3D)vectorGrid.GetPoint(x, y, z);
										vect = new Point3D(point.Z - lenghtZ / 2,
											-lenghtY / 10, -point.X + lenghtX / 2);
										double len = vect.Abs();
										if (len == 0) {
											len = 1;
										}
										if (randomizeSize) {
											vectors[x, y, z] = (Vector3D)(vect / len * random.NextDouble());
										}
										else {
											vectors[x, y, z] = (Vector3D)(vect / len);
										}
									}
								}
							}
						}
						break;
					case RandomGridType.Circles: {
							Point3D point;
							Point3D vect;
							for (int x = 0; x < sizeX + 1; x++) {
								for (int y = 0; y < sizeY + 1; y++) {
									for (int z = 0; z < sizeZ + 1; z++) {
										point = (Point3D)vectorGrid.GetPoint(x, y, z);
										vect = new Point3D(point.Z - lenghtZ / 2,
											0, -point.X + lenghtX / 2);
										double len = vect.Abs();
										if (len == 0) {
											len = 1;
										}
										if (randomizeSize) {
											vectors[x, y, z] = (Vector3D)(vect / len * random.NextDouble());
										}
										else {
											vectors[x, y, z] = (Vector3D)(vect / len);
										}
									}
								}
							}
						}
						break;
				}
			}
			else {
				vectors = new UniformDataArrayND(typeof(Vector2D), new int[] { sizeX + 1, sizeY + 1, sizeZ + 1 });

				switch (type) {
					case RandomGridType.Random: {
							for (int x = 0; x < sizeX + 1; x++) {
								for (int y = 0; y < sizeY + 1; y++) {
									for (int z = 0; z < sizeZ + 1; z++) {
										vectors[x, y, z] = new Vector2D(
											random.NextDouble() * 2 - 1,
											random.NextDouble() * 2 - 1);
									}
								}
							}
						}
						break;
					case RandomGridType.Circles:
					case RandomGridType.Well: {
							Point3D point;
							Point3D vect;
							for (int x = 0; x < sizeX + 1; x++) {
								for (int y = 0; y < sizeY + 1; y++) {
									for (int z = 0; z < sizeZ + 1; z++) {
										point = (Point3D)vectorGrid.GetPoint(x, y, z);
										vect = new Point2D(point.Z - lenghtZ / 2, -point.X + lenghtX / 2);
										double len = vect.Abs();
										if (len == 0) {
											len = 1;
										}
										if (randomizeSize) {
											vectors[x, y, z] = (Vector2D)(vect / len * random.NextDouble());
										}
										else {
											vectors[x, y, z] = (Vector2D)(vect / len);
										}
									}
								}
							}
						}
						break;
				}
			}
			vectorGrid.AddPointAttrs(vectorFieldName, vectors);
		}

		public override void Execute() {
			if (vectorGrid == null || !IsAllUserSettingsSame()) {
				GenerateGrid();
				SetOutput("vectorGrid", vectorGrid);
				IsAllUserSettingsSame();
			}
			else {
				SetOutput("vectorGrid", vectorGrid, true);
			}
		}
	}
}
