using System;
using Zcu.Mve.Core;
using System.Globalization;

namespace Zcu.Mve.Visualization
{
    /// <summary>
    /// General tranformation of a 3D object.
    /// </summary>
    public class Transform3D : Visualization.AbstractTransform
    {
        /// <summary>
        /// double matrix 4 x 4 representing 3D-transformation.
        /// </summary>
        private double[,] matrix = new double[4, 4];

        /// <summary>
        /// The toleration for equality operator.
        /// </summary>
        public static double Epsilon = Globals.EpsilonDoubleImplicit;

        /// <summary>
        /// Creates a Transform3D object and initialize it (set identity).
        /// </summary>
        public Transform3D()
        {
            SetIdentity();

            UpdateModificationTime();
        }

        /// <summary>
        /// Set identity transformation.
        /// </summary>
        public void SetIdentity()
        {
            for (int i = 0; i < 4; i++)
            {
                for (int j = 0; j < 4; j++)
                {
                    matrix[i, j] = 0.0f;
                }
            }
            for (int i = 0; i < 4; i++)
            {
                matrix[i, i] = 1.0f;
            }

            UpdateModificationTime();
        }

        /// <summary>
        /// Add scaling to current transformation.
        /// TODO: Optimize it!
        /// </summary>
        /// <param name="dx">Scaling in x-axis. 1 means no scale.</param>
        /// <param name="dy">Scaling in y-axis. 1 means no scale.</param>
        /// <param name="dz">Scaling in z-axis. 1 means no scale.</param>
        public void AddScale(double dx, double dy, double dz)
        {
            Transform3D t = new Transform3D();
            t.SetScale(dx, dy, dz);
            this.matrix = (this * t).matrix;

            UpdateModificationTime();
        }

        /// <summary>
        /// Set curent matrix to scaling matrix of given values.
        /// </summary>
        /// <param name="dx">Scaling in x-axis. 1 means no scale.</param>
        /// <param name="dy">Scaling in y-axis. 1 means no scale.</param>
        /// <param name="dz">Scaling in z-axis. 1 means no scale.</param>
        public void SetScale(double dx, double dy, double dz)
        {
            SetIdentity();
            matrix[0, 0] = dx;
            matrix[1, 1] = dy;
            matrix[2, 2] = dz;

            UpdateModificationTime();
        }

        /// <summary>
        /// Noncumulative rotation about axis X measured in radians.
        /// </summary>
        /// <param name="angle">Rotation angle masured in radians.</param>
        public void SetRotaionX(double angle)
        {
            SetIdentity();
            matrix[1, 1] = matrix[2, 2] = (double)Math.Cos(angle);
            matrix[2, 1] = (double)Math.Sin(angle);
            matrix[1, 2] = -matrix[2, 1];

            UpdateModificationTime();
        }

        /// <summary>
        /// Cumulative rotation about axis X.
        /// </summary>
        /// <param name="angle">Rotation angle masured in radians.</param>
        public void AddRotationX(double angle)
        {
            Transform3D rot = new Transform3D();
            rot.SetRotaionX(angle);
            this.matrix = (this * rot).matrix;

            UpdateModificationTime();
        }

        /// <summary>
        /// Noncumulative rotation about axis Y measured in radians.
        /// </summary>
        /// <param name="angle">Rotation angle masured in radians.</param>
        public void SetRotaionY(double angle)
        {
            SetIdentity();
            matrix[0, 0] = matrix[2, 2] = (double)Math.Cos(angle);
            matrix[0, 2] = (double)Math.Sin(angle);
            matrix[2, 0] = -matrix[0, 2];

            UpdateModificationTime();
        }

        /// <summary>
        /// Cumulative rotation about axis Y.
        /// </summary>
        /// <param name="angle">Rotation angle masured in radians.</param>
        public void AddRotationY(double angle)
        {
            Transform3D rot = new Transform3D();
            rot.SetRotaionY(angle);
            this.matrix = (this * rot).matrix;

            UpdateModificationTime();
        }

        /// <summary>
        /// Noncumulative rotation about axis Z measured in radians.
        /// </summary>
        /// <param name="angle">Rotation angle masured in radians.</param>
        public void SetRotaionZ(double angle)
        {
            SetIdentity();
            matrix[0, 0] = matrix[1, 1] = (double)Math.Cos(angle);
            matrix[1, 0] = (double)Math.Sin(angle);
            matrix[0, 1] = -matrix[1, 0];

            UpdateModificationTime();
        }

        /// <summary>
        /// Cumulative rotation about axis Z.
        /// </summary>
        /// <param name="angle">Rotation angle masured in radians.</param>
        public void AddRotationZ(double angle)
        {
            Transform3D rot = new Transform3D();
            rot.SetRotaionZ(angle);
            this.matrix = (this * rot).matrix;

            UpdateModificationTime();
        }

        /// <summary>
        /// Noncumlative translation.
        /// </summary>
        /// <param name="x">Translation in x-axis.</param>
        /// <param name="y">Translation in y-axis.</param>
        /// <param name="z">Translation in z-axis.</param>
        public void SetTranslation(double x, double y, double z)
        {
            SetIdentity();
            matrix[0, 3] = x;
            matrix[1, 3] = y;
            matrix[2, 3] = z;

            UpdateModificationTime();
        }

        /// <summary>
        /// Cumulative translation
        /// </summary>
        /// <param name="x">Translation in x-axis.</param>
        /// <param name="y">Translation in y-axis.</param>
        /// <param name="z">Translation in z-axis.</param>
        public void AddTranslation(double x, double y, double z)
        {
            Transform3D t = new Transform3D();
            t.SetTranslation(x, y, z);
            this.matrix = (this * t).matrix;

            UpdateModificationTime();
        }

        /// <summary>
        /// Creates a general rotation matrix. The rotation axis is specified by vertices passed as <paramref name="axisA"/> and <paramref name="axisB"/>, the rotation angle is specified by <paramref name="angleRadians"/>.
        /// </summary>
        /// <param name="axisA">First point specifying the rotation axis</param>
        /// <param name="axisB">Second point specifying the rotation axis</param>
        /// <param name="angleRadians">Angle of rotation in radians.</param>
        /// <returns>Resulting general rotation matrix</returns>
        public static Transform3D gerealRotation(Point3D axisA, Point3D axisB, double angleRadians)
        {
            Transform3D translateBack = new Transform3D();
            translateBack.SetTranslation(-axisA.X, -axisA.Y, -axisA.Z);
            double projectedX = axisB.X - axisA.X;
            double projectedY = axisB.Y - axisA.Y;
            double projectedZ = axisB.Z - axisA.Z;

            double lengthXZ = Math.Sqrt(projectedX * projectedX + projectedZ * projectedZ);
            double arg = projectedX / lengthXZ;
            if (arg > 1)
                arg = 1;
            if (arg < -1)
                arg = -1;
            double alpha = Math.Acos(arg);
            if (projectedZ < 0)
                alpha = Math.PI * 2 - alpha;

            if (lengthXZ == 0)
                alpha = 0;

            Transform3D rotate1 = new Transform3D();
            rotate1.SetRotaionY(alpha);
            Point3D projectedRotated = rotate1.Transform(new Point3D(projectedX, projectedY, projectedZ));

            double lengthXY = Math.Sqrt(projectedRotated.X * projectedRotated.X + projectedRotated.Y * projectedRotated.Y);
            arg = projectedRotated.X / lengthXY;
            if (arg > 1)
                arg = 1;
            if (arg < -1)
                arg = -1;
            double beta = Math.Acos(arg);
            if (projectedY < 0)
                beta = Math.PI * 2 - beta;

            if (lengthXY == 0)
                beta = 0;



            Transform3D rotate2 = new Transform3D();
            rotate2.SetRotaionZ(-beta);

            Transform3D mainRotate = new Transform3D();
            mainRotate.SetRotaionX(angleRadians);

            Transform3D rotate2inv = new Transform3D();
            rotate2inv.SetRotaionZ(beta);

            Transform3D rotate1inv = new Transform3D();
            rotate1inv.SetRotaionY(-alpha);

            Transform3D translate = new Transform3D();
            translate.SetTranslation(axisA.X, axisA.Y, axisA.Z);

            Transform3D result = translate * rotate1inv * rotate2inv * mainRotate * rotate2 * rotate1 * translateBack;
            //Transform3D result = translateBack * rotate1 * rotate2 * mainRotate * rotate2inv * rotate1inv * translate;
            return (result);
        }

        /// <summary>
        /// Transform the point.
        /// </summary>
        /// <param name="op">Point to be transform. This parameter won't be changed.</param>
        /// <returns>New instance of transformed point.</returns>
        public Point3D Transform(Point3D op)
        {
            Point3D tp = new Point3D();

            tp.X = matrix[0, 0] * op.X + matrix[0, 1] * op.Y + matrix[0, 2] * op.Z + matrix[0, 3];
            tp.Y = matrix[1, 0] * op.X + matrix[1, 1] * op.Y + matrix[1, 2] * op.Z + matrix[1, 3];
            tp.Z = matrix[2, 0] * op.X + matrix[2, 1] * op.Y + matrix[2, 2] * op.Z + matrix[2, 3];

            // homogeni slozka ignorovana -- never more, JK 12.3.2006

            double w = matrix[3, 0] * op.X + matrix[3, 1] * op.Y + matrix[3, 2] * op.Z + matrix[3, 3];

            tp.X /= w;
            tp.Y /= w;
            tp.Z /= w;

            return tp;
        }

        /// <summary>
        /// Composition of transformation. Matrix multiplication.
        /// </summary>
        /// <param name="t1"></param>
        /// <param name="t2"></param>
        /// <returns></returns>
        public static Transform3D operator *(Transform3D t1, Transform3D t2)
        {
            Transform3D t = new Transform3D();
            double sum = 0.0f;

            for (int i = 0; i < 4; i++)
            {
                for (int j = 0; j < 4; j++)
                {
                    sum = 0.0f;
                    for (int k = 0; k < 4; k++)
                    {
                        sum += t1[i, k] * t2[k, j];
                    }
                    t[i, j] = sum;
                }
            }
            return t;
        }

        /// <summary>
        /// Full access to particular matrix elements.
        /// </summary>
        public double this[int i, int j]
        {
            set
            {
                matrix[i, j] = value;

                UpdateModificationTime();
            }
            get
            {
                return matrix[i, j];
            }
        }

        /// <summary>
        /// Check the consistency of current data structure.
        /// </summary>
        /// <returns>True if structure is consistent.</returns>
        public override bool CheckConsistence()
        {
            // TODO: mozna nejaky determinant-test a nasledny warning
            return true;
        }

        /// <summary>
        /// Deep copy of data object and its subelements.
        /// </summary>
        /// <returns>Same data, different memory space.</returns>
        public override Zcu.Mve.Core.IDataObject DeepCopy()
        {
            Transform3D t = new Transform3D();

            for (int i = 0; i < 4; i++)
            {
                for (int j = 0; j < 4; j++)
                {
                    t.matrix[i, j] = this.matrix[i, j];
                }
            }

            return t;
        }

        /// <summary>
        /// Read data from Xml file to current structure.
        /// </summary>
        /// <param name="xmlTextReader">Stream to be read.</param>
        public override void ReadData(System.Xml.XmlTextReader xmlTextReader)
        {
            string[] lines = xmlTextReader.ReadString().Trim().Split(';');

            if (lines.Length != 4)
                throw new MveException(this, "Incorrect number of lines in Transform3D.");

            for (int i = 0; i < 4; i++)
            {
                string[] elems = lines[i].Trim().Split(' ');
                if (elems.Length != 4)
                    throw new MveException(this, "Incorrect number of elements in line " + i + " in Transform3D.");

                for (int j = 0; j < 4; j++)
                {
                    //this.matrix[i, j] = Int32.Parse(elems[j].Trim(), Globals.nfi);
                    this.matrix[i, j] = double.Parse(elems[j].Trim(), Globals.nfi);
                }
            }

            UpdateModificationTime();
        }

        /// <summary>
        /// Write current data to Xml file.
        /// </summary>
        /// <param name="xmlTextWriter">Stream to be write.</param>
        public override void WriteData(System.Xml.XmlTextWriter xmlTextWriter)
        {
            string str = "";
            for (int i = 0; i < 4; i++)
            {
                for (int j = 0; j < 4; j++)
                {
                    str += this.matrix[i, j].ToString(Globals.nfi);
                    if (j < 3)
                        str += " ";
                }
                if (i < 3)
                    str += "; ";
            }
            xmlTextWriter.WriteString(str);
        }

        // ===== Methods of the class Object ==============================================
        #region Overrided methods of the class Object

        /// <summary>
        /// Returns a String that represents the current Object.
        /// </summary>
        /// <returns>String that represents the current Object.</returns>
        public override string ToString()
        {
            string str = "Transform3D:\n";
            for (int i = 0; i < 4; i++)
                str += "   [" + matrix[i, 0] + ", " + matrix[i, 1] + ", " + matrix[i, 2] + ", " + matrix[i, 3] + "]\r\n";
            return str.Substring(0, str.Length - 2);
        }

        /// <summary>
        /// Determines whether the specified Objects are equal.
        /// </summary>
        /// <param name="a">First Object.</param>
        /// <param name="b">Second Object.</param>
        /// <returns>True if objects are equal. Otherwise, false.</returns>
        public static new bool Equals(object a, object b)
        {
            if (a.GetType() != typeof(Transform3D) || b.GetType() != typeof(Transform3D))
                return false;

            Transform3D x = (Transform3D)a;
            Transform3D y = (Transform3D)b;

            for (int i = 0; i < 4; i++)
                for (int j = 0; j < 4; j++)
                    if ((double)Math.Abs(x[i, j] - y[i, j]) > Transform3D.Epsilon)
                        return false;

            return true;
        }

        /// <summary>
        /// Determines whether the specified Object is equal to the current Object.
        /// </summary>
        /// <param name="obj">Specified Object to compare with the current Object.</param>
        /// <returns>True if the specified Object is equal to the current Object. Otherwise, false.</returns>
        public override bool Equals(object obj)
        {
            return Transform3D.Equals(this, obj);
        }

        /// <summary>
        /// Returns a hash code for the current Object.
        /// </summary>
        /// <returns>A hash code for the current Object.</returns>
        public override int GetHashCode()
        {
            int hashCode = this[0, 0].GetHashCode()
              ^ this[0, 1].GetHashCode()
              ^ this[0, 2].GetHashCode()
              ^ this[0, 3].GetHashCode();

            for (int i = 1; i < 3; i++)
                for (int j = 0; j < 3; j++)
                    hashCode = hashCode ^ this[i, j].GetHashCode();

            return hashCode;
        }

        #endregion

        // ===== Operators ================================================================
        #region Operators

        /// <summary>
        /// The equality operator (==) returns true if the values of its operands are equal, false otherwise.
        /// </summary>
        /// <param name="a">First operand.</param>
        /// <param name="b">Second operand.</param>
        /// <returns>True if the values of operands are equal, false otherwise.</returns>
        public static bool operator ==(Transform3D a, Transform3D b)
        {
            return Transform3D.Equals(a, b);
        }

        /// <summary>
        /// The inequality operator (!=) returns false if its operands are equal, true otherwise.
        /// </summary>
        /// <param name="a">First operand.</param>
        /// <param name="b">Second operand.</param>
        /// <returns>False if the values of operands are equal, true otherwise.</returns>
        public static bool operator !=(Transform3D a, Transform3D b)
        {
            return !(a == b);
        }

        #endregion

    } // class Transform3D
} // namespace