using System;
using System.Collections.Generic;
using System.Text;

using Zcu.Mve.Core;
using Zcu.Mve.Visualization;
using Zcu.Mve.Numerics;

using System.ComponentModel;
using System.Globalization;
using System.Xml;
using System.Diagnostics;

namespace Zcu.Mve.Visualization.Sources
{
  /// <summary>
  /// Source of Transform3D matrix. Values from input ports are preffered. After input port disconnecting, last values remains.
  /// </summary>
  [ModuleInfo("Jan Kaiser", "Source of Transform3D matrix. Values from input ports are preffered. "
            + "After input port disconnecting, last values remains.", CategoryType.Source, CategoryData.Number, IconName = "Transform3DSource.ico")]
  
  [PortInfo("outTransform3D", "Resulting transformation matrix.")]

  [PortInfo("inM00", "M[0,0] - transform matrix element.")]
  [PortInfo("inM01", "M[0,1] - transform matrix element.")]
  [PortInfo("inM02", "M[0,2] - transform matrix element.")]
  [PortInfo("inM03", "M[0,3] - transform matrix element.")]

  [PortInfo("inM10", "M[1,0] - transform matrix element.")]
  [PortInfo("inM11", "M[1,1] - transform matrix element.")]
  [PortInfo("inM12", "M[1,2] - transform matrix element.")]
  [PortInfo("inM13", "M[1,3] - transform matrix element.")]

  [PortInfo("inM20", "M[2,0] - transform matrix element.")]
  [PortInfo("inM21", "M[2,1] - transform matrix element.")]
  [PortInfo("inM22", "M[2,2] - transform matrix element.")]
  [PortInfo("inM21", "M[2,3] - transform matrix element.")]

  [PortInfo("inM30", "M[3,0] - transform matrix element.")]
  [PortInfo("inM31", "M[3,1] - transform matrix element.")]
  [PortInfo("inM32", "M[3,2] - transform matrix element.")]
  [PortInfo("inM33", "M[3,3] - transform matrix element.")]
  public class Transform3DSource : Module
  {
    private Transform3D transform3D;
    private bool[,] showInPort = new bool[4,4];

    #region Properties
    /// <summary>
    /// M00 - transform matrix element.
    /// </summary>
    [Category("1st row"), Description("M[id,j] - transform matrix element."), DefaultValue(1f)]
    public double M00
    {
      get
      {
        return transform3D[0, 0];
      }
      set
      {
        transform3D[0, 0] = value;
        //dataChanged = true;
      }
    }

    /// <summary>
    /// M01 - transform matrix element.
    /// </summary>
    [Category("1st row"), Description("M[id,j] - transform matrix element."), DefaultValue(0f)]
    public double M01
    {
      get
      {
        return transform3D[0, 1];
      }
      set
      {
        transform3D[0, 1] = value;
        //dataChanged = true;
      }
    }

    /// <summary>
    /// M02 - transform matrix element.
    /// </summary>
    [Category("1st row"), Description("M[id,j] - transform matrix element."), DefaultValue(0f)]
    public double M02
    {
      get
      {
        return transform3D[0, 2];
      }
      set
      {
        transform3D[0, 2] = value;
        //dataChanged = true;
      }
    }

    /// <summary>
    /// M03 - transform matrix element.
    /// </summary>
    [Category("1st row"), Description("M[id,j] - transform matrix element."), DefaultValue(0f)]
    public double M03
    {
      get
      {
        return transform3D[0, 3];
      }
      set
      {
        transform3D[0, 3] = value;
        //dataChanged = true;
      }
    }

    /// <summary>
    /// M10 - transform matrix element.
    /// </summary>
    [Category("2nd row"), Description("M[id,j] - transform matrix element."), DefaultValue(0f)]
    public double M10
    {
      get
      {
        return transform3D[1, 0];
      }
      set
      {
        transform3D[1, 0] = value;
        //dataChanged = true;
      }
    }

    /// <summary>
    /// M11 - transform matrix element.
    /// </summary>
    [Category("2nd row"), Description("M[id,j] - transform matrix element."), DefaultValue(1f)]
    public double M11
    {
      get
      {
        return transform3D[1, 1];
      }
      set
      {
        transform3D[1, 1] = value;
        //dataChanged = true;
      }
    }

    /// <summary>
    /// M12 - transform matrix element.
    /// </summary>
    [Category("2nd row"), Description("M[id,j] - transform matrix element."), DefaultValue(0f)]
    public double M12
    {
      get
      {
        return transform3D[1, 2];
      }
      set
      {
        transform3D[1, 2] = value;
        //dataChanged = true;
      }
    }

    /// <summary>
    /// M13 - transform matrix element.
    /// </summary>
    [Category("2nd row"), Description("M[id,j] - transform matrix element."), DefaultValue(0f)]
    public double M13
    {
      get
      {
        return transform3D[1, 3];
      }
      set
      {
        transform3D[1, 3] = value;
        //dataChanged = true;
      }
    }

    /// <summary>
    /// M20 - transform matrix element.
    /// </summary>
    [Category("3rd row"), Description("M[id,j] - transform matrix element."), DefaultValue(0f)]
    public double M20
    {
      get
      {
        return transform3D[2, 0];
      }
      set
      {
        transform3D[2, 0] = value;
        //dataChanged = true;
      }
    }

    /// <summary>
    /// M21 - transform matrix element.
    /// </summary>
    [Category("3rd row"), Description("M[id,j] - transform matrix element."), DefaultValue(0f)]
    public double M21
    {
      get
      {
        return transform3D[2, 1];
      }
      set
      {
        transform3D[2, 1] = value;
        //dataChanged = true;
      }
    }

    /// <summary>
    /// M22 - transform matrix element.
    /// </summary>
    [Category("3rd row"), Description("M[id,j] - transform matrix element."), DefaultValue(1f)]
    public double M22
    {
      get
      {
        return transform3D[2, 2];
      }
      set
      {
        transform3D[2, 2] = value;
        //dataChanged = true;
      }
    }

    /// <summary>
    /// M23 - transform matrix element.
    /// </summary>
    [Category("3rd row"), Description("M[id,j] - transform matrix element."), DefaultValue(0f)]
    public double M23
    {
      get
      {
        return transform3D[2, 3];
      }
      set
      {
        transform3D[2, 3] = value;
        //dataChanged = true;
      }
    }

    /// <summary>
    /// M30 - transform matrix element.
    /// </summary>
    [Category("4th row"), Description("M[id,j] - transform matrix element."), DefaultValue(0f)]
    public double M30
    {
      get
      {
        return transform3D[3, 0];
      }
      set
      {
        transform3D[3, 0] = value;
        //dataChanged = true;
      }
    }

    /// <summary>
    /// M31 - transform matrix element.
    /// </summary>
    [Category("4th row"), Description("M[id,j] - transform matrix element."), DefaultValue(0f)]
    public double M31
    {
      get
      {
        return transform3D[3, 1];
      }
      set
      {
        transform3D[3, 1] = value;
        //dataChanged = true;
      }
    }

    /// <summary>
    /// M32 - transform matrix element.
    /// </summary>
    [Category("4th row"), Description("M[id,j] - transform matrix element."), DefaultValue(0f)]
    public double M32
    {
      get
      {
        return transform3D[3, 2];
      }
      set
      {
        transform3D[3, 2] = value;
        //dataChanged = true;
      }
    }

    /// <summary>
    /// M33 - transform matrix element.
    /// </summary>
    [Category("4th row"), Description("M[id,j] - transform matrix element."), DefaultValue(1f)]
    public double M33
    {
      get
      {
        return transform3D[3, 3];
      }
      set
      {
        transform3D[3, 3] = value;
        //dataChanged = true;
      }
    }

    // Show Input Ports.......................................................................................................

    /// <summary>
    /// Shows input port for M00 transform matrix element.
    /// </summary>
    [Category("Show input port, 1st row"), Description("Shows input port for M[id,j] - transform matrix element."), DefaultValue(false)]
    public bool inM00
    {
      get
      {
        return showInPort[0, 0];
      }
      set
      {
        if (value)
        {
          AddInPort("inM00", typeof(Scalar), false, true);
        }
        else
        {
          RemoveInPort("inM00");
        }

        showInPort[0, 0] = value;
      }
    }

    /// <summary>
    /// Shows input port for M01 transform matrix element.
    /// </summary>
    [Category("Show input port, 1st row"), Description("Shows input port for M[id,j] - transform matrix element."), DefaultValue(false)]
    public bool inM01
    {
      get
      {
        return showInPort[0, 1];
      }
      set
      {
        if (value)
        {
          AddInPort("inM01", typeof(Scalar), false, true);
        }
        else
        {
          RemoveInPort("inM01");
        }

        showInPort[0, 1] = value;
      }
    }

    /// <summary>
    /// Shows input port for M02 transform matrix element.
    /// </summary>
    [Category("Show input port, 1st row"), Description("Shows input port for M[id,j] - transform matrix element."), DefaultValue(false)]
    public bool inM02
    {
      get
      {
        return showInPort[0, 2];
      }
      set
      {
        if (value)
        {
          AddInPort("inM02", typeof(Scalar), false, true);
        }
        else
        {
          RemoveInPort("inM02");
        }

        showInPort[0, 2] = value;
      }
    }

    /// <summary>
    /// Shows input port for M03 transform matrix element.
    /// </summary>
    [Category("Show input port, 1st row"), Description("Shows input port for M[id,j] - transform matrix element."), DefaultValue(false)]
    public bool inM03
    {
      get
      {
        return showInPort[0, 3];
      }
      set
      {
        if (value)
        {
          AddInPort("inM03", typeof(Scalar), false, true);
        }
        else
        {
          RemoveInPort("inM03");
        }

        showInPort[0, 3] = value;
      }
    }

    /// <summary>
    /// Shows input port for M10 transform matrix element.
    /// </summary>
    [Category("Show input port, 2nd row"), Description("Shows input port for M[id,j] - transform matrix element."), DefaultValue(false)]
    public bool inM10
    {
      get
      {
        return showInPort[1, 0];
      }
      set
      {
        if (value)
        {
          AddInPort("inM10", typeof(Scalar), false, true);
        }
        else
        {
          RemoveInPort("inM10");
        }

        showInPort[1, 0] = value;
      }
    }

    /// <summary>
    /// Shows input port for M11 transform matrix element.
    /// </summary>
    [Category("Show input port, 2nd row"), Description("Shows input port for M[id,j] - transform matrix element."), DefaultValue(false)]
    public bool inM11
    {
      get
      {
        return showInPort[1, 1];
      }
      set
      {
        if (value)
        {
          AddInPort("inM11", typeof(Scalar), false, true);
        }
        else
        {
          RemoveInPort("inM11");
        }

        showInPort[1, 1] = value;
      }
    }

    /// <summary>
    /// Shows input port for M12 transform matrix element.
    /// </summary>
    [Category("Show input port, 2nd row"), Description("Shows input port for M[id,j] - transform matrix element."), DefaultValue(false)]
    public bool inM12
    {
      get
      {
        return showInPort[1, 2];
      }
      set
      {
        if (value)
        {
          AddInPort("inM12", typeof(Scalar), false, true);
        }
        else
        {
          RemoveInPort("inM12");
        }

        showInPort[1, 2] = value;
      }
    }

    /// <summary>
    /// Shows input port for M13 transform matrix element.
    /// </summary>
    [Category("Show input port, 2nd row"), Description("Shows input port for M[id,j] - transform matrix element."), DefaultValue(false)]
    public bool inM13
    {
      get
      {
        return showInPort[1, 3];
      }
      set
      {
        if (value)
        {
          AddInPort("inM13", typeof(Scalar), false, true);
        }
        else
        {
          RemoveInPort("inM13");
        }

        showInPort[1, 3] = value;
      }
    }

    /// <summary>
    /// Shows input port for M20 transform matrix element.
    /// </summary>
    [Category("Show input port, 3rd row"), Description("Shows input port for M[id,j] - transform matrix element."), DefaultValue(false)]
    public bool inM20
    {
      get
      {
        return showInPort[2, 0];
      }
      set
      {
        if (value)
        {
          AddInPort("inM20", typeof(Scalar), false, true);
        }
        else
        {
          RemoveInPort("inM20");
        }

        showInPort[2, 0] = value;
      }
    }

    /// <summary>
    /// Shows input port for M21 transform matrix element.
    /// </summary>
    [Category("Show input port, 3rd row"), Description("Shows input port for M[id,j] - transform matrix element."), DefaultValue(false)]
    public bool inM21
    {
      get
      {
        return showInPort[2, 1];
      }
      set
      {
        if (value)
        {
          AddInPort("inM21", typeof(Scalar), false, true);
        }
        else
        {
          RemoveInPort("inM21");
        }

        showInPort[2, 1] = value;
      }
    }

    /// <summary>
    /// Shows input port for M22 transform matrix element.
    /// </summary>
    [Category("Show input port, 3rd row"), Description("Shows input port for M[id,j] - transform matrix element."), DefaultValue(false)]
    public bool inM22
    {
      get
      {
        return showInPort[2, 2];
      }
      set
      {
        if (value)
        {
          AddInPort("inM22", typeof(Scalar), false, true);
        }
        else
        {
          RemoveInPort("inM22");
        }

        showInPort[2, 2] = value;
      }
    }

    /// <summary>
    /// Shows input port for M23 transform matrix element.
    /// </summary>
    [Category("Show input port, 3rd row"), Description("Shows input port for M[id,j] - transform matrix element."), DefaultValue(false)]
    public bool inM23
    {
      get
      {
        return showInPort[2, 3];
      }
      set
      {
        if (value)
        {
          AddInPort("inM23", typeof(Scalar), false, true);
        }
        else
        {
          RemoveInPort("inM23");
        }

        showInPort[2, 3] = value;
      }
    }

    /// <summary>
    /// Shows input port for M30 transform matrix element.
    /// </summary>
    [Category("Show input port, 4th row"), Description("Shows input port for M[id,j] - transform matrix element."), DefaultValue(false)]
    public bool inM30
    {
      get
      {
        return showInPort[3, 0];
      }
      set
      {
        if (value)
        {
          AddInPort("inM30", typeof(Scalar), false, true);
        }
        else
        {
          RemoveInPort("inM30");
        }

        showInPort[3, 0] = value;
      }
    }

    /// <summary>
    /// Shows input port for M31 transform matrix element.
    /// </summary>
    [Category("Show input port, 4th row"), Description("Shows input port for M[id,j] - transform matrix element."), DefaultValue(false)]
    public bool inM31
    {
      get
      {
        return showInPort[3, 1];
      }
      set
      {
        if (value)
        {
          AddInPort("inM31", typeof(Scalar), false, true);
        }
        else
        {
          RemoveInPort("inM31");
        }

        showInPort[3, 1] = value;
      }
    }

    /// <summary>
    /// Shows input port for M32 transform matrix element.
    /// </summary>
    [Category("Show input port, 4th row"), Description("Shows input port for M[id,j] - transform matrix element."), DefaultValue(false)]
    public bool inM32
    {
      get
      {
        return showInPort[3, 2];
      }
      set
      {
        if (value)
        {
          AddInPort("inM32", typeof(Scalar), false, true);
        }
        else
        {
          RemoveInPort("inM32");
        }

        showInPort[3, 2] = value;
      }
    }

    /// <summary>
    /// Shows input port for M33 transform matrix element.
    /// </summary>
    [Category("Show input port, 4th row"), Description("Shows input port for M[id,j] - transform matrix element."), DefaultValue(false)]
    public bool inM33
    {
      get
      {
        return showInPort[3, 3];
      }
      set
      {
        if (value)
        {
          AddInPort("inM33", typeof(Scalar), false, true);
        }
        else
        {
          RemoveInPort("inM33");
        }

        showInPort[3, 3] = value;
      }
    }
    #endregion

    /// <summary>
    /// Default constructor. Adds output and input ports.
    /// </summary>
    public Transform3DSource()
    {
      AddOutPort("outTransform3D", typeof(Transform3D));

      for (int i = 0; i < 4; i++)
        for (int j = 0; j < 4; j++)
        {
          if (showInPort[i,j])
            AddInPort("inM" + i + "" + j, typeof(Scalar), false, true);
        }

      transform3D = new Transform3D();
    }
    
    /// <summary>
    /// Generates/updates transform matrix and exposes it on output port.
    /// </summary>
    public override void Execute()
    {
      Object o;

      for (int i = 0; i < 4; i++)
        for (int j = 0; j < 4; j++)
        {
          if (showInPort[i, j])
          {
            o = GetInput("inM" + i + "" + j);
          }
          else
          {
            o = null;
          }

          if (o != null)
          {
            Scalar s = (Scalar) o;

            if (transform3D[i, j] != (double) s.Value)
            {
              transform3D[i, j] = (double) s.Value;
            }
          }
        }

      SetOutput("outTransform3D", transform3D);
    }

    /// <summary>
    /// Writes Xml configuration of current module.
    /// </summary>
    /// <param name="config">XmlElement from map config file which belongs to current module.</param>
    public override void WriteConfig(System.Xml.XmlElement config)
    {
      XmlElement xmlProperties = config.OwnerDocument.CreateElement("Properties");

      XmlElement xmlTransform3D = config.OwnerDocument.CreateElement("Transform3D");
      xmlTransform3D.SetAttribute("type", transform3D.GetType().FullName);
      string str = "";
      for (int i = 0; i < 4; i++)
      {
        for (int j = 0; j < 4; j++)
        {
          str += this.transform3D[i, j].ToString(CultureInfo.InvariantCulture.NumberFormat);
          if (j < 3)
            str += " ";
        }
        if (i < 3)
          str += "; ";
      }
      xmlTransform3D.InnerText = str;
      xmlProperties.AppendChild(xmlTransform3D);

      XmlElement xmlShowInPort = config.OwnerDocument.CreateElement("ShowInPort");
      xmlShowInPort.SetAttribute("type", showInPort.GetType().FullName);
      str = "";
      for (int i = 0; i < 4; i++)
      {
        for (int j = 0; j < 4; j++)
        {
          str += this.showInPort[i, j].ToString(CultureInfo.InvariantCulture.NumberFormat);
          if (j < 3)
            str += " ";
        }
        if (i < 3)
          str += "; ";
      }
      xmlShowInPort.InnerText = str;
      xmlProperties.AppendChild(xmlShowInPort);

      config.AppendChild(xmlProperties);
    }

    /// <summary>
    /// Reads Xml configuration of current module.
    /// </summary>
    /// <param name="config">XmlElement from map config file which belongs to current module.</param>
    public override void ReadConfig(System.Xml.XmlElement config)
    {
      foreach (XmlElement xeConfig in config.ChildNodes)
      {
        string[] lines;

        if (xeConfig.Name == "Properties")
        {
          foreach (XmlElement xeProperties in xeConfig.ChildNodes)
          {
            switch (xeProperties.Name)
            {
              case "Transform3D":
                lines = xeProperties.InnerText.Trim().Split(';');

                if (lines.Length != 4)
                  throw new MveException(this, "Incorrect number of lines in Transform3D.");

                for (int i = 0; i < 4; i++)
                {
                  string[] elems = lines[i].Trim().Split(' ');
                  if (elems.Length != 4)
                    throw new MveException(this, "Incorrect number of elements in line " + i + " in Transform3D.");

                  for (int j = 0; j < 4; j++)
                  {
                    transform3D[i, j] = double.Parse(elems[j].Trim(), NumberStyles.Float, CultureInfo.InvariantCulture.NumberFormat);
                  }
                }
                break;
              case "ShowInPort":
                lines = xeProperties.InnerText.Trim().Split(';');

                if (lines.Length != 4)
                  throw new MveException(this, "Incorrect number of lines in Transform3D.");

                for (int i = 0; i < 4; i++)
                {
                  string[] elems = lines[i].Trim().Split(' ');
                  if (elems.Length != 4)
                    throw new MveException(this, "Incorrect number of elements in line " + i + " in Transform3D.");

                  for (int j = 0; j < 4; j++)
                  {
                    showInPort[i, j] = bool.Parse(elems[j].Trim());

                    if (showInPort[i, j]) AddInPort("inM" + i + "" + j, typeof(Scalar), false, true);
                  }
                }
                break;
              default:
                Debug.WriteLine("Unreaded value:" + xeProperties.Name + " - " + xeProperties.InnerText);
                break;
            }
          }
        }
      }
    }

    /// <summary>
    /// Creates new instance of module and copy all necessary member variables.
    /// This method is used by MapEditor in Copy/Paste process.
    /// Without overriding it returns implicit instance of current module.
    /// </summary>
    public override Module DeepCopy()
    {
      Transform3DSource transform3DSource = (Transform3DSource) base.DeepCopy();

      transform3DSource.transform3D = (Transform3D) this.transform3D.DeepCopy();

      return transform3DSource;
    }
  }
}
