using System;
using System.Globalization;
using System.Collections;
using System.IO;
using System.Diagnostics;
using System.Text.RegularExpressions;
using Zcu.Mve.Core;
using Zcu.Mve.Strings;
using System.ComponentModel;

namespace Zcu.Mve.Visualization.Loaders
{
  /// <summary>
  /// Module for loading of triangle mesh from .tri format file.
  /// Support [Vertices], [Points], [Vertices' Normals], [Triangles' Normals]
  /// </summary>
  [ModuleInfo("Miroslav Vavruka, Libor Va",
     "Module for loading of triangle mesh from .tri format file.",
     CategoryType.Loader,
     CategoryData.Mesh,
     Assemblied = "2004-07-01",
     IconName = "TriangleLoader.ico")]
  [PortInfo("Output", "Loaded Mesh.")]
  [PortInfo("FileName", "Name of the file to load from. Not required, when no link connected, then the value of the FileName property is used.")]
  public class TriangleLoader : Zcu.Mve.Core.Module
  {
    /// <summary>
    /// Name of the file containing data.
    /// </summary>
    private string fileName;

    private bool paramsChanged = true;

    TriangleMesh mesh = null;

    /// <summary>
    /// File to load data from. Should be in the .tri format. The property has no effect when there's a connection at the Filename port.
    /// </summary>
    [Editor(typeof(System.Windows.Forms.Design.FileNameEditor), typeof(System.Drawing.Design.UITypeEditor))]
    [Description("File to load data from. Should be in the .tri format. The property has no effect when there's a connection at the Filename port.")]
    public string FileName
    {
      get
      {
        return (fileName);
      }
      set
      {
        if (fileName != value)
        {
          fileName = value;
          paramsChanged = true;

          //Console.WriteLine("Tri: FileName changed");
        }
      }
    }

    private bool reload = false;

    /// <summary>
    /// Determines whether or not to reload data each time the modul is asked for result. Set to true when the contens of the input files change, otherwise keep false.
    /// </summary>
    [Description("Determines whether or not to reload data each time the modul is asked for result. Set to true when the contens of the input files change, otherwise keep false.")]
    public bool Reload
    {
      get
      {
        return (reload);
      }
      set
      {
        reload = value;
      }
    }

    private bool invertNormals = false;
    /// <summary>
    /// Use this option if you are importing files originally made for right-handed coordinate systems.
    /// </summary>
    [Description("Use this option if you are importing files originally made for right-handed coordinate systems."), DefaultValue(false)]
    public bool InvertNormals
    {
      get
      {
        return invertNormals;
      }
      set
      {
        invertNormals = value;
        paramsChanged = true;
      }
    }

	/// <summary>
	/// Analyze type of selected file based on file extension.
	/// </summary>
	/// <param name="path">Full filename (full path and filename).</param>
	/// <returns>Return type of selected input file - 1 - TRI format,
	///                                               2 - STL format,
	///                                               3 - PLY format,
	///                                               4 - CMX format,
	///                                               0 - ??? format.</returns>
	private byte tellFileType(string path)
	{
		if (path.EndsWith(".tri"))
		{
			return 1;
		}
		else if (path.EndsWith(".stl"))
		{
			return 2;
		}
		else if (path.EndsWith(".ply"))
		{
			return 3;
		}
		else if (path.EndsWith(".cmx"))
		{
			return 4;
		}
		else
		{
			return 0;
		}
	}

    /// <summary>
    /// Creates module and output port, initialize fileName.
    /// </summary>
    public TriangleLoader()
    {
      AddOutPort("Output", typeof(Visualization.TriangleMesh));
      AddInPort("FileName", typeof(MveString), false);
      this.fileName = "";
      this.SimulationStarted += new SimulationStartedEventHandler(TriangleLoader_OnSimulationStartEvent);
    } // TriangleLoader()

    void TriangleLoader_OnSimulationStartEvent(object source)
    {
      //paramsChanged = true;
    } 				

    /// <summary>
    /// Executes TriangleLoader functionality.
    /// </summary>
    public override void Execute()
    {
      if (this.GetInPort("FileName").Link != null)
      {
        MveString fn = (MveString) GetInput("FileName");
        FileName = fn.Value;
      }

      if (this.fileName == "")		// nezadano zadne jmeno souboru
        throw new MveException(this, this.GetType().Name + " \"" + this.Name + "\" error: Please enter the name of a file with data.");

      if (paramsChanged || reload || (mesh == null))
      {
        //Console.WriteLine("Paramschanged" + paramsChanged + " reload" + reload);
        StreamReader sr = null;

        try
        {
          sr = new StreamReader(fileName);

          // initializa arrraylists
          ArrayList points = new ArrayList();
          ArrayList cells = new ArrayList();
          ArrayList pointNormals = new ArrayList();
          ArrayList cellNormals = new ArrayList();

          // read data from file
          ReadTriFile(sr, points, cells, pointNormals, cellNormals);

          // create mesh and fill it with loaded data
          mesh = new TriangleMesh();
          mesh.Points = UniformDataArray.FromArrayList(typeof(Point3D), points);

          if (cells.Count != 0)
            mesh.AddCells(typeof(Triangle).Name, UniformDataArray.FromArrayList(typeof(Triangle), cells));

          if (cellNormals.Count != 0)
            mesh.AddCellAttrs(typeof(Triangle).Name, typeof(Normal3D).Name, UniformDataArray.FromArrayList(typeof(Normal3D), cellNormals));

          if (pointNormals.Count != 0)
            mesh.AddPointAttrs(typeof(Normal3D).Name, UniformDataArray.FromArrayList(typeof(Normal3D), pointNormals));

          //Console.WriteLine("Tri: Setoutput na data not same");
          // paramsChanged = false; <--- pridal Jan Kaiser, kvuli isdatasame
          paramsChanged = false;
          SetOutput("Output", mesh, false);
        } // try
        catch (FileNotFoundException ex)
        {
          Debug.WriteLine(this.GetType().Name + " '" + this.Name + "' error:");
          Debug.WriteLine(ex);
          throw new MveException(this, this.GetType().Name + " \"" + this.Name + "\" error: can not find file: \"" + this.fileName + "\"");
        }
        catch (DirectoryNotFoundException ex)
        {
          Debug.WriteLine(this.GetType().Name + " '" + this.Name + "' error:");
          Debug.WriteLine(ex);
          throw new MveException(this, this.GetType().Name + " \"" + this.Name + "\" error: can not find directory: \"" + this.fileName + "\"");
        }
        finally
        {
          if (sr != null)
            sr.Close();
        }
      }
      else
      {
        //pridal Jan Kaiser, kvuli zjistovani isdatasame
        // Console.WriteLine("Tri: Setoutput na data same");

        SetOutput("Output", mesh, true);
      }
    } // Execute()

    /// <summary>
    /// Support method, reads data from the opened stream
    /// </summary>
    /// <param name="sr">Strem to read.</param>
    /// <param name="pts">List of points.</param>
    /// <param name="trs">List of triangles.</param>
    /// <param name="ptsNorm">List of points normals.</param>
    /// <param name="trsNorm">List of triangles normals.</param>
    private void ReadTriFile(StreamReader sr, ArrayList pts, ArrayList trs, ArrayList ptsNorm, ArrayList trsNorm)
    {
      int nLine = 1;					// cislo prave nacitane radky
      NumberFormatInfo nfi = Globals.nfi;
      string line;
      line = sr.ReadLine();

      try
      {
        while (line != null)
        {
          switch (line.Trim())
          {
            case "[Vertices]":    // nacteni bodu
              line = sr.ReadLine();
              nLine++;
              while (line != null && (line.IndexOf("[") == -1))
              {
                AddPoint(pts, line);
                line = sr.ReadLine();
                if (line != null)
                  nLine++;
              }
              break;

            case "[Triangles]":    // nacteni trojuhelniku
              line = sr.ReadLine();
              nLine++;
              while (line != null && (line.IndexOf("[") == -1))
              {
                AddTriangle(trs, line);
                line = sr.ReadLine();
                if (line != null)
                  nLine++;
              }
              break;

            case "[Vertices' Normals]":	 // nacteni normal bodu
              line = sr.ReadLine();
              nLine++;
              while (line != null && (line.IndexOf("[") == -1))
              {
                AddNormal(ptsNorm, line);
                line = sr.ReadLine();
                if (line != null)
                  nLine++;
              }
              break;

            case "[Triangles' Normals]":	// nacteni normal trojuhelniku
              line = sr.ReadLine();
              nLine++;
              while (line != null && (line.IndexOf("[") == -1))
              {
                AddNormal(trsNorm, line);
                line = sr.ReadLine();
                if (line != null)
                  nLine++;
              }
              break;

            default:						// nacteni radky s 'nepotrebnymi' udaji
              line = sr.ReadLine();
              nLine++;
              break;
          } // switch
        } // while
      }
      catch (Exception ex)
      {
        Debug.WriteLine(this.GetType().Name + " '" + this.Name + "' error:");
        Debug.WriteLine(ex);
        throw new MveException(this, this.GetType().Name + " \"" + this.Name + "\" error: Invalid data format at line: " + nLine + " in file \"" + this.fileName + "\"");
      } // catch
    }

    /// <summary>
    /// Split the lino into 3 double values (point coordinates).
    /// </summary>
    /// <param name="line">Current line loaded from the file.</param>
    /// <param name="x">x-coordinate of Point.</param>
    /// <param name="y">y-coordinate of Point.</param>
    /// <param name="z">z-coordinate of Point.</param>
	private void SplitAndCorrectLineDouble(string line, out double x, out double y, out double z)
    {
      x = y = z = 0.0;
		char[] separators = {' ','\t'};
      string[] elems = line.Trim().Split(separators,StringSplitOptions.RemoveEmptyEntries);
      if (elems.Length != 3)
      {
        throw new MveException(this, "Invalind number of elements ...");
      }
      x = Double.Parse(elems[0], Globals.nfi);
      y = Double.Parse(elems[1], Globals.nfi);
      z = Double.Parse(elems[2], Globals.nfi);

    } // SplitAndCorrectLine()

    /// <summary>
    /// Split the lino into 3 integer values (for example indices).
    /// </summary>
    /// <param name="line">Current line loaded from the file.</param>
    /// <param name="x">First value.</param>
    /// <param name="y">Second value.</param>
    /// <param name="z">Third value.</param>
    private void SplitAndCorrectLineInt32(string line, out int x, out int y, out int z)
    {
      x = y = z = 0;

      string[] elems = line.Trim().Split(' ', '\t');
      if (elems.Length != 3)
      {
        throw new MveException(this, "Invalind number of elements ...");
      }
      x = Int32.Parse(elems[0], Globals.nfi);
      y = Int32.Parse(elems[1], Globals.nfi);
      z = Int32.Parse(elems[2], Globals.nfi);

    } // SplitAndCorrectLine()

    /// <summary>
    /// Parse line containing point coordinates and add the point into list.
    /// </summary>
    /// <param name="triangles">List of points.</param>
    /// <param name="line">Current line loaded from the file.</param>
    private void AddPoint(ArrayList triangles, string line)
    {
      double x, y, z;
      if (line == "")
        return;
      this.SplitAndCorrectLineDouble(line, out x, out y, out z);
      triangles.Add(new Point3D(x, y, z));
    } // InsertPoint()

    /// <summary>
    /// Parse line containing triangle indices and add the triangle into list.
    /// </summary>
    /// <param name="triangles">List of triangles.</param>
    /// <param name="line">Current line loaded from the file.</param>
    private void AddTriangle(ArrayList triangles, string line)
    {
      if (line == "")
        return;
      int i, j, k;

      this.SplitAndCorrectLineInt32(line, out i, out j, out k);

      triangles.Add(new Triangle(i, j, k));
    } // InsertCell()

    /// <summary>
    /// Parse line containing normal values and add the normal into list.
    /// </summary>
    /// <param name="normals">List of normals.</param>
    /// <param name="line">Current line loaded from the file.</param>
    private void AddNormal(ArrayList normals, string line)
    {
      double x, y, z;
      if (line == "")
        return;
      this.SplitAndCorrectLineDouble(line, out x, out y, out z);
      if (invertNormals)
      {
        normals.Add(new Normal3D(-x, -y, -z));
      }
      else
      {
        normals.Add(new Normal3D(x, y, z));
      }
    } // AddNormal()
  } // class TriangleLoader
} // namespace
