using System;
using System.Diagnostics;
using System.Xml;
using System.Collections;
using System.Collections.Generic;
using Zcu.Mve.Core;

namespace Zcu.Mve.Visualization
{
    /// <summary>
    /// Homogenous one dimensional array of IDataObjects. 
    /// </summary>
    public class UniformDataArray : MveArray
    {
        /// <summary>
        /// For XmlLoader use only. Use parametric construcotr instead.
        /// </summary>
        public UniformDataArray()
        {
            UpdateModificationTime();
        }

        /// <summary>
        /// Creates fixed size, one dimensional array of DataObjects
        /// </summary>
        /// <param name="type">Type of object (only this type is accepted). Must be inherited from Mve.Core.DataObject</param>
        /// <param name="count">Number of array elements</param>
        public UniformDataArray(Type type, int count)
            : base(type, count)
        {
            UpdateModificationTime();
        }

        /// <summary>
        /// Number of elements.
        /// </summary>
        public int Count
        {
            get
            {
                return this.Lengths[0];
            }
        }

        /// <summary>
        /// Creates UniformDataArray from ArrayList. 
        /// Elements are not copied. Reference only is passed to the new array.
        /// </summary>
        /// <param name="type">Type of uniform data array.</param>
        /// <param name="list">List to be converted (copied).</param>
        /// <returns>New instance of UniformDataArray.</returns>
        public static UniformDataArray FromArrayList(Type type, ArrayList list)
        {
            UniformDataArray ret = new UniformDataArray(type, list.Count);

            for (int i = 0; i < list.Count; i++)
                ret[i] = (IDataObject)list[i];

            return ret;
        }

        /// <summary>
        /// Creates a uniform data array from a list of IDataObject instances
        /// </summary>
        /// <typeparam name="TDataType">type of the data, must be IDataObject</typeparam>
        /// <param name="l">List to use for creating the new UniformDataArray.</param>
        /// <returns></returns>
        public static UniformDataArray FromList<TDataType>(List<TDataType> l) where TDataType : IDataObject
        {
            UniformDataArray ret = new UniformDataArray(typeof(TDataType), l.Count);

            for (int i = 0; i < l.Count; i++)
            {
                ret[i] = l[i];
            }
            return (ret);
        }

        /// <summary>
        /// Creates a uniform data array from the keys of a dictionary
        /// </summary>
        /// <typeparam name="TDataType">data type of the keys, must be IDataObject</typeparam>
        /// <typeparam name="TValueType">type of the values, has no effect</typeparam>
        /// <param name="l"></param>
        /// <returns></returns>
        public static UniformDataArray FromDictionaryKeys<TDataType, TValueType>(Dictionary<TDataType, TValueType> l) where TDataType : IDataObject
        {
            UniformDataArray ret = new UniformDataArray(typeof(TDataType), l.Count);

            int i = 0;
            foreach (TDataType e in l.Keys)
            {
                ret[i] = e;
                i++;
            }
            return (ret);
        }

        // OP 2009-08-03
        /// <summary>
        /// Creates a new UniformDataArray instance from a regular array.
        /// Elements are referenced, not copied.
        /// </summary>
        /// <typeparam name="TDataType">Type of elements of the input array, has to be IDataObject</typeparam>
        /// <param name="array">Input array</param>
        /// <returns>New instance of UniformDataArray referencing the same elements as the input array</returns>
        public static UniformDataArray FromArray<TDataType>(TDataType[] array) where TDataType : IDataObject
        {
            UniformDataArray res = new UniformDataArray(typeof(TDataType), array.Length);

            for (int i = 0; i < array.Length; i++)
            {
                res[i] = array[i];
            }
            return res;
        }

        /*
            /// <summary>
            /// Deep copy of data object and its subelements.
            /// </summary>
            /// <returns>same data, different memmory space.</returns>
            public override IDataObject DeepCopy()
            {
              UniformDataArray mveArray = new UniformDataArray(this.ElemType, lengths);
              mveArray.array = (Array) this.array.Clone();
      
              // Elements are DataObjects ... lets make deep copy.
              if (typeof(Zcu.Mve.Core.DataObject).IsInstanceOfType(this.ElemType))
              {
                CloneDimension(mveArray.array, this.Rank, new int [this.Rank]);
              }

              return mveArray;
            }
         */
    }
}
