using System;
using System.Collections.Generic;
using System.Text;
using Zcu.Mve.Core;
using Zcu.Mve.Visualization;
using System.ComponentModel;
using System.Xml;
using System.Diagnostics;

namespace Zcu.Mve.Visualization.Processing
{
  /// <summary>
  /// This enum is primarily intended to be used within UnstrGridModifyTransform module
  /// to allow user to choose what operation will be performed between two
  /// transform matrices.
  /// </summary>
  public enum TransformOperation
  {
    /// <summary>
    /// Former matrix will be replaced with the new one.
    /// </summary>
    Replace = 0,
    /// <summary>
    /// Mltiply former matrix with the new one.
    /// </summary>
    Multiply = 1
  }

  /// <summary>
  /// Modifies Transform matrix of UnstrGrid. Points of UnstrGrid remains untouched.
  /// </summary>
  [ModuleInfo("Jan Kaiser", "Modifies Transform matrix of UnstrGrid. Points of UnstrGrid remains untouched.", CategoryType.Filter, CategoryData.Mesh, IconName = "UnstrGridModifyTransform.ico")]
  [PortInfo("inData", "UnstrGrid whose transform matrix is to be changed.")]
  [PortInfo("inTransform", "Transform matrix.")]
  public class UnstrGridModifyTransform : Module
  {
    private bool dataChanged = true;
    private UnstrGrid output;
    private TransformOperation transformOperation;

    /// <summary>
    /// Option to replace or multiply former transform matrix.
    /// </summary>
    [Description("Option to replace or multiply former transform matrix."), DefaultValue(TransformOperation.Replace)]
    public TransformOperation TransformOperation
    {
      get
      {
        return transformOperation;
      }
      set
      {
        transformOperation = value;
        dataChanged = true;
      }
    }

    /// <summary>
    /// Default constructor. Only adds input ports, output port is added in runtime when link is connected to the Data input port.
    /// </summary>
    public UnstrGridModifyTransform()
    {
      AddInPort("inData", typeof(UnstrGrid));
      AddInPort("inTransform", typeof(Transform3D));

      this.PortConnected += new PortConnectedEventHandler(UnstrGridTransform_OnConnectEvent);
      this.PortDisconnected += new PortDisconnectedEventHandler(UnstrGridTransform_OnDisconnectEvent);

      transformOperation = TransformOperation.Replace;
    }

    /// <summary>
    /// Removes "outData" port when "inData" port is disconnected.
    /// </summary>
    private void UnstrGridTransform_OnDisconnectEvent(object source, OutPort portFrom, InPort portTo)
    {
      if (portTo.Name == "inData")
      {
        RemoveOutPort("outData");
      }
    }

    /// <summary>
    /// Adds "outData" port when "inData" port is connected.
    /// </summary>
    private void UnstrGridTransform_OnConnectEvent(object source, OutPort portFrom, InPort portTo)
    {
      if (portTo.Name == "inData")
      {
        AddOutPort("outData", portFrom.PortType);
      }
    }

    /// <summary>
    /// On Execute changes the transformation matrix of UnstrGrid.
    /// </summary>
    public override void Execute()
    {
      if (!IsAllDataSame() || (output == null) || dataChanged)
      {
        //Debug.WriteLine("UnstrGridModifyTransform YO");

        UnstrGrid input = (UnstrGrid) GetInput("inData");
        Transform3D transform = (Transform3D) GetInput("inTransform");

        //Console.WriteLine("UMT: " + transform);

        switch (transformOperation)
        {
          case TransformOperation.Replace:
            output = input.ShallowCopy();
            output.Transform = transform;
            break;
          case TransformOperation.Multiply:
            output = input.ShallowCopy();
            output.Transform = transform * ((Transform3D) input.Transform);
            break;
        }

        //Console.WriteLine("1:" + input.Points.ModificationTime + " 2:" + output.Points.ModificationTime);
        //Console.WriteLine("UMT O: " + output.Transform);

        SetOutput("outData", output);
        dataChanged = false;
      }
      else
      {
        SetOutput("outData", output, true);
      }
    }

    /// <summary>
    /// Reads module settings.
    /// </summary>
    /// <param name="config">XmlElement from map config file which belongs to current module.</param>
    public override void ReadConfig(System.Xml.XmlElement config)
    {
      foreach (XmlElement xeConfig in config.ChildNodes)
      {
        if (xeConfig.Name == "Properties")
        {
          foreach (XmlElement xeProperties in xeConfig.ChildNodes)
          {
            switch (xeProperties.Name)
            {
              case "TransformOperation":
                switch (xeProperties.InnerText.Trim())
                {
                  case "Multiply":
                    this.transformOperation = TransformOperation.Multiply;
                    break;
                  case "Replace":
                    this.transformOperation = TransformOperation.Replace;
                    break;
                  default:
                    this.transformOperation = TransformOperation.Replace;
                    break;
                }
                break;
            }
          }
        }
      }
    }

    /// <summary>
    /// Creates new instance of module and copy all necessary member variables.
    /// This method is used by MapEditor in Copy/Paste process.
    /// Without overriding it returns implicit instance of current module.
    /// </summary>
    public override Module DeepCopy()
    {
      UnstrGridModifyTransform unstrGridModifyTransform = (UnstrGridModifyTransform) base.DeepCopy();

      unstrGridModifyTransform.transformOperation = this.transformOperation;

      return unstrGridModifyTransform;
    }
  }
}
