using System;
using System.Collections.Generic;
using System.Text;
using System.ComponentModel;
using System.ComponentModel.Design;
using System.Diagnostics;
using System.Globalization;

using Zcu.Mve.Core;
using Zcu.Mve.Numerics;
using Zcu.Mve.Visualization;

namespace Zcu.Mve.Visualization.Processing
{
    /// <summary>
    /// Module for block using for module "cutting by polyhedron"
    /// Block have 3 important parametres (width, height, deepth) and own location in space.
    /// </summary>    
    [ModuleInfo("Roman Soukal", "Module for block using by \"convex polyhedron cutting\"", CategoryType.Filter, CategoryData.Volume, IconName = "VolumeCut3.ico")]
    [PortInfo("Output", "Output UnstrGrid block data")]
    public class BlockCutting : Module
    {
        /// <summary>
        /// Default constructor.
        /// </summary>
        public BlockCutting()
        {
            AddOutPort("Output", typeof(UnstrGrid));
        }

        #region Variables
        UnstrGrid result = null;
        private double height = 0.5, width = 0.5, deeph = 0.5;

        private string planeAttributesName = "Planes";
        private string blockCoordsS = "0, 0, 0";
        private double[] blockCoords = { 0, 0, 0 };
        private bool paramsChanged = false;
        #endregion

        /// <summary>
        /// Overrided
        /// </summary> 
        public override void Execute()
        {
            if (result==null || paramsChanged)
            {
                result = new UnstrGrid();

                UniformDataArray points = new UniformDataArray(typeof(Point3D), 7);
                points[0] = new Point3D(blockCoords[0] + 0, blockCoords[1] + 0, blockCoords[2] + 0);
                points[1] = new Point3D(blockCoords[0] + width, blockCoords[1] + 0, blockCoords[2] + 0);
                points[2] = new Point3D(blockCoords[0] + 0, blockCoords[1] + height, blockCoords[2] + 0);
                points[3] = new Point3D(blockCoords[0] + width, blockCoords[1] + height, blockCoords[2] + 0);
                points[4] = new Point3D(blockCoords[0] + 0, blockCoords[1] + 0, blockCoords[2] + deeph);
                points[5] = new Point3D(blockCoords[0] + width, blockCoords[1] + 0, blockCoords[2] + deeph);
                points[6] = new Point3D(blockCoords[0] + 0, blockCoords[1] + height, blockCoords[2] + deeph);

                result.Points = points;

                UniformDataArray triangles = new UniformDataArray(typeof(Triangle), 6);
                triangles[0] = new Triangle(0, 1, 2);
                triangles[1] = new Triangle(5, 4, 6);
                triangles[2] = new Triangle(0, 4, 1);
                triangles[3] = new Triangle(2, 3, 6);
                triangles[4] = new Triangle(1, 5, 3);
                triangles[5] = new Triangle(0, 2, 4);

                result.AddCells(planeAttributesName, triangles);

                SetOutput("Output", result);
                paramsChanged = false;
            }
            else
            {
                SetOutput("Output", result, true);
            }
        }

        #region Properties
        /// <summary>
        /// Set name of RegGridND attribute with each plane information. Default set to Planes.
        /// </summary>
        [Description("Set name of RegGridND attribute with each plane information. Default set to\"Planes\"")]
        public string PlaneAttributeName
        {
            set
            {
                if (value == string.Empty)
                {
                    throw new MveException(this, "Attribut property cannot be set to an empty string!");
                }
                else
                {
                    this.planeAttributesName = value;
                    this.paramsChanged = true;
                }
            }

            get { return this.planeAttributesName; }
        }


        /// <summary>
        /// Coordinates of left buttom corner of block 
        /// Example: a, b, c where a, b, c are floating point numbers.
        /// </summary>
        [Description("Coordinates of left buttom corner of block. Example: a, b, c where a, b, c are floating point numbers.")]
        public string BlockCoordinations
        {
            get { return blockCoordsS; }
            set
            {
                try
                {
                    string[] values = value.Split(new char[] { ',' });
                    if (values.Length != 3)
                        throw new MveException(this, "The plane is needed in correct form. " +
                            "Example: a, b, c where a, b, c are floating point numbers.");
                    blockCoords = new double[values.Length];
                    for (int i = 0; i < values.Length; i++)
                    {
                        blockCoords[i] = double.Parse(values[i].Trim(), NumberFormatInfo.InvariantInfo);
                    }
                    blockCoordsS = value;
                    paramsChanged = true;
                }
                catch (FormatException)
                {
                    blockCoords = null;
                    throw new MveException(this, "The plane is needed in correct form. " +
                        "Example: a, b, c where a, b, c are floating point numbers.");
                }
            }
        }

        /// <summary>
        /// Height of block.
        /// </summary>
        [Description("Set height of block. Default set to \"0.5\"")]
        public string Height
        {
            set
            {
                try
                {
                    height = double.Parse(value.Trim(), NumberFormatInfo.InvariantInfo);
                    this.paramsChanged = true;
                }
                catch (FormatException)
                {
                    throw new MveException(this, "The number is needed in correct form. " +
                        "Example: -5.2");
                }
            }

            get { return this.height.ToString(NumberFormatInfo.InvariantInfo); }
        }

        /// <summary>
        /// Width of block.
        /// </summary>
        [Description("Set width of block. Default set to \"0.5\"")]
        public string Width
        {
            set
            {
                try
                {
                    width = double.Parse(value.Trim(), NumberFormatInfo.InvariantInfo);
                    this.paramsChanged = true;
                }
                catch (FormatException)
                {
                    throw new MveException(this, "The number is needed in correct form. " +
                        "Example: -5.2");
                }
            }

            get { return this.width.ToString(NumberFormatInfo.InvariantInfo); }
        }

        /// <summary>
        /// Deeph of block.
        /// </summary>
        [Description("Set deeph of block. Default set to \"0.5\"")]
        public string Deeph
        {
            set
            {
                try
                {
                    deeph = double.Parse(value.Trim(), NumberFormatInfo.InvariantInfo);
                    this.paramsChanged = true;
                }
                catch (FormatException)
                {
                    throw new MveException(this, "The number is needed in correct form. " +
                        "Example: -5.2");
                }
            }

            get { return this.deeph.ToString(NumberFormatInfo.InvariantInfo); }
        }

        #endregion
    }
}
