﻿using System;
using System.Collections.Generic;
using System.Text;
using System.ComponentModel;
using System.ComponentModel.Design;
using System.Diagnostics;

using Zcu.Mve.Core;
using Zcu.Mve.Numerics;
using Zcu.Mve.Visualization;

namespace Zcu.Mve.Visualization.Processing
{
    /// <summary>
    /// Module for cutting by distance from point input point
    /// Module calculates distances from point for all the grid.
    /// </summary>    
    [ModuleInfo("Roman Soukal", "Module for point distance calculate used to cutting", CategoryType.Filter, CategoryData.Volume, IconName = "VolumeCut4.ico")]
    [PortInfo("Input Grid", "Input RegGridND data")]
    [PortInfo("Input Point", "Input PointND (Size N=3 is needed) to calculate a distance")]
    [PortInfo("Output Grid", "Output RegGridND data")]
    public class PointDistanceCutting : Module
    {
        /// <summary>
        /// Default constructor.
        /// </summary>
        public PointDistanceCutting()
        {
            AddInPort("Input Grid", typeof(RegGridND));
            AddInPort("Input Point", typeof(PointND));
            AddOutPort("Output Grid", typeof(RegGridND));
        }

        #region Variables
        private RegGridND data = null;
        private string pointAttributesName = "Values";
        private string cuttingAttributesName = "CuttingValues";
        private bool paramsChanged = false;
        #endregion

        /// <summary>
        /// Overrided
        /// </summary> 
        public override void Execute()
        {
            if (!IsAllDataSame() || paramsChanged)
            {
                RegGridND old = (RegGridND)this.GetInput("Input Grid");
                data = old.ShallowCopy();
                //IDataObject val = null;
                //if (!data.UserTags.TryGetValue("Min", out val))
                //{
                //    old.UserTags.TryGetValue("Min", out val);
                //    data.UserTags.Add("Min", val);
                //}
                //if (!data.UserTags.TryGetValue("Max", out val))
                //{
                //    old.UserTags.TryGetValue("Max", out val);
                //    data.UserTags.Add("Max", val);
                //}
                //if (!data.UserTags.TryGetValue("Distortion", out val))
                //{
                //    old.UserTags.TryGetValue("Distortion", out val);
                //    data.UserTags.Add("Distortion", val);
                //}

                PointND point = (PointND)this.GetInput("Input Point");
                if(point.Length!=3)
                    throw new MveException(this, "Size N=3 of PointND is expected!");
                MveArray values = data.GetPointAttrs(this.PointAttributeName);
                MveArray newValues = new MveArray(typeof(Scalar), values.Lengths);
                int[] sizes = values.Lengths;
                for (int i = 0; i < sizes[0]; i++)
                {
                    this.ProgressInfo(((float)i / sizes[0]));
                    for (int j = 0; j < sizes[1]; j++)
                        for (int k = 0; k < sizes[2]; k++)
                        {
                            int[] coords = { i, j, k };
                            PointND pValue = data.GetPoint(coords);
                            double distance1 = pValue.Values[0] - point.Values[0];
                            double distance2 = pValue.Values[1] - point.Values[1];
                            double distance3 = pValue.Values[2] - point.Values[2];
                            double distance = distance1 * distance1 + distance2 * distance2 + distance3 * distance3;

                            newValues.Array.SetValue(new Scalar(Math.Sqrt(distance)), coords);
                        }
                }
                data.AddPointAttrs(this.cuttingAttributesName, newValues);
                this.SetOutput("Output Grid", this.data, false);
                paramsChanged = false;
            }
            else
            {
                this.SetOutput("Output Grid", data, true);
            }
        }

        #region Properties

        /// <summary>
        /// Set name of RegGridND point attribute. Default set to Values.
        /// </summary>
        [Description("Set name of RegGridND point attribute. Default set to \"Values\"")]
        public string PointAttributeName
        {
            set
            {
                if (value == string.Empty)
                {
                    throw new MveException(this, "Attribut property cannot be set to an empty string!");
                }
                else
                {
                    this.pointAttributesName = value;
                    this.paramsChanged = true;
                }
            }

            get { return this.pointAttributesName; }
        }

        /// <summary>
        /// Set name of RegGridND attribute with cutting information. Default set to CuttingValues.
        /// </summary>
        [Description("Set name of RegGridND attribute with cutting information. Default set to\"CuttingValues\"")]
        public string CuttingAttributeName
        {
            set
            {
                if (value == string.Empty)
                {
                    throw new MveException(this, "Attribut property cannot be set to an empty string!");
                }
                else
                {
                    this.cuttingAttributesName = value;
                    this.paramsChanged = true;
                }
            }

            get { return this.cuttingAttributesName; }
        }

        #endregion
    }
}
