﻿using System;
using System.Threading;
using System.ComponentModel;
using System.Collections.Generic;

using Zcu.Mve.Core;
using Zcu.Mve.Numerics;
using Zcu.Mve.Visualization;

namespace Zcu.Mve.Visualization.Processing
{

    #region Structures
    /// <summary> Vertex of the cube or tetrahedron. It has coordinates, normal and value. </summary>
    public struct Vertex
    {
        /// <summary>
        /// vertex position
        /// </summary>
        public Vector3D coords;

        /// <summary>
        /// vertex normal
        /// </summary>
        public Vector3D normal;

        /// <summary>
        /// index of the vertex
        /// </summary>
        public int index;

        /// <summary>
        /// value associated with the vertex
        /// </summary>
        public double value;

        /// <summary>
        /// Constructs new vertex of the cube or tetrahedron. It has coordinates, normal and value.
        /// </summary>
        /// <param name="coords">coordinates of the vertex</param>
        /// <param name="normal">normal of the vertex</param>
        /// <param name="value">value associated with the vertex</param>
        /// <param name="id">identifier of the vertex</param>
        public Vertex(Vector3D coords, Vector3D normal, double value, int id)
        {
            this.coords = coords;
            this.normal = normal;
            this.value = value;
            index = id;
        }
    }

    /// <summary> Each cube (voxel, cell) is created from 8 vertices. </summary>
    public struct Cube
    {
        /// <summary>
        /// Array of vertics of the cube
        /// </summary>
        public Vertex[] p;

        /// <summary>
        /// contructs a new cube from given points (vertices)
        /// </summary>
        /// <param name="points"></param>
        public Cube(Vertex[] points)
        {
            p = points;
        }
    }

    #endregion

    /// <summary>
    /// Determines the data type of the dataset.
    /// </summary>
    public enum ValueDataType {
        /// <summary>
        /// Integer data type
        /// </summary>
        Integer,
        /// <summary>
        /// Scalar (double) data type
        /// </summary>
        Scalar };

    /// <summary>
    /// The iso-surface extractor module.
    /// </summary>
    public abstract class IsoSurfaceExtractor : Zcu.Mve.Core.Module
    {
        /// <summary>find min/max?</summary>
        protected bool findMinMax = false;
        /// <summary>did parameters change?</summary>
        protected bool paramsChanged = true;
        /// <summary>use threshold and iso value?</summary>
        protected bool useThresholdAsIsoValue = false;
        /// <summary>use linear interpolation?</summary>
        protected bool linearInterpolation = true;
        /// <summary>is it the first run?</summary>
        protected bool firstRunForThisData = true;

        /// <summary>minimum value</summary>
        protected double min = double.MaxValue;
        /// <summary>maximum value</summary>
        protected double max = double.MinValue;
        /// <summary>iso value</summary>
        protected double isoValue = 0;
        /// <summary>threshold</summary>
        protected double threshold = 50;

        /// <summary>size array</summary>
        protected int[] size = null;
        /// <summary>array of indices</summary>
        protected int[] indexList = null;
        /// <summary>integer data</summary>
        protected int[] dataI = null;

        /// <summary>distortions</summary>
        protected double[] distortion;
        /// <summary>inverse distortions</summary>
        protected double[] inverseDistortion;
        /// <summary>double data</summary>
        protected double[] dataD = null;

        /// <summary>point attribute name</summary>
        protected string pointAttributesName = "Values";
        /// <summary>normal attribute name</summary>
        protected string normalAttributeName = "Normal3D";

        /// <summary>current cube</summary>
        protected Cube cube;
        /// <summary>previous cube</summary>
        protected Cube previousCube;
        /// <summary>generated vertices</summary>
        protected Vertex[] vertList = null;
        /// <summary>list of vertices</summary>
        protected List<Vertex> pts = null;
        /// <summary>list of triangles</summary>
        protected List<Triangle> indices = null;
        /// <summary>input data</summary>
        protected RegGridND inData = null;
        /// <summary>volume data</summary>
        protected MveArray dataMveArray = null;
        /// <summary>output data</summary>
        protected TriangleMesh mesh = null;
        /// <summary>output array of points</summary>
        protected UniformDataArray points;
        /// <summary>output array of normals</summary>
        protected UniformDataArray normals;
        /// <summary>output array of triangles</summary>
        protected UniformDataArray triangles;
        /// <summary>data type</summary>
        protected ValueDataType valueDataType;

        /// <summary> Default constructor. Creates input and output ports. </summary>
        public IsoSurfaceExtractor()
        {
            AddInPort("InThreshold", typeof(Integer), false);
            AddInPort("InRegGridND", typeof(RegGridND));
            AddOutPort("Out", typeof(TriangleMesh));
        }

        /// <summary> This method initializes variables and call cube extracting method. </summary>
        protected virtual void ComputeIsoSurface()
        {
            IsoSurfaceExtractor.GCClean();

            Initialize();

            DecomposeVolumeIntoCubes(); // extract cubes

            Console.WriteLine("\n  Threshold:      {0}{1}\n  Triangles:       {2}\n   Points:           {3}\n",
                threshold, useThresholdAsIsoValue ? "" : "% (= " + isoValue + ")", indices.Count, pts.Count);

            CreateOutput(); // creates from computed data output data

            firstRunForThisData = false;
            paramsChanged = false;
        } // ComputeIsoSurface

        /// <summary> Initializes variables. </summary>
        protected virtual void Initialize()
        {
            dataMveArray = null;
            inData = null;
            points = null;
            normals = null;
            triangles = null;
            indices = null;
            vertList = null;
            indexList = null;
            mesh = null;

            if (!IsDataSame("InRegGridND"))
                firstRunForThisData = true;

            if (GetInput("InRegGridND") != null)
                inData = (RegGridND)GetInput("InRegGridND");

            if (GetInput("InThreshold") != null)
                threshold = (Integer)GetInput("InThreshold");

            dataMveArray = inData.GetPointAttrs(pointAttributesName);
            size = dataMveArray.Lengths;

            previousCube = cube = new Cube(new Vertex[8]);
            vertList = new Vertex[12];
            pts = new List<Vertex>();
            indices = new List<Triangle>();
            ProgressInfo(0f);

            if (inData.UserTags.ContainsKey("Min") && inData.UserTags.ContainsKey("Max"))
            {
                min = (Scalar)inData.UserTags["Min"];
                max = (Scalar)inData.UserTags["Max"];
            }
            else
                findMinMax = true;

            distortion = inData.Maxs;
            inverseDistortion = new double[distortion.Length];
            for (int i = 0; i < distortion.Length; i++)
                inverseDistortion[i] = 1 / distortion[i];                    

            if (firstRunForThisData)
                TypecastData();

            if (useThresholdAsIsoValue)
                isoValue = threshold;
            else
                isoValue = min + ((max - min) * (threshold / 100f));

            dataMveArray = null;
            inData = null;

            IsoSurfaceExtractor.GCClean();
        }

        #region Old typecasting without pointer utilisation
        /*/// <summary> Typecasts MveArray elements from MVE-datatypes to C# datatypes. </summary>
         protected virtual void TypecastData()
         {
             int size0size1 = size[0] * size[1];
             int index = 0, id = 0;
             double size0 = size[0];

             long ticks = Environment.TickCount;
             Console.Write("Converting data...");

             if (dataMveArray.ElemType == typeof(Zcu.Mve.Numerics.Integer))
             {
                 valueDataType = ValueDataType.Integer;
                 dataI = new int[size0size1 * size[2]];
                 for (int i = 0; i < size[0]; i++)
                 {
                     for (int j = 0; j < size[1]; j++)
                     {                        
                         index = i + j * size[0];
                         for (int k = 0; k < size[2]; k++)
                         {                           
                             id = index + k * size0size1;
                             dataI[id] = (Integer)dataMveArray[i, j, k];
                             if (findMinMax)
                             {
                                 if (dataI[id] <= min)
                                     min = dataI[id];
                                 if (dataI[id] >= max)
                                     max = dataI[id];
                             }
                         }
                     }
                     ProgressInfo(0.2 * (i + 1) / size0);
                 }//       
             }
             else if (dataMveArray.ElemType == typeof(Zcu.Mve.Numerics.Scalar))
             {
                 valueDataType = ValueDataType.Scalar;
                 dataD = new double[size0size1 * size[2]];
                 for (int i = 0; i < size[0]; i++)
                 {                    
                     for (int j = 0; j < size[1]; j++)
                     {
                      
                         index = i + j * size[0];
                         for (int k = 0; k < size[2]; k++)
                         {
                             id = index + k * size0size1;
                             dataD[id] = (Scalar)dataMveArray[i, j, k];
                             if (findMinMax)
                             {
                                 if (dataD[id] <= min)
                                     min = dataD[id];
                                 if (dataD[id] >= max)
                                     max = dataD[id];
                             }
                         }
                     }
                     ProgressInfo(0.2 * (i + 1) / size0);
                 }// 
             }                     
             Console.WriteLine("\t\tDone ({0} ms)", Environment.TickCount - ticks);

             dataMveArray = null;
             inData = null;

             IsoSurfaceExtractor.GCClean();
         }//*/
        #endregion

        /// <summary> Typecasts MveArray elements from MVE-datatypes to C# datatypes. </summary>
        protected virtual void TypecastData()
        {
            int size0size1 = size[0] * size[1];
            int index = 0, valueI = 0;
            double size0 = size[0], valueD = 0;

            long ticks = Environment.TickCount;
            Console.Write("Converting data...");

            if (dataMveArray.ElemType == typeof(Zcu.Mve.Numerics.Integer))
            {
                valueDataType = ValueDataType.Integer;
                dataI = new int[size0size1 * size[2]];

                unsafe
                {
                    fixed (int* dataPtr = &dataI[0])
                    {
                        for (int i = 0; i < size[0]; i++)
                        {
                            index = i;
                            for (int j = 0; j < size[1]; j++)
                            {
                                index = i + j * size[0];
                                for (int k = 0; k < size[2]; k++)
                                {
                                    valueI = (Integer)dataMveArray[i, j, k];

                                    if (findMinMax)
                                    {
                                        if (valueI < min)
                                            min = valueI;
                                        else if (valueI > max)
                                            max = valueI;
                                    }
                                    dataPtr[index + k * size0size1] = valueI;
                                }
                            }
                            ProgressInfo(0.2 * (i + 1) / size0);
                        }
                    }
                }
            }
            else if (dataMveArray.ElemType == typeof(Zcu.Mve.Numerics.Scalar))
            {
                valueDataType = ValueDataType.Scalar;
                dataD = new double[size0size1 * size[2]];

                unsafe
                {
                    fixed (double* dataPtr = &dataD[0])
                    {
                        for (int i = 0; i < size[0]; i++)
                        {
                            index = i;
                            for (int j = 0; j < size[1]; j++)
                            {
                                index = i + j * size[0];
                                for (int k = 0; k < size[2]; k++)
                                {
                                    valueD = (Scalar)dataMveArray[i, j, k];

                                    if (findMinMax)
                                    {
                                        if (valueD < min)
                                            min = valueD;
                                        else if (valueD > max)
                                            max = valueD;
                                    }
                                    dataPtr[index + k * size0size1] = valueD;
                                }
                            }
                            ProgressInfo(0.2 * (i + 1) / size0);
                        }
                    }
                }
            }
            Console.WriteLine("\t\tDone ({0} ms)", Environment.TickCount - ticks);

            dataMveArray = null;
            inData = null;

            IsoSurfaceExtractor.GCClean();
        }//*/

        /// <summary>
        /// This method makes cubes from values in two neghbouring slices. The convention of cube´s vertices numbering is shown below. 
        /// After each cube is created, it is commited to the next step - checked whether it is intersected by searched iso-surface and eventually appropriate triangles are generated.
        ///   4 ------ 5
        ///  /|       /|
        /// 7 ------ 6 |
        /// | |      | |
        /// | 0 -----| 1
        /// |/       |/ 
        /// 3 ------ 2  
        /// </summary> 
        protected virtual void DecomposeVolumeIntoCubes()
        {
            Console.Write("Extracting iso-surface...");
            long ticks = Environment.TickCount;

            double progress = firstRunForThisData ? 0.7 / size[0] : 0.9 / size[0];
            int size0size1 = size[0] * size[1], i1 = 0, j1 = 0, k1 = 0;
            int ysize0 = 0, y1size0 = 0, zsize0size1 = 0, z1size0size1 = 0;
            for (int i = 0; i < size[0] - 1; i++)
            {
                i1 = i + 1;
                for (int j = 0; j < size[1] - 1; j++)
                {
                    j1 = j + 1;
                    ysize0 = j * size[0];
                    y1size0 = j1 * size[0];
                    for (int k = 0; k < size[2] - 1; k++)
                    {
                        k1 = k + 1;
                        zsize0size1 = k * size0size1;
                        z1size0size1 = k1 * size0size1;

                        if (valueDataType == ValueDataType.Integer)
                            cube = CreateCubeI(i, j, k, i1, j1, k1, ysize0, y1size0, size0size1, zsize0size1, z1size0size1);
                        else if (valueDataType == ValueDataType.Scalar)
                            cube = CreateCubeD(i, j, k, i1, j1, k1, ysize0, y1size0, size0size1, zsize0size1, z1size0size1);

                        GenerateTriangles(cube, i, j, k);
                    }
                }
                ProgressInfo(firstRunForThisData ? 0.2 + ((i + 1) * progress) : ((i + 1) * progress));// */
            }

            Console.WriteLine("\tDone ({0} ms)", Environment.TickCount - ticks);
        }


        #region ***** Integer data specialized methods *****
        /// <summary>
        /// Creates a new integer cube.
        /// </summary>
        /// <param name="i">i</param>
        /// <param name="j">j</param>
        /// <param name="k">k</param>
        /// <param name="i1">i+1</param>
        /// <param name="j1">j+1</param>
        /// <param name="k1">k+1</param>
        /// <param name="ysize0">j * size</param>
        /// <param name="y1size0">(j+1) * size</param>
        /// <param name="size0size1">size</param>
        /// <param name="zsize0size1">k * size</param>
        /// <param name="z1size0size1">(k+1) * size</param>
        /// <returns>integer cube</returns>
        protected virtual Cube CreateCubeI(int i, int j, int k, int i1, int j1, int k1, int ysize0, int y1size0, int size0size1, int zsize0size1, int z1size0size1)
        {
            int x, y, z;
            if ((i > 0) && (j > 0) && (k > 0)) // take shared vertices form previous cube
            {
                cube.p[0] = previousCube.p[3];

                cube.p[1] = previousCube.p[2];

                x = i1; y = j; z = k1;
                cube.p[2].coords.X = x; cube.p[2].coords.Y = y; cube.p[2].coords.Z = z;
                cube.p[2].index = x + ysize0 + z1size0size1;
                cube.p[2].value = dataI[cube.p[2].index];

                x = i; y = j; z = k1;
                cube.p[3].coords.X = x; cube.p[3].coords.Y = y; cube.p[3].coords.Z = z;
                cube.p[3].index = x + ysize0 + z1size0size1;
                cube.p[3].value = dataI[cube.p[3].index];

                cube.p[4] = previousCube.p[7];

                cube.p[5] = previousCube.p[6];

                x = i1; y = j1; z = k1;
                cube.p[6].coords.X = x; cube.p[6].coords.Y = y; cube.p[6].coords.Z = z;
                cube.p[6].index = x + y1size0 + z1size0size1;
                cube.p[6].value = dataI[cube.p[6].index];

                x = i; y = j1; z = k1;
                cube.p[7].coords.X = x; cube.p[7].coords.Y = y; cube.p[7].coords.Z = z;
                cube.p[7].index = x + y1size0 + z1size0size1;
                cube.p[7].value = dataI[cube.p[7].index];
            }
            else
            {  // assign coordinates and values to each vertex of the cube
                x = i; y = j; z = k;
                cube.p[0].coords.X = x; cube.p[0].coords.Y = y; cube.p[0].coords.Z = z;
                cube.p[0].index = x + ysize0 + zsize0size1;
                cube.p[0].value = dataI[cube.p[0].index];

                x = i1; y = j; z = k;
                cube.p[1].coords.X = x; cube.p[1].coords.Y = y; cube.p[1].coords.Z = z;
                cube.p[1].index = x + ysize0 + zsize0size1;
                cube.p[1].value = dataI[cube.p[1].index];

                x = i1; y = j; z = k1;
                cube.p[2].coords.X = x; cube.p[2].coords.Y = y; cube.p[2].coords.Z = z;
                cube.p[2].index = x + ysize0 + z1size0size1;
                cube.p[2].value = dataI[cube.p[2].index];

                x = i; y = j; z = k1;
                cube.p[3].coords.X = x; cube.p[3].coords.Y = y; cube.p[3].coords.Z = z;
                cube.p[3].index = x + ysize0 + z1size0size1;
                cube.p[3].value = dataI[cube.p[3].index];

                x = i; y = j1; z = k;
                cube.p[4].coords.X = x; cube.p[4].coords.Y = y; cube.p[4].coords.Z = z;
                cube.p[4].index = x + y1size0 + zsize0size1;
                cube.p[4].value = dataI[cube.p[4].index];

                x = i1; y = j1; z = k;
                cube.p[5].coords.X = x; cube.p[5].coords.Y = y; cube.p[5].coords.Z = z;
                cube.p[5].index = x + y1size0 + zsize0size1;
                cube.p[5].value = dataI[cube.p[5].index];

                x = i1; y = j1; z = k1;
                cube.p[6].coords.X = x; cube.p[6].coords.Y = y; cube.p[6].coords.Z = z;
                cube.p[6].index = x + y1size0 + z1size0size1;
                cube.p[6].value = dataI[cube.p[6].index];

                x = i; y = j1; z = k1;
                cube.p[7].coords.X = x; cube.p[7].coords.Y = y; cube.p[7].coords.Z = z;
                cube.p[7].index = x + y1size0 + z1size0size1;
                cube.p[7].value = dataI[cube.p[7].index];
            }

            return cube;
        }

        /// <summary> Calculate normal. </summary>
        /// <param name="p">Point where we want calculate normal.</param>
        /// <param name="data">Loaded volume data.</param>
        /// <param name="size">Dimensions of input volume data array.</param>
        /// <returns>Normal.</returns>
        protected Vector3D CalcNormalI(Vertex p, int[] data, int[] size)
        {
            int i = (int)p.coords.X, j = (int)p.coords.Y, k = (int)p.coords.Z;
            int ksize0size1 = k * size[0] * size[1], size0size1 = size[0] * size[1], jsize0 = j * size[0];
            Vector3D norm = new Vector3D(0, 0, 0);

            //    norm.X = data[x + 1, y, z] - data[x - 1, y, z];
            //    norm.Y = data[x, y + 1, z] - data[x, y - 1, z];
            //    norm.Z = data[x, y, z + 1] - data[x, y, z - 1];

            // watchout the index in array
            if ((i > 0) && (i < size[0] - 1))
                norm.X = data[(i + 1) + jsize0 + ksize0size1] - data[(i - 1) + jsize0 + ksize0size1];

            if ((j > 0) && (j < size[1] - 1))
                norm.Y = data[i + (j + 1) * size[0] + ksize0size1] - data[i + (j - 1) * size[0] + ksize0size1];

            if ((k > 0) && (k < size[2] - 1))
                norm.Z = data[i + jsize0 + (k + 1) * size0size1] - data[i + jsize0 + (k - 1) * size0size1];

            return norm;
        } // of CalcNormalI   
        #endregion

        #region ***** Scalar data specialized methods *****
        /// <summary>
        /// Creates a new double cube.
        /// </summary>
        /// <param name="i">i</param>
        /// <param name="j">j</param>
        /// <param name="k">k</param>
        /// <param name="i1">i+1</param>
        /// <param name="j1">j+1</param>
        /// <param name="k1">k+1</param>
        /// <param name="ysize0">j * size</param>
        /// <param name="y1size0">(j+1) * size</param>
        /// <param name="size0size1">size</param>
        /// <param name="zsize0size1">k * size</param>
        /// <param name="z1size0size1">(k+1) * size</param>
        /// <returns>double cube</returns>
        protected virtual Cube CreateCubeD(int i, int j, int k, int i1, int j1, int k1, int ysize0, int y1size0, int size0size1, int zsize0size1, int z1size0size1)
        {
            int x, y, z;
            if ((i > 0) && (j > 0) && (k > 0)) // take shared vertices form previous cube
            {
                cube.p[0] = previousCube.p[3];

                cube.p[1] = previousCube.p[2];

                x = i1; y = j; z = k1;
                cube.p[2].coords.X = x; cube.p[2].coords.Y = y; cube.p[2].coords.Z = z;
                cube.p[2].index = x + ysize0 + z1size0size1;
                cube.p[2].value = dataD[cube.p[2].index];

                x = i; y = j; z = k1;
                cube.p[3].coords.X = x; cube.p[3].coords.Y = y; cube.p[3].coords.Z = z;
                cube.p[3].index = x + ysize0 + z1size0size1;
                cube.p[3].value = dataD[cube.p[3].index];

                cube.p[4] = previousCube.p[7];

                cube.p[5] = previousCube.p[6];

                x = i1; y = j1; z = k1;
                cube.p[6].coords.X = x; cube.p[6].coords.Y = y; cube.p[6].coords.Z = z;
                cube.p[6].index = x + y1size0 + z1size0size1;
                cube.p[6].value = dataD[cube.p[6].index];

                x = i; y = j1; z = k1;
                cube.p[7].coords.X = x; cube.p[7].coords.Y = y; cube.p[7].coords.Z = z;
                cube.p[7].index = x + y1size0 + z1size0size1;
                cube.p[7].value = dataD[cube.p[7].index];
            }
            else
            {  // assign coordinates and values to each vertex of the cube
                x = i; y = j; z = k;
                cube.p[0].coords.X = x; cube.p[0].coords.Y = y; cube.p[0].coords.Z = z;
                cube.p[0].index = x + ysize0 + zsize0size1;
                cube.p[0].value = dataD[cube.p[0].index];

                x = i1; y = j; z = k;
                cube.p[1].coords.X = x; cube.p[1].coords.Y = y; cube.p[1].coords.Z = z;
                cube.p[1].index = x + ysize0 + zsize0size1;
                cube.p[1].value = dataD[cube.p[1].index];

                x = i1; y = j; z = k1;
                cube.p[2].coords.X = x; cube.p[2].coords.Y = y; cube.p[2].coords.Z = z;
                cube.p[2].index = x + ysize0 + z1size0size1;
                cube.p[2].value = dataD[cube.p[2].index];

                x = i; y = j; z = k1;
                cube.p[3].coords.X = x; cube.p[3].coords.Y = y; cube.p[3].coords.Z = z;
                cube.p[3].index = x + ysize0 + z1size0size1;
                cube.p[3].value = dataD[cube.p[3].index];

                x = i; y = j1; z = k;
                cube.p[4].coords.X = x; cube.p[4].coords.Y = y; cube.p[4].coords.Z = z;
                cube.p[4].index = x + y1size0 + zsize0size1;
                cube.p[4].value = dataD[cube.p[4].index];

                x = i1; y = j1; z = k;
                cube.p[5].coords.X = x; cube.p[5].coords.Y = y; cube.p[5].coords.Z = z;
                cube.p[5].index = x + y1size0 + zsize0size1;
                cube.p[5].value = dataD[cube.p[5].index];

                x = i1; y = j1; z = k1;
                cube.p[6].coords.X = x; cube.p[6].coords.Y = y; cube.p[6].coords.Z = z;
                cube.p[6].index = x + y1size0 + z1size0size1;
                cube.p[6].value = dataD[cube.p[6].index];

                x = i; y = j1; z = k1;
                cube.p[7].coords.X = x; cube.p[7].coords.Y = y; cube.p[7].coords.Z = z;
                cube.p[7].index = x + y1size0 + z1size0size1;
                cube.p[7].value = dataD[cube.p[7].index];
            }

            return cube;
        }//*/

        /// <summary> Calculate normal. </summary>
        /// <param name="p">Point where we want calculate normal.</param>
        /// <param name="data">Loaded volume data.</param>
        /// <param name="size">Dimensions of input volume data array.</param>
        /// <returns>Normal.</returns>
        protected Vector3D CalcNormalD(Vertex p, double[] data, int[] size)
        {
            int i = (int)p.coords.X, j = (int)p.coords.Y, k = (int)p.coords.Z;
            int ksize0size1 = k * size[0] * size[1], size0size1 = size[0] * size[1], jsize0 = j * size[0];
            Vector3D norm = new Vector3D(0, 0, 0);

            //    norm.X = data[x + 1, y, z] - data[x - 1, y, z];
            //    norm.Y = data[x, y + 1, z] - data[x, y - 1, z];
            //    norm.Z = data[x, y, z + 1] - data[x, y, z - 1];

            // watchout the index in array            
            if ((i > 0) && (i < size[0] - 1))
                norm.X = data[(i + 1) + jsize0 + ksize0size1] - data[(i - 1) + jsize0 + ksize0size1];

            if ((j > 0) && (j < size[1] - 1))
                norm.Y = data[i + (j + 1) * size[0] + ksize0size1] - data[i + (j - 1) * size[0] + ksize0size1];

            if ((k > 0) && (k < size[2] - 1))
                norm.Z = data[i + jsize0 + (k + 1) * size0size1] - data[i + jsize0 + (k - 1) * size0size1];

            return norm;
        } // of CalcNormalD    
        #endregion

        /// <summary> This method generates triangles for each cell of the volume. </summary>
        /// <param name="cube">Cube to process.</param>
        /// <param name="i">x-axis index</param>
        /// <param name="j">y-axis index</param>
        /// <param name="k">z-axis index</param>
        protected abstract void GenerateTriangles(Cube cube, int i, int j, int k);

        /// <summary> Calculate new vertex. Interpolate point position and normal value. We can choose whether to use linear interpolation or the calculated point will be positioned in the middle of the input points. </summary>
        /// <param name="a">First input point.</param>
        /// <param name="b">Second input point.</param>
        /// <param name="isoValue">Iso value.</param>
        /// <param name="inverseDistortion">Distortion values powered by -1.</param>
        /// <param name="linearInterpolation">Use linear interpolation?</param>
        /// <returns>New vertex.</returns>
        protected Vertex Interpolate(Vertex a, Vertex b, double isoValue, double[] inverseDistortion, bool linearInterpolation)
        {
            if (linearInterpolation)
            {
                double val = (isoValue - a.value) / (b.value - a.value);

                a.coords.X = (a.coords.X + val * (b.coords.X - a.coords.X)) * inverseDistortion[0];
                a.coords.Y = (a.coords.Y + val * (b.coords.Y - a.coords.Y)) * inverseDistortion[1];
                a.coords.Z = (a.coords.Z + val * (b.coords.Z - a.coords.Z)) * inverseDistortion[2];

                a.normal.X = a.normal.X + val * (b.normal.X - a.normal.X);
                a.normal.Y = a.normal.Y + val * (b.normal.Y - a.normal.Y);
                a.normal.Z = a.normal.Z + val * (b.normal.Z - a.normal.Z);
            }
            else
            {
                a.coords.X = (a.coords.X + 0.5 * (b.coords.X - a.coords.X)) * inverseDistortion[0];
                a.coords.Y = (a.coords.Y + 0.5 * (b.coords.Y - a.coords.Y)) * inverseDistortion[1];
                a.coords.Z = (a.coords.Z + 0.5 * (b.coords.Z - a.coords.Z)) * inverseDistortion[2];

                a.normal.X = a.normal.X + 0.5 * (b.normal.X - a.normal.X);
                a.normal.Y = a.normal.Y + 0.5 * (b.normal.Y - a.normal.Y);
                a.normal.Z = a.normal.Z + 0.5 * (b.normal.Z - a.normal.Z);
            }

            return a;
        } // of Interpolate

        /// <summary>  Creates from computed data output data </summary>
        protected virtual void CreateOutput()
        {
            mesh = new TriangleMesh();
            points = new UniformDataArray(typeof(Point3D), pts.Count);
            for (int i = 0; i < pts.Count; i++)
                points[i] = new Point3D(pts[i].coords.X, pts[i].coords.Y, pts[i].coords.Z);
            mesh.Points = points;
            points = null;
            ProgressInfo(0.9f);

            normals = new UniformDataArray(typeof(Normal3D), pts.Count);
            for (int i = 0; i < pts.Count; i++)
                normals[i] = new Normal3D(pts[i].normal.X, pts[i].normal.Y, pts[i].normal.Z);
            mesh.AddPointAttrs(normalAttributeName, normals);
            pts = null;
            normals = null;
            ProgressInfo(0.95f);

            triangles = new UniformDataArray(typeof(Triangle), indices.Count);
            for (int i = 0; i < indices.Count; i++)
                triangles[i] = indices[i];
            mesh.AddCells("Triangles", triangles);

            if ((triangles.Count == 0) || (indices.Count == 0))
                throw new MveException(this, "No isosurface extracted! Try some other threshold value.");

            triangles = null;
            indices = null;
            vertList = null;
            indexList = null;
        }

        /// <summary> Sequence of commands, allowing better usage of garbage collector. </summary>
        public static void GCClean()
        {
            GC.Collect();
            GC.WaitForPendingFinalizers();
            GC.Collect();
            Thread.Sleep(10);
        } // GCClean   


        #region Properties
        /// <summary>
        /// Property for threshold variable.
        /// </summary>
        [CategoryAttribute("Extraction"), Description("Threshold of iso-surface value in percents. \nIf 'useThresholdAsIsoValue' flag is set to True,"
        + " is this value directly used as iso value to be searched (used for implicit surfaces visualization).")]
        public double Threshold
        {
            get { return (threshold); }
            set
            {
                threshold = value;
                paramsChanged = true;
            }
        }

        /// <summary>
        /// Property for linearInterpolation variable.
        /// </summary>
        [CategoryAttribute("Extraction"), Description("Interpolation type:\n  true = linear interpolation\n  false = center point.")]
        public bool LinearInterpolation
        {
            get { return (linearInterpolation); }
            set
            {
                linearInterpolation = value;
                paramsChanged = true;
            }
        }

        /// <summary>
        /// Set name of RegGridND point attributes. Default set to Values.
        /// </summary>
        [CategoryAttribute("Extraction"), Description("Set name of RegGridND point attributes. Default set to \"Values\"")]
        public string PointAttributeName
        {
            set
            {
                if (value == string.Empty)
                {
                    throw new MveException(this, "Attribut property cannot be set to an empty string!");
                }
                else
                {
                    pointAttributesName = value;
                    paramsChanged = true;
                }
            }

            get { return pointAttributesName; }
        }

        /// <summary>
        /// Set name of RegGridND normal attribute. Default set to Values.
        /// </summary>
        [CategoryAttribute("Extraction"), Description("Set name of normals point attribute. Default set to \"Normal3D\"")]
        public string NormalAttributeName
        {
            set
            {
                if (value == string.Empty)
                {
                    throw new MveException(this, "Attribut property cannot be set to an empty string!");
                }
                else
                {
                    normalAttributeName = value;
                    paramsChanged = true;
                }
            }
            get { return normalAttributeName; }
        }

        /// <summary>
        /// Property for linearInterpolation variable.
        /// </summary>
        [CategoryAttribute("Extraction"), Description("Specifies whether to use input threshold as percentual value from min/max interval or directly as iso value.")]
        public bool UseThresholdAsIsoValue
        {
            get { return (useThresholdAsIsoValue); }
            set
            {
                useThresholdAsIsoValue = value;
                paramsChanged = true;
            }
        }
        #endregion


        # region Lookup tables
        /// <summary>
        /// Edge lookup table.
        /// </summary>
        protected int[] edgeTable = new int[256]{
            0x0  , 0x109, 0x203, 0x30a, 0x406, 0x50f, 0x605, 0x70c, 0x80c, 0x905, 0xa0f, 0xb06,
            0xc0a, 0xd03, 0xe09, 0xf00, 0x190, 0x99 , 0x393, 0x29a, 0x596, 0x49f, 0x795, 0x69c, 
            0x99c, 0x895, 0xb9f, 0xa96, 0xd9a, 0xc93, 0xf99, 0xe90, 0x230, 0x339, 0x33 , 0x13a,
            0x636, 0x73f, 0x435, 0x53c, 0xa3c, 0xb35, 0x83f, 0x936, 0xe3a, 0xf33, 0xc39, 0xd30,
            0x3a0, 0x2a9, 0x1a3, 0xaa , 0x7a6, 0x6af, 0x5a5, 0x4ac, 0xbac, 0xaa5, 0x9af, 0x8a6, 
            0xfaa, 0xea3, 0xda9, 0xca0, 0x460, 0x569, 0x663, 0x76a, 0x66 , 0x16f, 0x265, 0x36c,
            0xc6c, 0xd65, 0xe6f, 0xf66, 0x86a, 0x963, 0xa69, 0xb60, 0x5f0, 0x4f9, 0x7f3, 0x6fa,
            0x1f6, 0xff , 0x3f5, 0x2fc, 0xdfc, 0xcf5, 0xfff, 0xef6, 0x9fa, 0x8f3, 0xbf9, 0xaf0,
            0x650, 0x759, 0x453, 0x55a, 0x256, 0x35f, 0x55 , 0x15c, 0xe5c, 0xf55, 0xc5f, 0xd56,
            0xa5a, 0xb53, 0x859, 0x950, 0x7c0, 0x6c9, 0x5c3, 0x4ca, 0x3c6, 0x2cf, 0x1c5, 0xcc ,
            0xfcc, 0xec5, 0xdcf, 0xcc6, 0xbca, 0xac3, 0x9c9, 0x8c0, 0x8c0, 0x9c9, 0xac3, 0xbca,
            0xcc6, 0xdcf, 0xec5, 0xfcc, 0xcc , 0x1c5, 0x2cf, 0x3c6, 0x4ca, 0x5c3, 0x6c9, 0x7c0,
            0x950, 0x859, 0xb53, 0xa5a, 0xd56, 0xc5f, 0xf55, 0xe5c, 0x15c, 0x55 , 0x35f, 0x256,
            0x55a, 0x453, 0x759, 0x650, 0xaf0, 0xbf9, 0x8f3, 0x9fa, 0xef6, 0xfff, 0xcf5, 0xdfc,
            0x2fc, 0x3f5, 0xff , 0x1f6, 0x6fa, 0x7f3, 0x4f9, 0x5f0, 0xb60, 0xa69, 0x963, 0x86a,
            0xf66, 0xe6f, 0xd65, 0xc6c, 0x36c, 0x265, 0x16f, 0x66 , 0x76a, 0x663, 0x569, 0x460,
            0xca0, 0xda9, 0xea3, 0xfaa, 0x8a6, 0x9af, 0xaa5, 0xbac, 0x4ac, 0x5a5, 0x6af, 0x7a6,
            0xaa , 0x1a3, 0x2a9, 0x3a0, 0xd30, 0xc39, 0xf33, 0xe3a, 0x936, 0x83f, 0xb35, 0xa3c,
            0x53c, 0x435, 0x73f, 0x636, 0x13a, 0x33 , 0x339, 0x230, 0xe90, 0xf99, 0xc93, 0xd9a,
            0xa96, 0xb9f, 0x895, 0x99c, 0x69c, 0x795, 0x49f, 0x596, 0x29a, 0x393, 0x99 , 0x190,
            0xf00, 0xe09, 0xd03, 0xc0a, 0xb06, 0xa0f, 0x905, 0x80c, 0x70c, 0x605, 0x50f, 0x406,
            0x30a, 0x203, 0x109, 0x0   }; // of edgeTable

        /// <summary>
        /// Triangle lookup table.
        /// </summary>
        protected int[,] triTable = new int[256, 16] {
            {-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
            {0, 8, 3, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
            {0, 1, 9, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
            {1, 8, 3, 9, 8, 1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
            {1, 2, 10, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
            {0, 8, 3, 1, 2, 10, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
            {9, 2, 10, 0, 2, 9, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
            {2, 8, 3, 2, 10, 8, 10, 9, 8, -1, -1, -1, -1, -1, -1, -1},
            {3, 11, 2, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
            {0, 11, 2, 8, 11, 0, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
            {1, 9, 0, 2, 3, 11, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
            {1, 11, 2, 1, 9, 11, 9, 8, 11, -1, -1, -1, -1, -1, -1, -1},
            {3, 10, 1, 11, 10, 3, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
            {0, 10, 1, 0, 8, 10, 8, 11, 10, -1, -1, -1, -1, -1, -1, -1},
            {3, 9, 0, 3, 11, 9, 11, 10, 9, -1, -1, -1, -1, -1, -1, -1},
            {9, 8, 10, 10, 8, 11, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
            {4, 7, 8, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
            {4, 3, 0, 7, 3, 4, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
            {0, 1, 9, 8, 4, 7, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
            {4, 1, 9, 4, 7, 1, 7, 3, 1, -1, -1, -1, -1, -1, -1, -1},
            {1, 2, 10, 8, 4, 7, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
            {3, 4, 7, 3, 0, 4, 1, 2, 10, -1, -1, -1, -1, -1, -1, -1},
            {9, 2, 10, 9, 0, 2, 8, 4, 7, -1, -1, -1, -1, -1, -1, -1},
            {2, 10, 9, 2, 9, 7, 2, 7, 3, 7, 9, 4, -1, -1, -1, -1},
            {8, 4, 7, 3, 11, 2, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
            {11, 4, 7, 11, 2, 4, 2, 0, 4, -1, -1, -1, -1, -1, -1, -1},
            {9, 0, 1, 8, 4, 7, 2, 3, 11, -1, -1, -1, -1, -1, -1, -1},
            {4, 7, 11, 9, 4, 11, 9, 11, 2, 9, 2, 1, -1, -1, -1, -1},
            {3, 10, 1, 3, 11, 10, 7, 8, 4, -1, -1, -1, -1, -1, -1, -1},
            {1, 11, 10, 1, 4, 11, 1, 0, 4, 7, 11, 4, -1, -1, -1, -1},
            {4, 7, 8, 9, 0, 11, 9, 11, 10, 11, 0, 3, -1, -1, -1, -1},
            {4, 7, 11, 4, 11, 9, 9, 11, 10, -1, -1, -1, -1, -1, -1, -1},
            {9, 5, 4, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
            {9, 5, 4, 0, 8, 3, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
            {0, 5, 4, 1, 5, 0, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
            {8, 5, 4, 8, 3, 5, 3, 1, 5, -1, -1, -1, -1, -1, -1, -1},
            {1, 2, 10, 9, 5, 4, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
            {3, 0, 8, 1, 2, 10, 4, 9, 5, -1, -1, -1, -1, -1, -1, -1},
            {5, 2, 10, 5, 4, 2, 4, 0, 2, -1, -1, -1, -1, -1, -1, -1},
            {2, 10, 5, 3, 2, 5, 3, 5, 4, 3, 4, 8, -1, -1, -1, -1},
            {9, 5, 4, 2, 3, 11, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
            {0, 11, 2, 0, 8, 11, 4, 9, 5, -1, -1, -1, -1, -1, -1, -1},
            {0, 5, 4, 0, 1, 5, 2, 3, 11, -1, -1, -1, -1, -1, -1, -1},
            {2, 1, 5, 2, 5, 8, 2, 8, 11, 4, 8, 5, -1, -1, -1, -1},
            {10, 3, 11, 10, 1, 3, 9, 5, 4, -1, -1, -1, -1, -1, -1, -1},
            {4, 9, 5, 0, 8, 1, 8, 10, 1, 8, 11, 10, -1, -1, -1, -1},
            {5, 4, 0, 5, 0, 11, 5, 11, 10, 11, 0, 3, -1, -1, -1, -1},
            {5, 4, 8, 5, 8, 10, 10, 8, 11, -1, -1, -1, -1, -1, -1, -1},
            {9, 7, 8, 5, 7, 9, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
            {9, 3, 0, 9, 5, 3, 5, 7, 3, -1, -1, -1, -1, -1, -1, -1},
            {0, 7, 8, 0, 1, 7, 1, 5, 7, -1, -1, -1, -1, -1, -1, -1},
            {1, 5, 3, 3, 5, 7, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
            {9, 7, 8, 9, 5, 7, 10, 1, 2, -1, -1, -1, -1, -1, -1, -1},
            {10, 1, 2, 9, 5, 0, 5, 3, 0, 5, 7, 3, -1, -1, -1, -1},
            {8, 0, 2, 8, 2, 5, 8, 5, 7, 10, 5, 2, -1, -1, -1, -1},
            {2, 10, 5, 2, 5, 3, 3, 5, 7, -1, -1, -1, -1, -1, -1, -1},
            {7, 9, 5, 7, 8, 9, 3, 11, 2, -1, -1, -1, -1, -1, -1, -1},
            {9, 5, 7, 9, 7, 2, 9, 2, 0, 2, 7, 11, -1, -1, -1, -1},
            {2, 3, 11, 0, 1, 8, 1, 7, 8, 1, 5, 7, -1, -1, -1, -1},
            {11, 2, 1, 11, 1, 7, 7, 1, 5, -1, -1, -1, -1, -1, -1, -1},
            {9, 5, 8, 8, 5, 7, 10, 1, 3, 10, 3, 11, -1, -1, -1, -1},
            {5, 7, 0, 5, 0, 9, 7, 11, 0, 1, 0, 10, 11, 10, 0, -1},
            {11, 10, 0, 11, 0, 3, 10, 5, 0, 8, 0, 7, 5, 7, 0, -1},
            {11, 10, 5, 7, 11, 5, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
            {10, 6, 5, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
            {0, 8, 3, 5, 10, 6, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
            {9, 0, 1, 5, 10, 6, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
            {1, 8, 3, 1, 9, 8, 5, 10, 6, -1, -1, -1, -1, -1, -1, -1},
            {1, 6, 5, 2, 6, 1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
            {1, 6, 5, 1, 2, 6, 3, 0, 8, -1, -1, -1, -1, -1, -1, -1},
            {9, 6, 5, 9, 0, 6, 0, 2, 6, -1, -1, -1, -1, -1, -1, -1},
            {5, 9, 8, 5, 8, 2, 5, 2, 6, 3, 2, 8, -1, -1, -1, -1},
            {2, 3, 11, 10, 6, 5, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
            {11, 0, 8, 11, 2, 0, 10, 6, 5, -1, -1, -1, -1, -1, -1, -1},
            {0, 1, 9, 2, 3, 11, 5, 10, 6, -1, -1, -1, -1, -1, -1, -1},
            {5, 10, 6, 1, 9, 2, 9, 11, 2, 9, 8, 11, -1, -1, -1, -1},
            {6, 3, 11, 6, 5, 3, 5, 1, 3, -1, -1, -1, -1, -1, -1, -1},
            {0, 8, 11, 0, 11, 5, 0, 5, 1, 5, 11, 6, -1, -1, -1, -1},
            {3, 11, 6, 0, 3, 6, 0, 6, 5, 0, 5, 9, -1, -1, -1, -1},
            {6, 5, 9, 6, 9, 11, 11, 9, 8, -1, -1, -1, -1, -1, -1, -1},
            {5, 10, 6, 4, 7, 8, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
            {4, 3, 0, 4, 7, 3, 6, 5, 10, -1, -1, -1, -1, -1, -1, -1},
            {1, 9, 0, 5, 10, 6, 8, 4, 7, -1, -1, -1, -1, -1, -1, -1},
            {10, 6, 5, 1, 9, 7, 1, 7, 3, 7, 9, 4, -1, -1, -1, -1},
            {6, 1, 2, 6, 5, 1, 4, 7, 8, -1, -1, -1, -1, -1, -1, -1},
            {1, 2, 5, 5, 2, 6, 3, 0, 4, 3, 4, 7, -1, -1, -1, -1},
            {8, 4, 7, 9, 0, 5, 0, 6, 5, 0, 2, 6, -1, -1, -1, -1},
            {7, 3, 9, 7, 9, 4, 3, 2, 9, 5, 9, 6, 2, 6, 9, -1},
            {3, 11, 2, 7, 8, 4, 10, 6, 5, -1, -1, -1, -1, -1, -1, -1},
            {5, 10, 6, 4, 7, 2, 4, 2, 0, 2, 7, 11, -1, -1, -1, -1},
            {0, 1, 9, 4, 7, 8, 2, 3, 11, 5, 10, 6, -1, -1, -1, -1},
            {9, 2, 1, 9, 11, 2, 9, 4, 11, 7, 11, 4, 5, 10, 6, -1},
            {8, 4, 7, 3, 11, 5, 3, 5, 1, 5, 11, 6, -1, -1, -1, -1},
            {5, 1, 11, 5, 11, 6, 1, 0, 11, 7, 11, 4, 0, 4, 11, -1},
            {0, 5, 9, 0, 6, 5, 0, 3, 6, 11, 6, 3, 8, 4, 7, -1},
            {6, 5, 9, 6, 9, 11, 4, 7, 9, 7, 11, 9, -1, -1, -1, -1},
            {10, 4, 9, 6, 4, 10, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
            {4, 10, 6, 4, 9, 10, 0, 8, 3, -1, -1, -1, -1, -1, -1, -1},
            {10, 0, 1, 10, 6, 0, 6, 4, 0, -1, -1, -1, -1, -1, -1, -1},
            {8, 3, 1, 8, 1, 6, 8, 6, 4, 6, 1, 10, -1, -1, -1, -1},
            {1, 4, 9, 1, 2, 4, 2, 6, 4, -1, -1, -1, -1, -1, -1, -1},
            {3, 0, 8, 1, 2, 9, 2, 4, 9, 2, 6, 4, -1, -1, -1, -1},
            {0, 2, 4, 4, 2, 6, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
            {8, 3, 2, 8, 2, 4, 4, 2, 6, -1, -1, -1, -1, -1, -1, -1},
            {10, 4, 9, 10, 6, 4, 11, 2, 3, -1, -1, -1, -1, -1, -1, -1},
            {0, 8, 2, 2, 8, 11, 4, 9, 10, 4, 10, 6, -1, -1, -1, -1},
            {3, 11, 2, 0, 1, 6, 0, 6, 4, 6, 1, 10, -1, -1, -1, -1},
            {6, 4, 1, 6, 1, 10, 4, 8, 1, 2, 1, 11, 8, 11, 1, -1},
            {9, 6, 4, 9, 3, 6, 9, 1, 3, 11, 6, 3, -1, -1, -1, -1},
            {8, 11, 1, 8, 1, 0, 11, 6, 1, 9, 1, 4, 6, 4, 1, -1},
            {3, 11, 6, 3, 6, 0, 0, 6, 4, -1, -1, -1, -1, -1, -1, -1},
            {6, 4, 8, 11, 6, 8, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
            {7, 10, 6, 7, 8, 10, 8, 9, 10, -1, -1, -1, -1, -1, -1, -1},
            {0, 7, 3, 0, 10, 7, 0, 9, 10, 6, 7, 10, -1, -1, -1, -1},
            {10, 6, 7, 1, 10, 7, 1, 7, 8, 1, 8, 0, -1, -1, -1, -1},
            {10, 6, 7, 10, 7, 1, 1, 7, 3, -1, -1, -1, -1, -1, -1, -1},
            {1, 2, 6, 1, 6, 8, 1, 8, 9, 8, 6, 7, -1, -1, -1, -1},
            {2, 6, 9, 2, 9, 1, 6, 7, 9, 0, 9, 3, 7, 3, 9, -1},
            {7, 8, 0, 7, 0, 6, 6, 0, 2, -1, -1, -1, -1, -1, -1, -1},
            {7, 3, 2, 6, 7, 2, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
            {2, 3, 11, 10, 6, 8, 10, 8, 9, 8, 6, 7, -1, -1, -1, -1},
            {2, 0, 7, 2, 7, 11, 0, 9, 7, 6, 7, 10, 9, 10, 7, -1},
            {1, 8, 0, 1, 7, 8, 1, 10, 7, 6, 7, 10, 2, 3, 11, -1},
            {11, 2, 1, 11, 1, 7, 10, 6, 1, 6, 7, 1, -1, -1, -1, -1},
            {8, 9, 6, 8, 6, 7, 9, 1, 6, 11, 6, 3, 1, 3, 6, -1},
            {0, 9, 1, 11, 6, 7, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
            {7, 8, 0, 7, 0, 6, 3, 11, 0, 11, 6, 0, -1, -1, -1, -1},
            {7, 11, 6, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
            {7, 6, 11, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
            {3, 0, 8, 11, 7, 6, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
            {0, 1, 9, 11, 7, 6, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
            {8, 1, 9, 8, 3, 1, 11, 7, 6, -1, -1, -1, -1, -1, -1, -1},
            {10, 1, 2, 6, 11, 7, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
            {1, 2, 10, 3, 0, 8, 6, 11, 7, -1, -1, -1, -1, -1, -1, -1},
            {2, 9, 0, 2, 10, 9, 6, 11, 7, -1, -1, -1, -1, -1, -1, -1},
            {6, 11, 7, 2, 10, 3, 10, 8, 3, 10, 9, 8, -1, -1, -1, -1},
            {7, 2, 3, 6, 2, 7, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
            {7, 0, 8, 7, 6, 0, 6, 2, 0, -1, -1, -1, -1, -1, -1, -1},
            {2, 7, 6, 2, 3, 7, 0, 1, 9, -1, -1, -1, -1, -1, -1, -1},
            {1, 6, 2, 1, 8, 6, 1, 9, 8, 8, 7, 6, -1, -1, -1, -1},
            {10, 7, 6, 10, 1, 7, 1, 3, 7, -1, -1, -1, -1, -1, -1, -1},
            {10, 7, 6, 1, 7, 10, 1, 8, 7, 1, 0, 8, -1, -1, -1, -1},
            {0, 3, 7, 0, 7, 10, 0, 10, 9, 6, 10, 7, -1, -1, -1, -1},
            {7, 6, 10, 7, 10, 8, 8, 10, 9, -1, -1, -1, -1, -1, -1, -1},
            {6, 8, 4, 11, 8, 6, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
            {3, 6, 11, 3, 0, 6, 0, 4, 6, -1, -1, -1, -1, -1, -1, -1},
            {8, 6, 11, 8, 4, 6, 9, 0, 1, -1, -1, -1, -1, -1, -1, -1},
            {9, 4, 6, 9, 6, 3, 9, 3, 1, 11, 3, 6, -1, -1, -1, -1},
            {6, 8, 4, 6, 11, 8, 2, 10, 1, -1, -1, -1, -1, -1, -1, -1},
            {1, 2, 10, 3, 0, 11, 0, 6, 11, 0, 4, 6, -1, -1, -1, -1},
            {4, 11, 8, 4, 6, 11, 0, 2, 9, 2, 10, 9, -1, -1, -1, -1},
            {10, 9, 3, 10, 3, 2, 9, 4, 3, 11, 3, 6, 4, 6, 3, -1},
            {8, 2, 3, 8, 4, 2, 4, 6, 2, -1, -1, -1, -1, -1, -1, -1},
            {0, 4, 2, 4, 6, 2, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
            {1, 9, 0, 2, 3, 4, 2, 4, 6, 4, 3, 8, -1, -1, -1, -1},
            {1, 9, 4, 1, 4, 2, 2, 4, 6, -1, -1, -1, -1, -1, -1, -1},
            {8, 1, 3, 8, 6, 1, 8, 4, 6, 6, 10, 1, -1, -1, -1, -1},
            {10, 1, 0, 10, 0, 6, 6, 0, 4, -1, -1, -1, -1, -1, -1, -1},
            {4, 6, 3, 4, 3, 8, 6, 10, 3, 0, 3, 9, 10, 9, 3, -1},
            {10, 9, 4, 6, 10, 4, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
            {4, 9, 5, 7, 6, 11, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
            {0, 8, 3, 4, 9, 5, 11, 7, 6, -1, -1, -1, -1, -1, -1, -1},
            {5, 0, 1, 5, 4, 0, 7, 6, 11, -1, -1, -1, -1, -1, -1, -1},
            {11, 7, 6, 8, 3, 4, 3, 5, 4, 3, 1, 5, -1, -1, -1, -1},
            {9, 5, 4, 10, 1, 2, 7, 6, 11, -1, -1, -1, -1, -1, -1, -1},
            {6, 11, 7, 1, 2, 10, 0, 8, 3, 4, 9, 5, -1, -1, -1, -1},
            {7, 6, 11, 5, 4, 10, 4, 2, 10, 4, 0, 2, -1, -1, -1, -1},
            {3, 4, 8, 3, 5, 4, 3, 2, 5, 10, 5, 2, 11, 7, 6, -1},
            {7, 2, 3, 7, 6, 2, 5, 4, 9, -1, -1, -1, -1, -1, -1, -1},
            {9, 5, 4, 0, 8, 6, 0, 6, 2, 6, 8, 7, -1, -1, -1, -1},
            {3, 6, 2, 3, 7, 6, 1, 5, 0, 5, 4, 0, -1, -1, -1, -1},
            {6, 2, 8, 6, 8, 7, 2, 1, 8, 4, 8, 5, 1, 5, 8, -1},
            {9, 5, 4, 10, 1, 6, 1, 7, 6, 1, 3, 7, -1, -1, -1, -1},
            {1, 6, 10, 1, 7, 6, 1, 0, 7, 8, 7, 0, 9, 5, 4, -1},
            {4, 0, 10, 4, 10, 5, 0, 3, 10, 6, 10, 7, 3, 7, 10, -1},
            {7, 6, 10, 7, 10, 8, 5, 4, 10, 4, 8, 10, -1, -1, -1, -1},
            {6, 9, 5, 6, 11, 9, 11, 8, 9, -1, -1, -1, -1, -1, -1, -1},
            {3, 6, 11, 0, 6, 3, 0, 5, 6, 0, 9, 5, -1, -1, -1, -1},
            {0, 11, 8, 0, 5, 11, 0, 1, 5, 5, 6, 11, -1, -1, -1, -1},
            {6, 11, 3, 6, 3, 5, 5, 3, 1, -1, -1, -1, -1, -1, -1, -1},
            {1, 2, 10, 9, 5, 11, 9, 11, 8, 11, 5, 6, -1, -1, -1, -1},
            {0, 11, 3, 0, 6, 11, 0, 9, 6, 5, 6, 9, 1, 2, 10, -1},
            {11, 8, 5, 11, 5, 6, 8, 0, 5, 10, 5, 2, 0, 2, 5, -1},
            {6, 11, 3, 6, 3, 5, 2, 10, 3, 10, 5, 3, -1, -1, -1, -1},
            {5, 8, 9, 5, 2, 8, 5, 6, 2, 3, 8, 2, -1, -1, -1, -1},
            {9, 5, 6, 9, 6, 0, 0, 6, 2, -1, -1, -1, -1, -1, -1, -1},
            {1, 5, 8, 1, 8, 0, 5, 6, 8, 3, 8, 2, 6, 2, 8, -1},
            {1, 5, 6, 2, 1, 6, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
            {1, 3, 6, 1, 6, 10, 3, 8, 6, 5, 6, 9, 8, 9, 6, -1},
            {10, 1, 0, 10, 0, 6, 9, 5, 0, 5, 6, 0, -1, -1, -1, -1},
            {0, 3, 8, 5, 6, 10, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
            {10, 5, 6, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
            {11, 5, 10, 7, 5, 11, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
            {11, 5, 10, 11, 7, 5, 8, 3, 0, -1, -1, -1, -1, -1, -1, -1},
            {5, 11, 7, 5, 10, 11, 1, 9, 0, -1, -1, -1, -1, -1, -1, -1},
            {10, 7, 5, 10, 11, 7, 9, 8, 1, 8, 3, 1, -1, -1, -1, -1},
            {11, 1, 2, 11, 7, 1, 7, 5, 1, -1, -1, -1, -1, -1, -1, -1},
            {0, 8, 3, 1, 2, 7, 1, 7, 5, 7, 2, 11, -1, -1, -1, -1},
            {9, 7, 5, 9, 2, 7, 9, 0, 2, 2, 11, 7, -1, -1, -1, -1},
            {7, 5, 2, 7, 2, 11, 5, 9, 2, 3, 2, 8, 9, 8, 2, -1},
            {2, 5, 10, 2, 3, 5, 3, 7, 5, -1, -1, -1, -1, -1, -1, -1},
            {8, 2, 0, 8, 5, 2, 8, 7, 5, 10, 2, 5, -1, -1, -1, -1},
            {9, 0, 1, 5, 10, 3, 5, 3, 7, 3, 10, 2, -1, -1, -1, -1},
            {9, 8, 2, 9, 2, 1, 8, 7, 2, 10, 2, 5, 7, 5, 2, -1},
            {1, 3, 5, 3, 7, 5, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
            {0, 8, 7, 0, 7, 1, 1, 7, 5, -1, -1, -1, -1, -1, -1, -1},
            {9, 0, 3, 9, 3, 5, 5, 3, 7, -1, -1, -1, -1, -1, -1, -1},
            {9, 8, 7, 5, 9, 7, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
            {5, 8, 4, 5, 10, 8, 10, 11, 8, -1, -1, -1, -1, -1, -1, -1},
            {5, 0, 4, 5, 11, 0, 5, 10, 11, 11, 3, 0, -1, -1, -1, -1},
            {0, 1, 9, 8, 4, 10, 8, 10, 11, 10, 4, 5, -1, -1, -1, -1},
            {10, 11, 4, 10, 4, 5, 11, 3, 4, 9, 4, 1, 3, 1, 4, -1},
            {2, 5, 1, 2, 8, 5, 2, 11, 8, 4, 5, 8, -1, -1, -1, -1},
            {0, 4, 11, 0, 11, 3, 4, 5, 11, 2, 11, 1, 5, 1, 11, -1},
            {0, 2, 5, 0, 5, 9, 2, 11, 5, 4, 5, 8, 11, 8, 5, -1},
            {9, 4, 5, 2, 11, 3, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
            {2, 5, 10, 3, 5, 2, 3, 4, 5, 3, 8, 4, -1, -1, -1, -1},
            {5, 10, 2, 5, 2, 4, 4, 2, 0, -1, -1, -1, -1, -1, -1, -1},
            {3, 10, 2, 3, 5, 10, 3, 8, 5, 4, 5, 8, 0, 1, 9, -1},
            {5, 10, 2, 5, 2, 4, 1, 9, 2, 9, 4, 2, -1, -1, -1, -1},
            {8, 4, 5, 8, 5, 3, 3, 5, 1, -1, -1, -1, -1, -1, -1, -1},
            {0, 4, 5, 1, 0, 5, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
            {8, 4, 5, 8, 5, 3, 9, 0, 5, 0, 3, 5, -1, -1, -1, -1},
            {9, 4, 5, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
            {4, 11, 7, 4, 9, 11, 9, 10, 11, -1, -1, -1, -1, -1, -1, -1},
            {0, 8, 3, 4, 9, 7, 9, 11, 7, 9, 10, 11, -1, -1, -1, -1},
            {1, 10, 11, 1, 11, 4, 1, 4, 0, 7, 4, 11, -1, -1, -1, -1},
            {3, 1, 4, 3, 4, 8, 1, 10, 4, 7, 4, 11, 10, 11, 4, -1},
            {4, 11, 7, 9, 11, 4, 9, 2, 11, 9, 1, 2, -1, -1, -1, -1},
            {9, 7, 4, 9, 11, 7, 9, 1, 11, 2, 11, 1, 0, 8, 3, -1},
            {11, 7, 4, 11, 4, 2, 2, 4, 0, -1, -1, -1, -1, -1, -1, -1},
            {11, 7, 4, 11, 4, 2, 8, 3, 4, 3, 2, 4, -1, -1, -1, -1},
            {2, 9, 10, 2, 7, 9, 2, 3, 7, 7, 4, 9, -1, -1, -1, -1},
            {9, 10, 7, 9, 7, 4, 10, 2, 7, 8, 7, 0, 2, 0, 7, -1},
            {3, 7, 10, 3, 10, 2, 7, 4, 10, 1, 10, 0, 4, 0, 10, -1},
            {1, 10, 2, 8, 7, 4, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
            {4, 9, 1, 4, 1, 7, 7, 1, 3, -1, -1, -1, -1, -1, -1, -1},
            {4, 9, 1, 4, 1, 7, 0, 8, 1, 8, 7, 1, -1, -1, -1, -1},
            {4, 0, 3, 7, 4, 3, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
            {4, 8, 7, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
            {9, 10, 8, 10, 11, 8, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
            {3, 0, 9, 3, 9, 11, 11, 9, 10, -1, -1, -1, -1, -1, -1, -1},
            {0, 1, 10, 0, 10, 8, 8, 10, 11, -1, -1, -1, -1, -1, -1, -1},
            {3, 1, 10, 11, 3, 10, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
            {1, 2, 11, 1, 11, 9, 9, 11, 8, -1, -1, -1, -1, -1, -1, -1},
            {3, 0, 9, 3, 9, 11, 1, 2, 9, 2, 11, 9, -1, -1, -1, -1},
            {0, 2, 11, 8, 0, 11, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
            {3, 2, 11, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
            {2, 3, 8, 2, 8, 10, 10, 8, 9, -1, -1, -1, -1, -1, -1, -1},
            {9, 10, 2, 0, 9, 2, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
            {2, 3, 8, 2, 8, 10, 0, 1, 8, 1, 10, 8, -1, -1, -1, -1},
            {1, 10, 2, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
            {1, 3, 8, 9, 1, 8, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
            {0, 9, 1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
            {0, 3, 8, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
            {-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1}
        }; // of triTable
        #endregion
    }
}
