using System;
using System.Collections.Generic;
using System.Text;
using Zcu.Mve.Core;
using Zcu.Mve.Visualization;
using System.Diagnostics;
using System.IO;
using Zcu.Mve.Numerics;
using System.ComponentModel;
using System.Drawing;


namespace Zcu.Mve.Visualization.Renderers
{
    /// <summary>
    /// Wrapper for Pre-Integrated Volume Renderer.
    /// </summary>
    [ModuleInfo("Martin Holub","Wrapper for Pre-Integrated Volume Renderer", CategoryType.Output, CategoryData.Volume)]
    [PortInfo("Input3DGrid", "Three dimensional RegGridND to display")]
    [PortInfo("PaletteInput", "Scalar value to be colored")]
    [PortInfo("PaletteOutput", "Color for scalar value")]
    public class VolumeRendererWrapper:Module
    {

        #region private attributes

        private const byte UBYTE_MAX_VALUE = 255;
        private const ushort USHORT_MAX_VALUE = 65535;
        
        private string scalarAttributeName = "";                            //name of cellAttrs to be rendered
        private string directory = @".\data\volume_renderer\";             //path to directory where volume renderer is
        private string file = "volume.exe";                                 //name of execute file of volume renderer
        private string targetFile = "";                                     //name of file where data for volume renderer will be
        private string anotherArguments = "";                               //another arguments for volume renderer
        
        private bool colored = false;                                       //if true, color map will be used

        private string arguments = "";                                      //all arguments  for volume renderer in one string
        
        private OutputDataType outDataType = OutputDataType.UShort;         //indicate if data for volume renderer will be represented as a byte or ushort

       

        /// <summary>
        /// Indicates what data type have to be data for Pre-Integrated Volume Renderer.
        /// </summary>
        public enum OutputDataType { 
            
            /// <summary>
            /// Data for Pre-Integrated Volume Renderer will be represented as a byte
            /// </summary>
            Byte, 
            
            /// <summary>
            /// Data for Pre-Integrated Volume Renderer will be represented as a unsigned short
            /// </summary>
            UShort };
        
        #endregion

        #region properties

            /// <summary>
        /// Indicates if data from RegGridND will be represented as a byte or ushort
        /// </summary>
        [Category("Data Options")]
        [Description("Indicates if data from RegGridND will be represented as a byte or ushort.")]
        public OutputDataType OutDataType
        {
            get { return outDataType; }
            set
            {
                if ((value == OutputDataType.UShort) && (colored == true))
                {
                    throw new MveException(this, "Colored is true, data can be byte only");
                }
                else
                {
                    outDataType = value;
                }
            }
        }
                
        /// <summary>
        /// If true, the color map will be created and used. Module needs to be connected with any palette
        /// </summary>
        [Category("Color Options")]
        [Description("If true, the color map will be created and used. Module needs to be connected with any palette")]
        public bool Colored
        {
            get { return colored; }
            set
            {
                if (value != colored)
                {
                    colored = value;

                    if (colored)
                    {
                        this.AddInPort("PaletteOutput", typeof(ColorRGBA), true, false);
                        this.AddOutPort("PaletteInput", typeof(Scalar), false);
                        outDataType = OutputDataType.Byte;
                    }
                    else
                    {
                        this.RemoveInPort("PaletteOutput");
                        this.RemoveOutPort("PaletteInput");
                    }
                }

            }
        }

        
        /// <summary>
        /// Name of target file, where data will be saved
        /// </summary>
        [Category("Renderer Setup")]
        [Description("Name of target file, where data will be saved")]
        public string TargetFile
        {
            get { return targetFile; }
            set
            {
                if (value == null)
                {
                    throw new MveException(this, "TargetFile has null value");
                }
                else
                {
                    targetFile = value;
                }
            }
        }
        


        /// <summary>
        /// Another arguments for renderer
        /// </summary>
        [Category("Renderer Setup")]
        [Description("Another arguments for renderer")]
        public string AnotherArguments
        {
            get { return anotherArguments; }
            set
            {
                if (value == null)
                {
                    throw new MveException(this, "AnotherArguments has null value");
                }
                else
                {
                    anotherArguments = value;
                }
            }
        }


        

        
        /// <summary>
        /// Working directory, where the execute file of Pre-Integrated Volume Renderer is.
        /// </summary>
        [Category("Renderer Setup")]
        [Description("Working directory, where the execute file of Pre-Integrated Volume Renderer is.")]
        public string Directory
        {
            get { return directory; }
            set
            {
                if (value == null)
                {
                    throw new MveException(this, "Directory has null value");
                }
                else
                {
                    directory = value;
                }
            }
        }
                
        

        /// <summary>
        /// Execute file of Pre-Integrated Volume Renderer
        /// </summary>
        [Category("Renderer Setup")]
        [Description("Execute file of Pre-Integrated Volume Renderer")]
        public string File
        {
            get { return file; }
            set
            {
                if (value == null)
                {
                    throw new MveException(this, "File has null value");
                }
                else
                {
                    file = value;
                }
            }
        }
        
        
        /// <summary>
        /// Name of scalar attributes
        /// </summary>
        [Category("Data Options")]
        [Description("Name of scalar cell attribute to be rendered")]
        public string ScalarAttributeName
        {
            get { return scalarAttributeName; }
            set
            {
                if (value == null)
                {
                    throw new MveException(this, "ScallarAttributeName has null value");
                }
                else
                {
                    scalarAttributeName = value;
                }
            }
        }

        #endregion

        #region constructor

        /// <summary>
        /// Contructor
        /// </summary>
        public VolumeRendererWrapper()
        {
            AddInPort("Input3DGrid", typeof(RegGridND));
        }

        #endregion

        #region Execute method

        /// <summary>
        /// Main method of this module
        /// </summary>
        public override void Execute()
        {
            if ((IsAllDataSame() == false) || (IsAllUserSettingsSame() == false))
            {
                arguments = "";

                int[] dimensions = new int[1];
                double[] thickness = new double[1];
               
                RegGridND grid = (RegGridND)GetInput("Input3DGrid");

                try
                {
                    CheckForErrors();
                    dimensions = AnalyzeGridDimensions(grid);

                    if (scalarAttributeName == "")
                    {
                        foreach (string s in grid.GetCellAttrIds())
                        {
                            if ((grid.GetCellAttrs(s).ElemType == typeof(Scalar)) || (grid.GetCellAttrs(s).ElemType == typeof(Integer)))
                                scalarAttributeName = s;
                        }

                        if (scalarAttributeName == "")
                        {
                            throw new MveException(this, "Input RegGridND contains no cell attribute of type Scalar or Integer.");
                        }

                    }
                }
                catch (MveException exception)
                {
                    Console.WriteLine(exception.Message);
                    Console.WriteLine("Module stopped");
                    return;
                }

                UniformDataArrayND inputCells = (UniformDataArrayND)grid.GetCellAttrs(scalarAttributeName);

                UniformDataArrayND cells = (UniformDataArrayND)inputCells.DeepCopy();
                
                thickness = AnalyzeGridThickness(grid);
                ProcessData(cells, dimensions, thickness);

                CreateColorMap(colored);
            }

            RunRenderer();
        }

        #endregion

        #region private methods

        /// <summary>
        /// Checks for possible errors and throws MveException
        /// </summary>
        private void CheckForErrors()
        {
            if (file == "")
            {
                throw new MveException(this, "The name of the execute file of volume rederer is not set");
            }

            if (directory == "")
            {
                throw new MveException(this, "The path to the directory of volume renderer is not set");
            }

            if (targetFile == "")
            {
                throw new MveException(this, "The name of target file for data for volume renderer is not set");
            }
        }

        /// <summary>
        /// Creates and saves a color map for dataset.
        /// </summary>
        /// <param name="create">if true, color map will be created and saved</param>
        private void CreateColorMap(bool create)
        {
            float o = 0;
            ColorRGBA color = new ColorRGBA();
            if (create == true)
            {
                string targetTABLE = targetFile + ".table";

                StreamWriter sw = new StreamWriter(directory + targetTABLE);

                for (int i = 0; i < 256; i++)
                {
                    o = (float)(i / 255.0f);
                    SetOutput("PaletteInput", (Scalar)o);
                    UpdateInput("PaletteOutput");
                    color = (ColorRGBA)GetInput("PaletteOutput");

                    sw.WriteLine(color.Red + " " + color.Green + " " + color.Blue + " " + color.Alpha);
                }
                sw.Close();
               
                arguments = " " + targetTABLE;
            }
        }

        /// <summary>
        /// Analyze scaling for all dimensions
        /// </summary>
        /// <param name="grid">grid to be analyzed</param>
        /// <returns>double field with scale for all dimensions</returns>
        private double[] AnalyzeGridThickness(RegGridND grid)
        {
            PointND first = grid.GetPoint(0, 0, 0);
            PointND second = grid.GetPoint(1, 1, 1);

            double[] thickness = new double[3];

            thickness[0] = second.Values[0] - first.Values[0];
            thickness[1] = second.Values[1] - first.Values[1];
            thickness[2] = second.Values[2] - first.Values[2];

            return thickness;
        }


        /// <summary>
        /// Analyze how many dimensions given grid has.
        /// </summary>
        /// <param name="grid">grid to be analyzed</param>
        /// <returns>int field with dimensions length of given grid</returns>
        private int[] AnalyzeGridDimensions(RegGridND grid)
        {
            int[] dimensions = grid.Dimensions;

            if (dimensions.Length != 3)
            {
                throw new MveException(this, "Grid dimension must be 3!");
            }

            return dimensions;
        }


        /// <summary>
        /// Process given data
        /// </summary>
        /// <param name="cells">field with data for volume renderer</param>
        /// <param name="dimensions">dimensions of cells field</param>
        /// <param name="thickness">scaling of dataset (slice thickness)</param>
        private void ProcessData(UniformDataArrayND cells, int[] dimensions, double[] thickness)
        {
            if (outDataType == OutputDataType.Byte)
            {
                if (cells[0, 0, 0] is Integer)
                {
                    ResampleData(cells, dimensions, (Integer)cells[0, 0, 0], UBYTE_MAX_VALUE);
                    CreateRawFile(cells, dimensions, (Integer)cells[0, 0, 0], UBYTE_MAX_VALUE);
                    CreateDatFile(dimensions, thickness, "BYTE");
                }
                else
                {
                    ResampleData(cells, dimensions, (Scalar)cells[0, 0, 0], UBYTE_MAX_VALUE);
                    CreateRawFile(cells, dimensions, (Scalar)cells[0, 0, 0], UBYTE_MAX_VALUE);
                    CreateDatFile(dimensions, thickness, "BYTE");
                }

            }
            if (outDataType == OutputDataType.UShort)
            {
                if (cells[0, 0, 0] is Integer)
                {
                    ResampleData(cells, dimensions, (Integer)cells[0, 0, 0], USHORT_MAX_VALUE);
                    CreateRawFile(cells, dimensions, (Integer)cells[0, 0, 0], USHORT_MAX_VALUE);
                    CreateDatFile(dimensions, thickness, "USHORT");
                }
                else
                {
                    ResampleData(cells, dimensions, (Scalar)cells[0, 0, 0], USHORT_MAX_VALUE);
                    CreateRawFile(cells, dimensions, (Scalar)cells[0, 0, 0], USHORT_MAX_VALUE);
                    CreateDatFile(dimensions, thickness, "USHORT");
                }
            }
        }

        /// <summary>
        /// If needed, resample input data for specified data type.
        /// </summary>
        /// <param name="cells">field with data for volume renderer</param>
        /// <param name="dimensions">dimensions of cells field</param>
        /// <param name="number">indicates if input data are Integer or Scalar</param>
        /// <param name="maxValue">maximal value for specified data type</param>
        private void ResampleData(UniformDataArrayND cells, int[] dimensions, Integer number, byte maxValue)
        {
            double[] minMax = FindMinMax(cells, dimensions);
            
            if ((minMax[0] < 0) || (minMax[1] > maxValue))
            {
                for (int x = 0; x < dimensions[0]; x++)
                {
                    for (int y = 0; y < dimensions[1]; y++)
                    {
                        for (int z = 0; z < dimensions[2]; z++)
                        {
                            long value = (Integer)cells[x, y, z];
                            cells[x,y,z] = (Integer)(maxValue * (value - minMax[0]) / (minMax[1] - minMax[0]));
                        }
                    }
                }
            }
        }


        private void ResampleData(UniformDataArrayND cells, int[] dimensions, Integer number, ushort maxValue)
        {
            double[] minMax = FindMinMax(cells, dimensions);

            if ((minMax[0] < 0) || (minMax[1] > maxValue))
            {
                for (int x = 0; x < dimensions[0]; x++)
                {
                    for (int y = 0; y < dimensions[1]; y++)
                    {
                        for (int z = 0; z < dimensions[2]; z++)
                        {
                            long value = (Integer)cells[x, y, z];
                            cells[x, y, z] = (Integer)(maxValue * (value - minMax[0]) / (minMax[1] - minMax[0]));
                        }
                    }
                }
            }
        }

        private void ResampleData(UniformDataArrayND cells, int[] dimensions, Scalar number, byte maxValue)
        {
            double[] minMax = FindMinMax(cells, dimensions);

            if ((minMax[0] < 0) || (minMax[1] > maxValue))
            {
                for (int x = 0; x < dimensions[0]; x++)
                {
                    for (int y = 0; y < dimensions[1]; y++)
                    {
                        for (int z = 0; z < dimensions[2]; z++)
                        {
                            double value = (Scalar)cells[x, y, z];
                            cells[x, y, z] = (Scalar)(maxValue * (value - minMax[0]) / (minMax[1] - minMax[0]));
                        }
                    }
                }
            }
        }

        private void ResampleData(UniformDataArrayND cells, int[] dimensions, Scalar number, ushort maxValue)
        {
            double[] minMax = FindMinMax(cells, dimensions);

            if ((minMax[0] < 0) || (minMax[1] > maxValue))
            {
                for (int x = 0; x < dimensions[0]; x++)
                {
                    for (int y = 0; y < dimensions[1]; y++)
                    {
                        for (int z = 0; z < dimensions[2]; z++)
                        {
                            double value = (Scalar)cells[x, y, z];
                            cells[x, y, z] = (Scalar)(maxValue * (value - minMax[0]) / (minMax[1] - minMax[0]));
                        }
                    }
                }
            }
        }

        


        /// <summary>
        /// Searchs minimal and maximal value in data set.
        /// </summary>
        /// <param name="cells">dataset</param>
        /// <param name="dimensions">dimensions of dataset</param>
        /// <returns></returns>
        private double[] FindMinMax(UniformDataArrayND cells, int[] dimensions)
        {
            double min = double.MaxValue;
            double max = double.MinValue;


            for (int x = 0; x < dimensions[0]; x++)
            {
                for (int y = 0; y < dimensions[1]; y++)
                {
                    for (int z = 0; z < dimensions[2]; z++)
                    {


                        double value = (Integer)cells[x, y, z];

                        if (value > max)
                        {
                            max = value;

                        }

                        if (value < min)
                        {
                            min = value;
                        }
                    }
                }
            }

            return new double[2] { min, max };
        }

        /// <summary>
        /// Creates a .dat file filled with informations about dataset
        /// </summary>
        /// <param name="dimensions">dimensions of dataset</param>
        /// <param name="thickness">scaling of dataset</param>
        /// <param name="type">data type of elements (e.g. BYTE, USHORT...)</param>
        private void CreateDatFile(int[] dimensions, double[] thickness, string type)
        {
            string targetDAT = targetFile + ".dat";
            string targetRAW = targetFile + ".raw";

            StreamWriter strw = new StreamWriter(directory + targetDAT);

            strw.WriteLine("ObjectFileName: " + targetRAW);
            strw.WriteLine("TaggedFileName: ---");
            strw.WriteLine("Resolution:      " + dimensions[0] + " " + dimensions[1] + " " + dimensions[2]);
            string th = "SliceThickness: " + thickness[0] + " " + thickness[1] + " " + thickness[2];
            th = th.Replace(',', '.');
            strw.WriteLine(th);
            strw.WriteLine("Format:         " + type);
            strw.WriteLine("NbrTags:        0");
            strw.WriteLine("ObjectType:     TEXTURE_VOLUME_OBJECT");
            strw.WriteLine("ObjectModel:    DENSITY");
            strw.WriteLine("GridType:       EQUIDISTANT");
            strw.Close();

        }

        


        /// <summary>
        /// Creates a .raw file filled with raw data
        /// </summary>
        /// <param name="cells">dataset</param>
        /// <param name="dimensions">dimensions of dataset</param>
        /// <param name="number">one number of type Scalar</param>
        /// <param name="maxValue">max value of ushort</param>
        private void CreateRawFile(UniformDataArrayND cells, int[] dimensions, Scalar number, ushort maxValue)
        {

            string targetRAW = targetFile + ".raw";

            FileStream fs = new FileStream(directory + targetRAW, FileMode.Create);
            BinaryWriter bw = new BinaryWriter(fs);

            for (int x = 0; x < dimensions[0]; x++)
            {
                ProgressInfo((float)x / dimensions[0]);
                for (int y = 0; y < dimensions[1]; y++)
                {
                    for (int z = 0; z < dimensions[2]; z++)
                    {
                        double value = (Scalar)cells[x, y, z];
                        bw.Write((ushort)value );
                    }
                }
            }

            bw.Close();
        }

        /// <summary>
        /// Creates a .raw file filled with raw data
        /// </summary>
        /// <param name="cells">dataset</param>
        /// <param name="dimensions">dimensions of dataset</param>
        /// <param name="number">one number of type Scalar</param>
        /// <param name="maxValue">max value of byte</param>
        private void CreateRawFile(UniformDataArrayND cells, int[] dimensions, Scalar number, byte maxValue)
        {

            string targetRAW = targetFile + ".raw";

            FileStream fs = new FileStream(directory + targetRAW, FileMode.Create);
            BinaryWriter bw = new BinaryWriter(fs);

            for (int x = 0; x < dimensions[0]; x++)
            {
                ProgressInfo((float)x / dimensions[0]);
                for (int y = 0; y < dimensions[1]; y++)
                {
                    for (int z = 0; z < dimensions[2]; z++)
                    {
                        double value = (Scalar)cells[x, y, z];
                        bw.Write((byte)value);
                    }
                }
            }

            bw.Close();
        }

        /// <summary>
        /// Creates a .raw file filled with raw data
        /// </summary>
        /// <param name="cells">dataset</param>
        /// <param name="dimensions">dimensions of dataset</param>
        /// <param name="number">one number of type Integer</param>
        /// <param name="maxValue">max value of ushort</param>
        private void CreateRawFile(UniformDataArrayND cells, int[] dimensions, Integer number, ushort maxValue)
        {

            string targetRAW = targetFile + ".raw";

            FileStream fs = new FileStream(directory + targetRAW, FileMode.Create);
            BinaryWriter bw = new BinaryWriter(fs);

            for (int x = 0; x < dimensions[0]; x++)
            {
                ProgressInfo((float)x / dimensions[0]);
                for (int y = 0; y < dimensions[1]; y++)
                {
                    for (int z = 0; z < dimensions[2]; z++)
                    {
                        long value = (Integer)cells[x, y, z];
                        bw.Write((ushort)value);
                    }
                }
            }

            bw.Close();
        }

        /// <summary>
        /// Creates a .raw file filled with raw data
        /// </summary>
        /// <param name="cells">dataset</param>
        /// <param name="dimensions">dimensions of dataset</param>
        /// <param name="number">one number of type Integer</param>
        /// <param name="maxValue">max value of ushort</param>
        private void CreateRawFile(UniformDataArrayND cells, int[] dimensions, Integer number, byte maxValue)
        {

            string targetRAW = targetFile + ".raw";

            FileStream fs = new FileStream(directory + targetRAW, FileMode.Create);
            BinaryWriter bw = new BinaryWriter(fs);

            for (int x = 0; x < dimensions[0]; x++)
            {
                ProgressInfo((float)x / dimensions[0]);
                for (int y = 0; y < dimensions[1]; y++)
                {
                    for (int z = 0; z < dimensions[2]; z++)
                    {
                        long value = (Integer)cells[x, y, z];
                        bw.Write((byte)value);
                    }
                }
            }

            bw.Close();

            
        }

           
        /// <summary>
        /// Runs a thread with Pre-Integrated Volume renderer
        /// </summary>
        private void RunRenderer()
        {
            string targetDAT = targetFile + ".dat";

            arguments = targetDAT + arguments + " " + anotherArguments;

            ProcessStartInfo startInfo = new ProcessStartInfo();
            startInfo.WorkingDirectory = directory;
            startInfo.FileName = directory + file;
            startInfo.Arguments = arguments;
            Process.Start(startInfo);
        }

        #endregion
    }

}
